! VNY 25.01.2002; The subroutine dlocate can crash if the
! input xx-array consists of only one element. 
! Protection against this crash has been added
! VNY 29.01.03 Subroutine Marcquardts has been modified (adjusted
! in accordance with the Numerical Recipes book
! VNY 26.01.2004 Changes during the visit to Vilspa 26-28.01.2004
! VNY 02.02.2004 Removing flux calculation for the wavelengths outside
! the range defined by CCF's, that is, 1800-3600 for Grism1 and
! 3000-6000 Angstrom for Grism2
! VNY 06.02.2004 Computing the src/bkg ratio
!
module omgextraction_mod

  use dal
  use cal
  use sasdal
  use types
  use errorhandling 
  use sort_mod

  implicit none

  public

contains


!***********************************************************************
! ExtractSpectrum 
! For extraction of one spectrum from grismSetH using the x and y 
! coordinates of the zero-order image and the extraction region parameters.
!***********************************************************************
! Input parameters:
! -----------------
! grismSetH - handle to the input image, which is rotated so that
!             the dispersion direction is aligned with respect to
!             the y-direction
! xZero - x-coordinate of the image of the zero-order spectrum
! yZero - y-coordinate of the image of the zero-order spectrum 
! spectrumHalfWidth - half-width of the spectrum extraction region  
! bkgOffsetLeft - offset of the left background extraction region from
!             the left edge of the spectrum extraction region 
! bkgWidthLeft - width of the left background extraction region 
! bkgOffsetRight - offset of the  right background extraction region
!             from the right edge of the spectrum extraction region
! bkgWidthRight - width of the right background extraction region
! size -      length of the extraction regions (must be the size
!             in the vertical direction of the input image
! Output parameters:
! ------------------ 
! wavelengthArray - wavelengths corresponding to the extracted spectrum
! spectrumArray - extracted spectrum 
! bkgArray - extracted background (average from the left and right
!            background strips
!******************************************************************
  subroutine ExtractSpectrum(grismSetH, xZero, yZero, spectrumHalfWidth, &
       bkgOffsetLeft, bkgWidthLeft, bkgOffsetRight, bkgWidthRight, &
       wavelengthArray, spectrumArray, spectrumErrArray, bkgArray, &
       bkgErrArray, fluxArray, fluxErrArray, sizeY, extractionMode, &
       subtractBkg)
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xZero, yZero
    real(kind=double), intent(in) :: spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft
    real(kind=double), intent(in) :: bkgOffsetRight, bkgWidthRight 
    integer, intent(in) :: sizeY, extractionMode
    logical, intent(in) :: subtractBkg
    real(kind=double), intent(out), dimension(0:sizeY-1) :: wavelengthArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: spectrumErrArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: bkgArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: bkgErrArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: fluxArray
    real(kind=double), intent(out), dimension(0:sizeY-1) :: fluxErrArray
    !-----
    integer :: i, j, ny, nx
    real(kind=double) :: deltaY
    real(kind=double), dimension(:), pointer :: countsScaleFactor
    real(kind=double), dimension(:), pointer :: bkgVectR, bkgVectL, specVect
    real(kind=double), dimension(:), pointer :: bkgVect2, bkgVect3
    real(kind=double), dimension(:), pointer :: smoothBkgVectL, smoothBkgVectR
    !-----
    real(kind=double), dimension(:), pointer :: lambda
    real(kind=double), dimension(:), pointer :: isfLambda
    real(kind=double), dimension(:), pointer :: isf
    real(kind=double), dimension(:), pointer :: isfError
    integer(kind=int32) :: filterIdInt !+
    integer(kind=int32) ccdMode !+  
    !----

    character(len=9) :: filter !+                         ! filter keyword in the input file
    integer(kind=int32) :: binX                                        ! Bin x.
    integer(kind=int32) :: binY                                       ! Bin y.
    integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
    integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
    logical :: binBpe !+
    integer :: jbkg1, jbkg2, nPixBkgR, nPixBkgL, jcross
    integer :: jyspc1, jyspc2
    real(kind=double) :: averBkgR, averRight,numRight, numBkgR
    real(kind=double) :: averBkgL, averLeft, numLeft, numBkgL
    type(ArrayT) :: inArrayH   
    real(kind=single) :: exposureTime         ! The exposure time for an input image
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    integer :: xBkgL, xBkgR
    integer :: jspc1, jspc2, nPixSpec, nPts, iR, iL
    real(kind=double) :: source2spectrum, spectrumLength
    real(kind=double) :: averStrip, numStrip, averSpec, numSpec
    real(kind=double) :: levelLR, nLevel, a0, a1, wSpec, dSpec
    real(kind=double) :: sum1, sum2, sumTest, sumGauss, sumWeight
    real(kind=double), dimension (:), pointer :: xinput
    real(kind=double), dimension (:), pointer :: yinput
    integer :: ios, isfSize, lambdaSize, jZero
    real(kind=double) :: modZero
    real(kind=double) :: avStrip, avStripN
    integer :: jstart, jend
    !--- VNY 08.01.2004
    real(kind=double), dimension(9) :: param
    integer(kind=int32) :: numparam, numpoints
    !---
    real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
    real(kind=double), dimension(9) ::dydptest ! this must be uncomm. for opt. extr.
    !
    real(kind=double), dimension (:), pointer :: yp
    real(kind=double), dimension (:), pointer :: ysp
    real(kind=double), dimension (:), pointer :: temp
    real(kind=double), dimension (:), pointer :: dispersion
    real(kind=double) :: slp1, slpn, tension
    real(kind=double) :: eps,smooth, delta, averSpecOpt, averSpecLowess, averSpecGauss
    integer(kind=int32) :: ierr, islpsw, numsteps
    real(kind=double) :: sbRatio
    !
    numparam=9
    !   real :: sigm ! this must be uncomm. for opt. extr. 
    !---
    !   real :: sum1, sum2, sumTest, sumGauss, sumWeight
    !---
    if (.not. subtractBkg) then
       call warning('OMGRISM','In this version the background rates are subtracted from the spectrum')
    endif
    !---VNY 28.01.2004
    if (extractionMode .gt. 2) then
       call warning('OMGRISM','This version works with extractionMode < 3')
    endif

    ! the image contains the information on the science window
    inArrayH = array(grismSetH, "PRIMARY", read)
    exposureTime = real32Attribute(grismSetH, "EXPOSURE") ! Image exposure time
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  
    !ny=sizeY
    !nx=sizeX
    allocate(countsScaleFactor(0:ny-1))
    if (hasAttribute(grismSetH,"FILTER")) then
       filter = stringAttribute(grismSetH, "FILTER")
    else
       call warning('OMGRISM','Keyword FILTER missing in the input file ')
    endif

    spectrumArray=0.0
    spectrumErrArray=0.0
    bkgArray=0.0
    bkgErrArray=0.0
    fluxArray=0.0
    fluxErrArray=0.0

    ccdMode = IMAGE

    !**************************************
    ! Get the filter Id integer value
    !**************************************
    filterIdInt = filterIdToInt(filter)
    !*******************
    ! Set the CAL state
    !********************
    call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)
    !
    allocate(lambda(0:ny-1))
    allocate(isfLambda(0:ny-1))
    allocate(isf(0:ny-1))
    allocate(isfError(0:ny-1))
    lambda=0.0
    ! VNY 21.11.03
    ios=1
    !  open (1,file="isfsas_uv.dat", iostat=ios)
    !  write(*,*) "OMGEXTRACTION L140 iostat=", ios
    !  close(1)

    call CAL_omGetGrismWavelength(lambda)
    call CAL_omGetGrismFlux(isfLambda, isf, isfError)
    isfSize=size(isfLambda)
    lambdaSize=size(lambda)
    !write(*,*) " isfL(0)=", isfLambda(0), " isfL(isfSize-1)=", isfLambda(isfSize-1)
    !write(*,*) " isf(0)=", isf(0), " isf(isfSize-1)=", isf(isfSize-1)

    ! Calculate image binning.  
    binBpe= .false.
    if (hasAttribute(grismSetH,"BINBPE")) then       
       binBpe= booleanAttribute(grismSetH, "BINBPE")
    else
       call warning('OMGRISM','Keyword BINBPE missing in the input file ')
    endif
    if (hasAttribute(grismSetH,"BINAX1")) then       
       binAx1 = int32Attribute(grismSetH, "BINAX1")
       binX = binAx1
       if (binX .lt. 1.0) binX = 1.0
    else
       call warning('OMGRISM','Keyword BINAX1 missing in the input file ')
       binX=1.0
    endif

    if (hasAttribute(grismSetH,"BINAX2")) then       
       binAx2 = int32Attribute(grismSetH, "BINAX2")
       binY = binAx2
       if (binY .lt. 1.0) binY = 1.0
    else
       call warning('OMGRISM','Keyword BINAX2 missing in the input file ')
       binY=1.0
    endif

    if (binBPE) then
       binX = 2.0
       binY = 2.0
    end if

    !write (*,*) "OMGRISMEXTRACTION L201"
    !write(*,*) "OMGRISMEXTRACTION L235: xZero=", xZero, " yZero=", yZero, " size=", sizeY
    !write(*,*) "OMGRISMEXTRACTION L236: binBpe=", binBpe, " binAx1=", binAx1, " binAx2=", binAx2
    !write(*,*) " filter=", filter, " binX=", binX, " binY=", binY

    ! compute zero to first order separation (depending o the filter)
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       source2spectrum= 1100.0 
       spectrumLength= 1000.0
    else
       source2spectrum = 800.0
       spectrumLength = 400.0
    endif
    if (binY .gt. 1.0) then 
       source2spectrum=source2spectrum/binY
       spectrumLength=spectrumLength/binY
    endif

    !write(*,*) " source2spectrum=", source2spectrum, " spectrumLength=", spectrumLength

    wavelengthArray=0.0

    ! Compute wavelengths
    countsScaleFactor=1.0
    do j=0,ny-1
       deltaY=real(j)+0.5-yZero ! yZero here is the Y-coordinate of the source' zero order image
       ! Y-binning is taken into account for calculation of the wavelengths 
       if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
          countsScaleFactor(j)=0.240774*binY+2.0*0.000047387*deltaY*binY*binY ! this is per unbinned pixelY
          wavelengthArray(j)=0.240774*deltaY*binY+0.000047387*deltaY*deltaY*binY*binY+ &
               77.3484 ! UV grism
          wavelengthArray(j)=wavelengthArray(j)*10.0 ! wavelengths in Angstroms
          countsScaleFactor(j)=countsScaleFactor(j)*10.0
       else
          countsScaleFactor(j)=0.5626*binY
          wavelengthArray(j)=0.5626*deltaY*binY+20.0898 
          ! for the visual grism (filter="Grism2 "
          wavelengthArray(j)=wavelengthArray(j)*10.0
          countsScaleFactor(j)=countsScaleFactor(j)*10.0
       endif
    enddo ! j=0,ny-1

    ! VNY 26.11.2003 Computing the wavelengths based on CCF
    wavelengthArray=0.0
    jZero=int(yZero)
    ! the wavelength value of each pixcel should be interpolated
    modZero=1.0+yZero-real(jZero)

    do j=jZero, ny-1
       i=(j-jZero)*binY
       if (i .eq. 0) then
          wavelengthArray(j)=lambda(i)-(lambda(i+1)-lambda(i))+(lambda(i+1)-lambda(i))*modZero
       else
          if (i .lt. lambdaSize) then
             wavelengthArray(j)=lambda(i-1)+(lambda(i)-lambda(i-1))*modZero
          endif
       endif
    enddo ! j=jZero+1,ny-1

    !  write(*,*) " ************************ "
    !  do j=0, 10
    !     write(*,*) " j=", j, " isfLambda=", isfLambda(j), " isf=", isf(j), " isfError=", isfError(j)
    !  enddo
    !  do j=isfSize-10, isfSize-1
    !     write(*,*) " j=", j, " isfLambda=", isfLambda(j), " isf=", isf(j), " isfError", isfError(j)
    !  enddo

    ! Compute background around the spectrum strip
    ! Doing the right strip of the background
    jbkg1=xZero+spectrumHalfWidth+bkgOffsetRight
    jbkg2=jbkg1+bkgWidthRight
    iR=(jbkg1+jbkg2)/2 ! iR and iL are for Gaussian fit
    nPixBkgR=jbkg2-jbkg1+1
    allocate(bkgVectR(0:ny-1))
    allocate(bkgVectL(0:ny-1))
    allocate(specVect(0:ny-1))
    allocate(bkgVect2(0:ny-1))
    allocate(bkgVect3(0:ny-1))
    allocate(smoothBkgVectL(0:ny-1))
    allocate(smoothBkgVectR(0:ny-1))
    bkgVectR=0.0
    averBkgR=0.0
    numBkgR=0.0
    if (jbkg1 .lt. nx-1) then
       do j=0, ny-1
          averRight=0.0
          numRight=0.0
          do jcross=jbkg1,jbkg2
             if (jcross .le. nx-1) then
                if (oswImage(jcross,j) .gt. -0.9) then
                   averRight=averRight+oswImage(jcross,j)
                   numRight=numRight+1.0
                endif
             endif
          enddo
          if (numRight .gt. 0.0) then
             averRight=averRight/numRight
          endif
          bkgVectR(j)=averRight
          averBkgR=averBkgR+bkgVectR(j)
          numBkgR=numBkgR+1.0  
       enddo
    endif
    if (numBkgR .gt. 0.0) then 
       averBkgR=averBkgR/numBkgR
       xBkgR=jbkg1
    else
       xBkgR=0
    endif

    ! Doing the left strip of the background
    jbkg1=xZero-spectrumHalfWidth-bkgOffsetLeft
    jbkg2=jbkg1-bkgWidthLeft
    iL=(jbkg2+jbkg1)/2
    nPixBkgL=jbkg1-jbkg2+1
    bkgVectL=0.0
    averBkgL=0.0
    numBkgL=0.0
    if (jbkg1 .gt. 0) then
       do j=0, ny-1
          averLeft=0.0
          numLeft=0.0
          do jcross=jbkg1,jbkg2,-1
             if (jcross .ge. 0 .and. jcross .lt. nx-1) then
                if (oswImage(jcross,j) .gt. -0.9) then
                   if (jcross .lt. nx-1) then
                      averLeft=averLeft+oswImage(jcross,j)
                      numLeft=numLeft+1.0
                   endif
                endif
             endif
          enddo
          if (numLeft .gt. 0.0) then
             averLeft=averLeft/numLeft
          endif
          bkgVectL(j)=averLeft ! bkgVectL value is per pixel
          averBkgL=averBkgL+bkgVectL(j)
          numBkgL=numBkgL+1.0
       enddo
    endif
    if (numBkgL .gt. 0.0) then
       averBkgL=averBkgL/numBkgL ! averBkgL is per pixel
       xBkgL=jbkg1
    else
       xBkgL=0
    endif

    ! Compute the average spectrum within the extraction area
    jspc1=int(xZero-spectrumHalfWidth-0.5)
    jspc2=int(xZero+spectrumHalfWidth+0.5)
    nPixSpec=jspc2-jspc1+1
    jyspc1=int(yZero+source2spectrum-spectrumLength/2.0)
    jyspc2=int(yZero+source2spectrum+spectrumLength/2.0)
    averStrip=0.0
    numStrip=0.0
    do j=jyspc1,jyspc2
       if (j .ge. 0 .and. j .le. ny-1) then
          averSpec=0.0
          numSpec=0.0
          do jcross=jspc1,jspc2
             if (jcross .ge. 0 .and. jcross .le. nx-1) then
                if (oswImage(jcross,j) .gt. -0.9) then
                   averSpec=averSpec+oswImage(jcross,j)
                   numSpec=numSpec+1.0
                endif
             endif
          enddo
          ! VNY test 07.01.2004
          !        jstart=jspc1-10
          !        jend=jspc2+10
          !        if (jstart .lt. 0) jstart=0
          !        if (jend .gt. nx-1) jend=nx-1
          !        if (j .eq. jyspc1+10) then
          !           write(*,*) " j=", j, ":"
          !           write(*,*) "------------"
          !           do jcross=jstart, jend
          !              write (*,*) " jcross=", jcross, " i=", oswImage(jcross,j)
          !           enddo
          !           write(*,*) "------------"
          !        endif
          !---
          if (numSpec .gt. 1.0) then
             averSpec=averSpec/numSpec ! per pixel
          endif
          specVect(j)=averSpec
          averStrip=averStrip+averSpec
          numStrip=numStrip+1.0
       endif
    enddo
    if (numStrip .gt. 1.0) averStrip=averStrip/numStrip ! per pixel
    !  write(*,*) " averStrip=", averStrip, " above the bkg=", averStrip-(averBkgL+averBkgR)/2.0

    !*******************************************************************************
    !smoothing the background 
    !
    do j=0,ny-1
       smoothBkgVectL(j)=smoothMedian(bkgVectL,j,ny-1)
       smoothBkgVectR(j)=smoothMedian(bkgVectR,j,ny-1)
    enddo

    do j=0,ny-1
       bkgVectL(j)=smoothAver(smoothBkgVectL,j,ny-1)
       bkgVectR(j)=smoothAver(smoothBkgVectR,j,ny-1)
    enddo

    avStrip=0.0
    avStripN=0.0
    ! Subtract the background
    do j=0,ny-1
       ! Calculation of average value for the strip
       a0=0.0 ! average value
       a1=0.0 ! number of averaging pixels in the strip     
       ! Compute the mean background value 
       levelLR=0.0
       nLevel=0.0
       ! across the left bkg-strip
       if (xBkgL .gt. 0.0) then
          levelLR=levelLR+bkgVectL(j)
          nLevel=nLevel+1.0
       endif
       ! across the right bkg-strip
       if (xBkgR .gt. 0.0) then
          levelLR=levelLR+bkgVectR(j)
          nLevel=nLevel+1.0
       endif
       ! if there is only one background strip then 
       ! levelLR would correspond to this one strip
       if (nLevel .gt. 1.0) levelLR=levelLR/nLevel 

       averSpec=0.0 ! average value for the spectrum cross-section with 
       ! interpolation of the background
       numSpec=0.0
       if (jspc1 .gt. 0 .and. jspc2 .lt. nx-1 .and. jspc2 .gt. jspc1) then
          ! if both edges are within the OSW then interpolate
          ! supposing that the background levels at the left and right
          ! edges of the strip are bkgVectL(j) and bkgVectR(j), respectively
          wSpec=real(jspc2-jspc1+2)
          dSpec=bkgVectR(j)-bkgVectL(j)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          do jcross=jspc1,jspc2
             wSpec=real(jcross-jspc1+1)
             levelLR=bkgVectL(j)+dSpec*wSpec
             if (jcross .lt. nx .and. jcross .gt. 0) then
                if (oswImage(jcross,j) .gt. -0.9) then
                   averSpec=averSpec+oswImage(jcross,j)-levelLR
                   !              averSpec=averSpec+oswImage(jcross,j)-bkgVectR(j)
                endif
             endif
          enddo
          !otherwise compute just the average value    
       else
          do jcross=jspc1,jspc2
             if (jcross .gt. 0 .and. jcross .lt. nx-1) then
                if (oswImage(jcross,j) .gt. -0.9) then
                   averSpec=averSpec+oswImage(jcross,j)-levelLR
                endif
             endif
          enddo
       endif
       ! VNY test 08.01.2004
       jstart=jspc1-10
       jend=jspc2+10
       if (jstart .lt. 0) jstart=0
       if (jend .gt. nx-1) jend=nx-1
       !     if (j .eq. jyspc1+10) then
       !        write(*,*) " j=", j, ":"
       !        write(*,*) "------------"
       !        do jcross=jstart, jend
       !           write (*,*) " jcross=", jcross, " i=", oswImage(jcross,j)
       !        enddo
       !        write(*,*) "------------"
       !        write(*,*) " iR=", iR, " iL=", iL, " iR-il-9=", iR-iL-9
       !     endif
       !---
       !------
       !     if (j .ge. 600 .and. j .lt. 603) then
       !        write(*,*) " j=", j, " averSpec=", averSpec, " levelLR=", levelLR
       !     endif
       !---

       !nPts=iR-iL-9
       nPts=jspc2-jspc1+1
       !if (j .eq. jyspc1+10) then
       !write (*,*) "L510: j=", j, " nPts=", nPts, " jspc1=", &
       !     jspc1, " jspc2=", jspc2, " averSpec=", averSpec
       !endif

       allocate(xinput(1:nPts))
       allocate(yinput(1:nPts))
       !---VNY 15.01.2004
       allocate(yp(1:nPts))
       !allocate(temp(1:nPts*9))
       allocate(temp(1:nPts))
       allocate(ysp(1:nPts))
       allocate(dispersion(1:nPts))

       ! VNY 14.01.2004
       !numparam=4
       numparam=9
       ! Different  extraction modes
       do i=jspc1,jspc2!1, nPts !iL+5, iR-5
          !xinput(i)=real(i)
          xinput(i-jspc1+1)=real(i)
          !        if (i+il+4 .gt. 0 .and. i+il+4 .lt. nx)  then
          if (i .gt. 0 .and. i .lt. nx)  then
             !           yinput(i)=oswImage(i+il+4,j)
             yinput(i-jspc1+1)=oswImage(i,j)
          endif
       enddo

       !  Gaussian fit of the strip
       ! Calculate parameters of the gaussian weight finction
       ! VNY 22.09.2003 Commented to avoid an error message for bright sources
       !                The following line must be uncommented when applying the optimal extraction algorithm
       !                     call gaussiana(xinput,yinput, nPts, param)
       !^^^^^^^^^^^^^^^^^^^^^^
       !---VNY 08.01.2004
!!       if (j .eq. jyspc1+10) then
!!          write(*,*) "%%%%%%%%%%% computing the gaussian approximation for j=", j
!!          !call gaussiana(xinput,yinput, nPts, param)
!!          ! VNY 27.01.2004
!!          call gaussianal(xinput,yinput, nPts, param)
!!          numparam=9
!!          slp1=0.0
!!          slpn=0.0
!!          tension=0.9
!!          islpsw=1 ! means that dispersion is scalar in CURVS
!!          ierr=0
!!          dispersion=10.0 ! dispersion of the data points
!!          smooth=real(2*nPts) ! smoothing factor
!!          eps= 0.8 ! sqrt(2.0/real(nPts))
!!          !!           call curvs(nPts, xinput, yinput, dispersion, islpsw, smooth, &
!!          !!                eps, yp, ysp, tension, temp, ierr)
!!          !tension=0.001
!!          !islpsw=3 
!!          !call curv1(nPts, xinput, yinput, slp1, slpn, islpsw, yp, temp, tension, ierr)
!!          !!           write (*,*) " L523 ierr=", ierr
!!          !        call gaussiana2(xinput,yinput, nPts, param, numparam)
!!          write(*,*) " L565: computed parameters: param=", param
!!          !sum1=gFunc(param, numparam, xinput, yinput, nPts)
!!          !write(*,*) " L517; residual=", sum1
!!          !        write(*,*) " RESULT AFTER CALLING GAUSSIANA:"
!!          !           write(*,*) " ysp=", ysp
!!          !           write(*,*) " tension=", tension
!!          write(*,*) " --------------------------------"
!!
!!          do i=1, nPts
!!             xtest=xinput(i)
!!             ! VNY 14.01.2004
!!             !!           !call fgauss(xtest,param,ytest,dydptest,numparam)
!!             !!           call fgaussup(xtest,param,ytest,dydptest,numparam)
!!             ! VNY 15.01.2003!
!!             !              ytest=curv2(xtest, nPts, xinput, yinput, ysp, tension)
!!             ! VNY 27.01.2004
!!             call fgaussl(xtest,param,ytest,dydptest,numparam)
!!             write(*,*) "i=", i, " x(i)=", xtest, " y(i)=",yinput(i), " ytest=", ytest
!!             !!           !write(*,*) " dydp=", dydptest
!!          enddo
!!          write(*,*) "%%%%%%%%%%%"
!!       endif
       ! VNY 22.01.2004
       !--- 

       averSpecLowess=0.0
       if (extractionMode .eq. 2) then
          !if (j .eq. jyspc1+10) then
          !write(*,*) "%%%%%%%%%%% computing the lowess approximation for j=", j
          smooth=0.3 ! smoothing factor for the lowess approximation algorithm
          numsteps=0
          delta=0.0
          call lowess(xinput, yinput, nPts, smooth, numsteps, delta, &
               ysp, yp, dispersion)

          !write(*,*) " --------------------------------"

          !do i=jspc1, jspc2!1, nPts
          !   xtest=real(i)
          !   write(*,*) "i=", i, " x(i)=", xtest, " y(i)=",yinput(i-jspc1+1), " ytest=", ysp(i-jspc1+1)
          !enddo
          ! Compute averSpecLowess
          averSpecLowess=0.0
          wSpec=real(nPts)!jspc2-jspc1+1)
          dSpec=ysp(nPts)-ysp(1)!bkgVectR(j)-bkgVectL(j)
          if (wSpec .gt. 1.0) dSpec=dSpec/wSpec
          do jcross=jspc1,jspc2!1,nPts
             wSpec=real(jcross-jspc1)
             levelLR=ysp(1)+dSpec*wSpec
             averSpecLowess=averSpecLowess+ysp(jcross-jspc1+1)-levelLR       
          enddo
          !write(*,*) " L608: averSpec=", averSpec, " averSpecLowess=", averSpecLowess
          !write(*,*) "%%%%%%%%%%%"
          !endif
       endif ! extractionMode .eqe. 2
       !--- 

       ! Calculate the sums
       sum1=0.0
       sum2=0.0
       sumTest=0.0
       sumGauss=0.0
       !..........................................
       !
       ! The following peice of code should be uncommented for the optimal extraction algorithm
       !                     ! Optimal extraction
       !                     do i=1, nPts
       !                        xtest=xinput(i)
       !                        !numparam=4
       !                        call fgauss(xtest,param,ytest,dydptest,numparam)
       !                        ytest=ytest-param(1)
       !                        sigm=sqrt(yinput(i))
       !                        !sigm=sqrt(abs(yinput(i)-param(1)))
       !                        !sigm=1.0
       !                        if (sigm .eq. 0.0) sigm=1.0
       !                        ! VNY 6.02.03 Using the smoothed background
       !!                        sum1=sum1+(yinput(i)-param(1))*ytest*ytest/(sigm*sigm) 
       !                        sum1=sum1+(yinput(i)-bkgVect3(j))*ytest*ytest/(sigm*sigm) 
       !                        sum2=sum2+ytest*ytest/(sigm*sigm)
       !!                        sumTest=sumTest+(yinput(i)-param(1))
       !                        sumTest=sumTest+(yinput(i)-bkgVect3(j))
       !                        sumGauss=sumGauss+ytest ! sumGauss is the integral over the gaussian calculated 
       !!                                                 with the use of parameters determined by gaussiana()
       !                     enddo
       !...........................................
       sumWeight=sum1 ! sumWeight is the result of optimal extraction
       if (sum2 .gt. 0.0) then 
          sumWeight=sumWeight/sum2
       endif
       !if (j .gt. y1 .and. j .le. y1+10) then
       !!   write(*,*) "L891; j=", j, "a0=", a0, " a1=", a1, &
       !!        " iL+4=", iL+4, " nPts=", nPts
       !   write(*,*) " param=", param
       !   write(*,*) " sum1=", sum1, " sum2=", sum2, " s1/s2=", sumWeight, &
       !        "sumTest=", sumTest, " sumGauss=", sumGauss
       !   write(*,*) "........"
       !endif

       !if (j .gt. y1+9 .and. j .le. y1+10) then
       !   write(*,*) "%%%%%%%%%%% j=", j
       !   do i=1, nPts
       !      xtest=xinput(i)
       !      call fgauss(xtest,param,ytest,dydptest,numparam)
       !      write(*,*) "i=", i, " y(i)=",yinput(i), " ytest=", ytest
       !   enddo
       !   write(*,*) "%%%%%%%%%%%"
       !endif
       deallocate(xinput)
       deallocate(yinput)
       deallocate(yp)
       deallocate(temp)
       deallocate(ysp)
       deallocate(dispersion)
       !endif !(j .ge. y1 .and. j .le. y2)
       !----------------------------------------------------
       !
       !**********************************
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be uncommented for the optimal extraction algorithm
       !                     rateSourceOutArray(j)=sumWeight  ! Optimal extraction value 
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! rateSourceOutArray(j)=sumGauss ! approximation of data with a Gaussian
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be uncommented for the normal extraction algorithm
       !rateSourceOutArray(j)=a0 !oswImage(j1,i) ! this is just the average across the spectrum
       !                 !    !if (a1 .gt. 0.0) rateSourceOutArray(j)=rateSourceOutArray(j)/a1
       !rateSourceErrOutArray(j)=0.0
       !if (rateSourceOutArray(j) .gt. 1.0) rateSourceErrOutArray(j)=sqrt(rateSourceOutArray(j))
       !---VNY 30.08.03
       !!spectrumArray(j)=a0 ! this is just the average across the spectrum
       ! VNY 13.11.2003
       ! VNY 27.01.2004
       ! For the extractionMode=0 the output spectrum array takes the average across 
       spectrumArray(j)=averSpec ! this is the average with interpolation of the backgroung
       ! If extractionMode=2 then the output is the integral of the spline-smoothed (lowess) cross-section
       if (extractionMode .eq. 2) spectrumArray(j)=averSpecLowess
       !------
       !     if (j .ge. 600 .and. j .lt. 603) then
       !        write(*,*) " j=", j, " averSpec=", averSpec, " levelLR=", levelLR
       !     endif
       !---
       !---
       avStrip=avStrip+averSpec
       avStripN=avStripN+1.0
       !---
       ! across the spectrum
       spectrumErrArray(j)=0.0
       if (spectrumArray(j) .gt. 1.0) spectrumErrArray(j)= &
            sqrt(spectrumArray(j))
       !---
       ! VNY 17.10.2003 here the ISF calibration must go
       ! VNY 18.11.2003 the ISF calibration can be implemented outside this subroutine
       !     fluxOutArray0(j)=nullValue
       !     fluxErrOutArray0(j)=nullValue
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       ! The following line should be  uncommented  for normal extraction
       !if (a1 .ge. 2.0) rateSourceErrOutArray(j)=rateSourceErrOutArray(j)/sqrt(a1-1.0)
       !---
       if (a1 .ge. 2.0) spectrumErrArray(j)=spectrumErrArray(j)/sqrt(a1-1.0)
       !---
       ! The following line should be uncommented for the optimal extraction algorithm
       !                  if (sumWeight .ge. 2.0) rateSourceErrOutArray(j)=rateSourceErrOutArray(j)/sqrt(sumWeight-1.0)
       !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
       !                  rateBackOutArray(j)=levelLR
       ! The following line should be uncommeted for the optimal extraction algorithm
       !                 rateBackOutArray(j)=param(1)
       !.............The three following lines should be uncommented for the normal extraction
       !rateBackOutArray(j)=bkgVect2(j)
       !rateBackErrOutArray(j)=0.0
       !if (rateBackOutArray(j) .gt. 0.0) rateBackErrOutArray(j)=sqrt(rateBackOutArray(j)) 
       !--VNY 30.08.03
       bkgArray(j)=bkgVect2(j)
       ! VNY 13.11.2003 
       bkgArray(j)=levelLR !(bkgVectR(j)+bkgVectL(j))/2

       bkgErrArray(j)=0.0
       if (bkgArray(j) .gt. 0.0) bkgErrArray(j)=sqrt(bkgArray(j)) 
       !---

       !!                  if (levelLR .gt. 0.0) rateBackErrOutArray(j)=sqrt(levelLR)
       ! The following line should be uncommented for the optimal extraction algorithm 
       !                  if (param(1) .gt. 0.0) rateBackErrOutArray(j)=sqrt(param(1))
       !^^^^^^^^^^^^^^^^^^^^^^^
       !  Scaling counts
       if (exposureTime .gt. 0.0) then
          spectrumArray(j)=spectrumArray(j)/exposureTime
          spectrumErrArray(j)=spectrumErrArray(j)/exposureTime
          bkgArray(j)=bkgArray(j)/exposureTime
          bkgErrArray(j)=bkgErrArray(j)/exposuretime
       endif
       if (countsScaleFactor(j) .gt. 0.0) then
          spectrumArray(j)=spectrumArray(j)/countsScaleFactor(j)
          spectrumErrArray(j)=spectrumErrArray(j)/countsScaleFactor(j)
          bkgArray(j)=bkgArray(j)/countsScaleFactor(j)
          bkgErrArray(j)=bkgErrArray(j)/countsScaleFactor(j)
       endif
       ! Compute the flux and its error
       !     call CountsToFlux(wavelengthArray(j), spectrumArray(j), spectrumErrArray(j), &
       !          fluxArray(j), fluxErrArray(j), isfLambda, isf, isfError, isfSize)
    enddo
    ! VNY 05.12.2003
    if (avStripN .gt. 1.0) then
       avStrip=avStrip/avStripN
    endif

    do j=jZero, ny-1
       call CountsToFlux(wavelengthArray(j), spectrumArray(j), spectrumErrArray(j), &
            fluxArray(j), fluxErrArray(j), isfLambda, isf, isfError)
    enddo
    !---
    deallocate(bkgVectR)
    deallocate(bkgVectL)
    deallocate(specVect)
    deallocate(bkgVect2)
    deallocate(bkgVect3)
    deallocate(smoothBkgVectL)
    deallocate(smoothBkgVectR)
    !deallocate(lambda)
    !deallocate(isf)
    !deallocate(isfError)
    deallocate(countsScaleFactor)
  end subroutine ExtractSpectrum

!**********************************************************************
! Subroutine EvaluateSpectrum for evaluation of the signal-to-noise
! ratio within a given limits of the extraction strip
!******************************************************************
  subroutine EvaluateSpectrum(spectrumArray, bkgArray, sizeY, blueEdge, redEdge, sbRatio)
    integer, intent(in) :: blueEdge, redEdge 
    integer, intent(in) :: sizeY
    real(kind=double), intent(in), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(in), dimension(0:sizeY-1) :: bkgArray
    real(kind=double), intent(out) :: sbRatio
    integer :: i, j, correctedRedEdge, redMinusBlue
    real(kind=double) :: averSpec, averBkg
    integer, dimension(0:1024) :: iBlue, iRed
    integer :: nBlue, nRed, nBlueRed, switch, flag
    real(kind=double) :: nAver

    iBlue=0
    iRed=0
    nBlue=0
    nRed=0
    averSpec=0.0
    nAver=0.0
    averBkg=0.0
    sbRatio=0.0
    switch=0
    correctedRedEdge=redEdge
    if (redEdge .ge. sizeY) correctedRedEdge=sizeY-1
    redMinusBlue=correctedRedEdge-blueEdge
    redMinusBlue=int(0.8*real(redMinusBlue))
!    do i=blueEdge,correctedRedEdge
    do i=blueEdge,sizeY-1
       averSpec=averSpec+spectrumArray(i)
       nAver=nAver+1.0
       averBkg=averBkg+bkgArray(i)
    enddo
    if (averBkg .ne. 0.0) then
       sbRatio=(averSpec+averBkg)/averBkg
    endif
    !write(*,*) " L826 averSpec=", averSpec, " averBkg=", averBkg, " sbRatio=", sbRatio
! VNY 12.02.2004 Check if the sbRatio>2 was caused by a zero-order
! within the spectrum extraction area
    if (nAver .gt. 1.0) averSpec=averSpec/nAver
    if (sbRatio .gt. 2.0) then
       ! Search for the maximum
       flag=0
!       do i=blueEdge, correctedRedEdge
       do i=blueEdge, sizeY-1
          flag=0
          if (switch .eq. 0) then
             if (spectrumArray(i) .gt. averSpec) then
                iBlue(nBlue)=i
                nBlue=nBlue+1
                switch=1
                flag=1
             endif! spectrumArray(i) .gt. averSpec
          endif ! switch .eq. 0
          if (switch .eq. 1 .and. flag .eq. 0) then
             if (spectrumArray(i) .lt. averSpec) then
                iRed(nRed)=i
                nRed=nRed+1
                switch=0
             endif ! spectrumArray(i) .lt. averSpec
          endif ! switch .eq. 1
       enddo ! i=blueEdge, correctedRedEdge
       nBlueRed=min(nBlue,nRed)

       flag=1
       do i=0,nBlueRed-1
          !write(*,*) " i=", i, " iBlue=", iBlue(i), " iRed=", iRed(i), " delta=", &
          !     iRed(i)-iBlue(i), " redMinusBlue=", redMinusBlue
          if (iRed(i)-iBlue(i) .ge. redMinusBlue/4) then
             flag=0 ! do not recalculate the sbRatio
          endif
       enddo ! i=0,nBlueRed
       if (flag .eq. 1) then ! recalculate sbRatio
          do i=blueEdge,correctedRedEdge
             ! Check if i lies within one of the exlusion regions
             switch=1
             do j=0,nBlueRed-1
                if (i .ge. iBlue(j)-20 .and. i .le. iRed(j)+20) then
                   switch=0
                endif
             enddo ! j=0,nBluered-1
             if (switch .eq. 1) then
                averSpec=averSpec+spectrumArray(i)
                nAver=nAver+1.0
                averBkg=averBkg+bkgArray(i)
             endif ! switch .eq. 1
          enddo
          if (averBkg .ne. 0.0) then
             sbRatio=(averSpec+averBkg)/averBkg
          endif
       endif ! flag .eq. 1
    endif ! sbRatio .gt. 2.0
  end subroutine EvaluateSpectrum

!******************************************************************
! Subroutine for finding a zero-order position based on the 
! coordinates of the first-order.
! yZero is the output coordinate of the zero-order
! but at the beginning it must contain the expected position
! of the zero-order
!******************************************************************
  subroutine FindZeroOrder(grismSetH, xFirst, yFirst, &
       spectrumHalfLength, yZero, &
       maxValue, spectrumHalfWidth, shapeCriterium1, verbFlag, &
       maxValue2, meanValue2, yZero2)  
    type( DataSetT ),intent(in) :: grismSetH
    real(kind=double), intent(in) :: xFirst, yFirst
    real(kind=double), intent(in) :: spectrumHalfLength
    real(kind=double), intent(in) :: spectrumHalfWidth
    real(kind=double), intent(in) :: shapeCriterium1
    integer, intent(in) :: verbFlag ! auxiliary flag for test output
    real(kind=double), intent(out) :: yZero2 ! 
    real(kind=double), intent(inout) :: yZero
    real(kind=double), intent(out) :: maxValue
    real(kind=double), intent(out) :: maxValue2
    real(kind=double), intent(out) :: meanValue2
    real(kind=double), dimension(:), pointer :: verticalStrip
    integer :: i, j, ny, nx
    integer :: iLeft, iRight, jBlue, jRed, jMaxValue
    real(kind=double) :: stripWidth
    type(ArrayT) :: inArrayH   
    real(kind=single), pointer :: oswImage(:, :)
    integer(kind=int32) :: dim(2)
    real(kind=double) :: maxValueFirst
    real(kind=double) :: convValue, nMean
    integer :: jMaxValue2
    !
    ! the input image image 
    inArrayH = array(grismSetH, "PRIMARY", read)
    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)  
    allocate(verticalStrip(0:ny-1))
    iLeft=int(xFirst-spectrumHalfWidth)
    if (iLeft .lt. 0) iLeft=0
    iRight=int(xFirst+spectrumHalfWidth)
    if (iRight .gt. nx-1) iRight=nx-1
    stripWidth=real(iRight-iLeft+1)
    ! Fill the strip in with the values from the image
    do j=0,ny-1
       verticalStrip(j)=0.0
       do i=iLeft,iRight
          verticalStrip(j)=verticalStrip(j)+oswImage(i,j)
       enddo ! i=iLeft,iRight
       if (stripWidth .gt. 1.0) verticalStrip(j)=verticalStrip(j)/stripWidth
    enddo ! j=0,ny-1
    ! Compute maximum of the spectrum itself
    jBlue=int(yFirst-spectrumHalfLength)
    jRed=int(yFirst+spectrumHalfLength)
    !if (verbFlag .eq.1 ) then
    !   write(*,*) "OMGEXTRACTION L941 jBlue=", jBlue, " jRed=", jRed
    !endif
    if (jBlue .lt. 0) jBlue=0
    if (jRed .gt. ny-1) jRed=ny-1
    maxValueFirst=0.0
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValueFirst) maxValueFirst=verticalStrip(j)
    enddo ! j=jBlue,jRed
    ! Set the interval for searching the maximum near the 
    ! expected position of the zero-order
!    jBlue=int(yZero)-400
!    jRed=jBlue+450
    jBlue=int(yZero)-ny/4.0
    jRed=int(yZero)+ny/4.0
    if (jBlue .lt. 0) jBlue=0
    if (jRed .gt. ny-1) jRed=ny-1
    !if (verbFlag .eq. 1 ) then
    !   write(*,*) " L957 spectrumHalfLength=", spectrumHalfLength
    !   write(*,*) "OMGEXTRACTION L958 jBlue=", jBlue, " jRed=", jRed
    !endif
    ! Search for the maximum in between jBlue and jRed
    maxValue=0.0
    jMaxValue=-1
    do j=jBlue,jRed
       if (verticalStrip(j) .gt. maxValue) then
         !write(*,*) " j=",j, " v=", verticalStrip(j), " ********", &
         !     " current MAX=", maxValue
          maxValue=verticalStrip(j)
          jMaxValue=j
       else
          !write(*,*) " j=",j, " v=", verticalStrip(j)  
       endif
    enddo ! j=jBlue,jRed
    !write(*,*) " L957 jMaxValue=", jMaxValue, " maxValue=", maxValue, " maxValueFirst=", maxValueFirst
    yZero=-1
    if (jMaxValue .gt. -1 .and. maxValue .gt. maxValueFirst*2.0) yZero=jMaxValue
    ! VNY 03.03.2004 Test of the ExpConv
    jMaxValue2=-1
    maxValue2=0.0
    meanValue2=0.0
    nMean=0.0
    do j=jBlue,jRed
       convValue=ConvoluteExp(verticalStrip, ny, j, shapeCriterium1)
       meanValue2=meanValue2+convValue
       nMean=nMean+1
       if (convValue .gt. maxValue2) then
          maxValue2=convValue
          jMaxValue2=j
       endif
       !if (verbFlag .eq. 1) then
       !   write(*,*) " j=", j, " conv=", convValue
       !end if ! verbFlag .eq. 1
    enddo ! j=jBlue, jRed
    if (nMean .gt. 1.0) meanValue2=meanValue2/nMean
    yZero2=real(jMaxValue2)
    deallocate(verticalStrip)
  end subroutine FindZeroOrder

!******************************************************************
! Convolutes an exponent with the spectra at a given point
!******************************************************************
real function ConvoluteExp(spectrumArray, sizeY, currentY, expWidth)
    integer, intent(in) :: sizeY, currentY
    real(kind=double), intent(in), dimension(0:sizeY-1) :: spectrumArray
    real(kind=double), intent(in) :: expWidth
    integer :: k
    real(kind=double) :: a1, a2, sum1, sum2
    sum1=0.0
    sum2=0.0
    do k=0,sizeY-1
       a1=abs(k-currentY)
       a1=2.0*a1/expWidth
       a2=exp(-a1)
       sum1=sum1+spectrumArray(k)*a2
       sum2=sum2+a2
    enddo
    if (sum2 .gt. 0.0) sum1=sum1/sum2
    ConvoluteExp=sum1
  end function ConvoluteExp


!*******************************************************************
! Subroutine for conversion of count rates to fluxes
!*******************************************************************
subroutine CountsToFlux(wavelength, countRate, countRateErr, flux, fluxErr, &
     calWavelength, calFlux, calFluxErr)
real(kind=double), intent(in) :: wavelength
real(kind=double), intent(in) :: countRate
real(kind=double), intent(in) :: countRateErr
real(kind=double), intent(in), dimension(:) :: calWavelength
real(kind=double), intent(in), dimension(:) :: calFlux
real(kind=double), intent(in), dimension(:) :: calFluxErr
real(kind=double), intent(out) :: flux
real(kind=double), intent(out) :: fluxErr
!---
integer :: cSize
integer :: j, indexW
real(kind=double) :: fluxIncrement, wIncrement, extrapolatedFlux, normCountRateErr 
real(kind=double) :: limitWavelength

flux=0.0
fluxErr=0.0
cSize= size(calWavelength)
limitWavelength=calWavelength(csize)

! return if the wavelength is out of range of CAL
if (wavelength .lt. calWavelength(1) .or. wavelength .gt. calWavelength(cSize)) then
   flux=0.0
   fluxErr=0.0
   return
endif

indexW=-1
do j=1,cSize
   if (calWavelength(j) .gt. wavelength) then
      indexW=j
      exit 
   endif
enddo

if (indexW .gt. -1 .and. indexW .ne. cSize .and. &
     wavelength .lt. limitWavelength) then 

   if (indexW .gt. 0 .and. indexW .lt. cSize) then
      fluxIncrement=calFlux(indexW+1)-calFlux(indexW)
      wIncrement=calWavelength(indexW+1)-calWavelength(indexW)
      fluxIncrement=fluxIncrement/wIncrement
      wIncrement=wavelength-calWavelength(indexW)
      extrapolatedFlux=calFlux(indexW)+fluxIncrement*wIncrement
      flux=countRate*extrapolatedFlux
      if (abs(countRate) .gt. 1.0e-6) then
         normCountRateErr=countRateErr/abs(countRate)
      else
         normCountRateErr=0.0
      endif
      fluxErr=calFluxErr(indexW)*calFluxErr(indexW)/10000.0+ &
           normCountRateErr*normCountRateErr
      fluxErr=sqrt(fluxErr)
!      fluxErr=calFluxErr(indexW)/100.0
      fluxErr=countRate*fluxErr*extrapolatedFlux
   endif
   if (indexW .eq. cSize) then
      fluxIncrement=calFlux(indexW)-calFlux(indexW-1)
      wIncrement=calWavelength(indexW)-calWavelength(indexW-1)
      fluxIncrement=fluxIncrement/wIncrement
      wIncrement=wavelength-calWavelength(indexW-1)
      extrapolatedFlux=calFlux(indexW-1)+fluxIncrement*wIncrement
      if (abs(countRate) .gt. 1.0e-6) then
         normCountRateErr=countRateErr/abs(countRate)
      else
         normCountRateErr=0.0
      endif
      flux=countRate*extrapolatedFlux
      fluxErr=calFluxErr(indexW)*calFLuxErr(indexW)/10000.0+ &
           normCountRateErr*normCountRateErr
      fluxErr=sqrt(fluxErr)
!      fluxErr=calFluxErr(indexW)/100.0
      fluxErr=countRate*fluxErr*extrapolatedFlux
   endif
endif

! Find the wavelenght index in cal-array

end subroutine CountsToFlux

!----
! Median smoothing function 11 points
!real function smoothMedian(vect, j, nmax)
!  integer, intent(in) :: nmax
!  real, intent(in) :: vect(0:nmax)
!  integer, intent(in) :: j
!  real :: a(11) ! smoothing for +-5 elements
!  integer :: k,l
!  real :: dum
!  ! Copy the current 11 elements of the vector (indexed +-5)
!  do k=1,11 !j-5,j+5
!     if (j+k-6 .lt. 0 .or. j+k-6 .gt. nmax) then
!        a(k) = 9.E+20
!     else
!        a(k)=vect(j+k-6)
!     endif
!  enddo
!  ! ordering the vector a
!  do k=1,10
!     do l=k+1, 11
!        if (a(l) .lt. a(k)) then
!           dum=a(k)
!           a(k)=a(l)
!           a(l)=dum
!        endif
!     enddo
!  enddo
!  smoothMedian=a(6)
!end function smoothMedian

!----
! Median smoothing function 41 points
real function smoothMedian(vect, j, nmax)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: j
  real(kind=double) :: a(41) ! smoothing for +-5 elements
  integer :: k,l
  real(kind=double) :: dum
  ! Copy the current 11 elements of the vector (indexed +-5)
  do k=1,41 !j-5,j+5
     if (j+k-21 .lt. 0 .or. j+k-21 .gt. nmax) then
        a(k) = 9.E+20
     else
        a(k)=vect(j+k-21)
     endif
  enddo
  ! ordering the vector a
  do k=1,40
     do l=k+1, 41
        if (a(l) .lt. a(k)) then
           dum=a(k)
           a(k)=a(l)
           a(l)=dum
        endif
     enddo
  enddo
  smoothMedian=a(21)
end function smoothMedian

!---smoothing by averaging
real function smoothAver(vect, j, nmax)
  integer, intent(in) :: nmax
  real(kind=double), intent(in) :: vect(0:nmax)
  integer, intent(in) :: j
  integer :: k
  real(kind=double) :: sum, nsum
  sum=0.0
  nsum=0.0
  do k=1,41
     if (j+k-21 .gt. 0 .and. j+k-21 .le. nmax) then
        sum=sum+vect(j+k-21)
        nsum=nsum+1.0
     endif
  enddo
  if (nsum .gt. 0.0) sum=sum/nsum
  smoothAver=sum  
end function smoothAver
!---
!---test of Gaussian function 28.01.03
real(double) function gFunc(param, numparam, xinput, yinput, numpoints)
   integer, intent(in)      :: numparam, numpoints
   real(kind=double), intent(in)   :: param(numparam)
   real(kind=double), intent(in)   :: xinput(numpoints), yinput(numpoints)
   real(kind=double)    :: residual
   integer    :: i
! The function is: y=param(1)+param(2)*exp(-((x-param(3))/param(4))**2)
! The result is the sum of residuals squared
   gFunc=0.0
   do i=1,numpoints
      if (param(4) .gt. 0.0) then
         residual=param(1)+param(2)*exp(-((xinput(i)-param(3))/param(4))**2)
      else
         residual=param(1)+param(2)*exp(-((xinput(i)-param(3)))**2)
      endif
      residual=(yinput(i)-residual)
      gFunc=gFunc+residual*residual
   enddo
end function gFunc

! Test 2 of the Gaussian function
subroutine fgauss(x, p, y, dydp, np)
! The function is: y=p(1)+p(2)*exp(-((x-p(3))/p(4))**2)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  y=p(1)+p(2)*ex
  dydp(1)=1.0
  dydp(2)=ex
  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
end subroutine fgauss

! Superposition of the Gaussian and a linear function
subroutine fgaussl(x, p, y, dydp, np)
! The function is: y=p(1) + p(5)*(x-p(3)) + p(2)*exp(-((x-p(3))/p(4))**2)
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac

  y=0.0
  dydp=0.0
  if (p(4) .eq. 0.0) p(4)=1.0
  arg=(x-p(3))/p(4)
  ex=exp(-arg*arg)
  fac=p(2)*ex*2.0*arg
  y=p(1)+p(2)*ex +p(5)*x
  dydp(1)=1.0
  dydp(2)=ex
  dydp(3)=fac/p(4)
  dydp(4)=fac*arg/p(4)
  dydp(5)=x-p(3)
end subroutine fgaussl

! Superposition  of the Gaussian function
subroutine fgaussup(x, p, y, dydp, np)
!
! The function computes the superposition of various gaussians:
! y(x;a)=\sum_{k=1}^K B_k exp[-((x-E_k)/G_k)**2],
! which is a sum of K Gaussians (np/3), each having a variable position,
! amplitude and width. The parameters are stored in the 
! order B_1,E_1,G_1, B_2,E_2,G_2, ... , B_K,E_K,G_K:
! p(i)=B_k, p(i+1)=E_k, p(i+2)=G_k, k=1,...,np/3
!
  integer(kind=int32), intent(in) :: np ! number of parameters
  real(kind=double), intent(in) :: x ! x-point
  real(kind=double), intent(in out) :: p(np) ! parameters
  real(kind=double), intent(out) :: y ! value of the function
  real(kind=double), intent(out) :: dydp(np) ! derivatives of the function
  real(kind=double) :: arg, ex, fac
  integer(kind=int32) :: i

  y=0.0
  dydp=0.0
  do i=1,np-1,3
     if (p(i+2) .eq. 0.0) p(i+2)=1.0
     arg=(x-p(i+1))/p(i+2)
     ex=exp(-arg*arg)
     fac=p(i)*ex*2.0*arg
     y=y+p(i)*ex
     dydp(i)=ex
     dydp(i+1)=fac/p(i+2)
     dydp(i+2)=fac*arg/p(i+2)
  enddo
  return
end subroutine fgaussup

!---
subroutine gaussiana(xinput,yinput, nPts, param)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a four-parameter gaussian curve y=p1+p2*exp(-(x-p3**2)/p4**2).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(4)
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(4) :: fitList
  real(kind=double), dimension(4,4) :: covMatrix
  real(kind=double), dimension(4,4) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(4) :: atry, beta, da
  real(kind=double), dimension(4,4) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  ! 08.01.2004
  real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
  real(kind=double), dimension(4) ::dydptest ! this must be uncomm. for opt. extr. 
  
  numparam=4
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  do i1=1,nPts
     meany=meany+yinput(i1)
  enddo
  if (nPts .gt. 0) then
     meany=meany/real(nPts)
  else
     write(*,*) "Error (no input points for Gaussian approximation"
  endif
!  write(*,*) " GAUSSIANA L927: meany=", meany

  !Initial values for the input parameters
!  param(1)=0.0 ! 2.0
!  param(2)=meany ! 6.0
!  param(3)=nPts/2.0 ! 10.0
!  param(4)=param(3)/3.0 ! 5.0
  param(1)=meany ! 10.0 ! meany/10.0
  param(2)=10.0 !meany !*1.2
  param(3)=nPts/2.0 ! 10.0
  param(4)=1.0 ! param(3)/3.0 ! 5.0
  !write(*,*) "******************000000000000000000"
  !write(*,*) " L943 param(1)=", param(1), " param(2)=", param(2), &
  !     " param(3)=", param(3), " param(4)=", param(4)
  siginput=6.0
  aLambda=-10

  
  call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
       covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, atry, beta, da, oneda, ochisqr)
  chi0=chiSqr
  !write(*,*) "  First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, " aLambda1=", aLambda 
  !write(*,*) "  First   p1=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "*********************1111111111111111"

  !write(*,*) " Consecutive steps :"
  istep=0
  fchi: do 
     istep=istep+1
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, atry, beta, da, oneda, ochiSqr)
     if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
     chi0=chiSqr
  enddo fchi
  !write (*,*) " Intermediate chiSqr=", chiSqr, " istep=", istep
  !write (*,*) " Intermediate parameters=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************2222222222222222222"
  finchi0=chiSqr
  !write(*,*) " steps :"
  jstep=0
  fchi0 : do
     jstep=jstep+1
     istep=0
     aLambda=-10
     fchi1: do 
        istep=istep+1
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, chiSqr, fgauss, aLambda, atry, beta, da, oneda, ochiSqr)
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
        chi0=chiSqr
     enddo fchi1
     if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
     finchi0=chiSqr
  enddo fchi0
  !write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
  !write (*,*) " Final parameters2=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgauss(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************3333333333333333333"
end subroutine gaussiana

!---

! Approximation of the input data with a Gaussian plus linear functions
!---
subroutine gaussianal(xinput,yinput, nPts, param)
  ! This subroutine approximates a series of input data xinput[1:nPts], yinput[1:nPts]
  ! with a five-parameter gaussian curve superimposed with a linear function 
  ! y=p1+p2*exp(-(x-p3**2)/p4**2) + p(5)*(x-p(3)).
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(5)
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(5) :: fitList
  real(kind=double), dimension(5,5) :: covMatrix
  real(kind=double), dimension(5,5) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(5) :: atry, beta, da
  real(kind=double), dimension(5,5) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1
  real(kind=double)  :: xtest , ytest  
  real(kind=double), dimension(5) ::dydptest 
  real(kind=double) :: maxy
  integer(kind=int32) :: imaxy 
  
  numparam=5
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  maxy=0.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     meany=meany+yinput(i1)
  enddo
  if (nPts .gt. 0) then
     meany=meany/real(nPts)
  else
     write(*,*) "Error (no input points for Gaussian approximation"
  endif
  !write(*,*) " GAUSSIANAL L1216: meany=", meany, " maxy=", maxy, " imaxy=", imaxy

  !Initial values for the input parameters
  param(1)=meany/10.0
  param(2)=maxy
  param(3)=xinput(imaxy)
  param(4)=1.0 
  param(5)=0.0

  !write(*,*) "******************STEP:  000000000000000000"
  !write(*,*) " L1226 param(1)=", param(1), " param(2)=", param(2), &
  !     " param(3)=", param(3), " param(4)=", param(4), " param(5)=", param(5)
  siginput=6.0
  aLambda=-10

  call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
       covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, da, oneda, ochisqr)
  chi0=chiSqr
  !write(*,*) "  First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, " aLambda1=", aLambda 
  !write(*,*) "  First   p1=", param
  do i1=1, nPts
     xtest=xinput(i1)
     call fgaussl(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "*********************STEP: 1111111111111111"

  !write(*,*) " Consecutive steps :"
  istep=0
  fchi: do 
     istep=istep+1
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, da, oneda, ochiSqr)
     if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
     chi0=chiSqr
  enddo fchi
  !write (*,*) " Intermediate chiSqr=", chiSqr, " istep=", istep
  !write (*,*) " Intermediate parameters=", param
  do i1=1, nPts
     xtest=xinput(i1)
     call fgaussl(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************2222222222222222222"
  finchi0=chiSqr
  !write(*,*) " steps :"
  jstep=0
! VNY 28.01.2004
  aLambda=-10
  fchi0 : do
     jstep=jstep+1
     istep=0
! VNY 28.01.2004
!     aLambda=-10
     fchi1: do 
        istep=istep+1
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, chiSqr, fgaussl, aLambda, atry, beta, da, oneda, ochiSqr)
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
        chi0=chiSqr
     enddo fchi1
     if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
     finchi0=chiSqr
  enddo fchi0
  !write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
  !write (*,*) " Final parameters2=", param
  do i1=1, nPts
     xtest=xinput(i1)
     call fgaussl(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************3333333333333333333"
end subroutine gaussianal

!---


subroutine gaussiana2(xinput,yinput, nPts, param, numparam)
! This subroutine approximates a series of input data 
! xinput[1:nPts], yinput[1:nPts]
! with a superposition of various gaussians:
! y(x;a)=\sum_{k=1}^K B_k exp[-((x-E_k)/G_k)**2],
! which is a sum of K Gaussians (np/3), each having a variable position,
! amplitude and width. The parameters are stored in the 
! order B_1,E_1,G_1, B_2,E_2,G_2, ... , B_K,E_K,G_K:
! p(i)=B_k, p(i+1)=E_k, p(i+2)=G_k, k=1,...,numparam/3
! numparam should be ge 3
! 
  integer(kind=int32) :: nPts
  real(kind=double) :: xinput(nPts)
  real(kind=double) :: yinput(nPts)
  integer(kind=int32) :: numparam
  real(kind=double) :: param(numparam)
  real(kind=double) :: meany
  real(kind=double) :: siginput(nPts)
  real(kind=double) :: aLambda
  integer(kind=int32), dimension(numparam) :: fitList
  real(kind=double), dimension(numparam,numparam) :: covMatrix
  real(kind=double), dimension(numparam,numparam) :: alphaMatrix
  real(kind=double) :: chiSqr
  real(kind=double), dimension(numparam) :: atry, beta, da
  real(kind=double), dimension(numparam,numparam) :: oneda
  real(kind=double) :: ochiSqr, chi0, finchi0
  integer(kind=int32) :: istep, jstep, i1,i
  ! 08.01.2004
  real(kind=double)  :: xtest , ytest ! this must be uncomm. for opt. extr. 
  real(kind=double), dimension(numparam) ::dydptest ! this must be uncomm. for opt. extr. 
  real(kind=double) :: signature, amplitude, amplitude0, maxy
  integer(kind=int32) :: imaxy

  !write(*,*) " L1079: Starting GAUSSIANA2, numparam=", numparam
  !numparam=3
  atry=0.
  beta=0.
  da=0.
  oneda=0.
  ochiSqr=0.
  fitList=1

  ! averaging y
  meany=0.0
  maxy=0.0
  imaxy=nPts/2
  do i1=1,nPts
     if (yinput(i1) .gt. maxy) then
        maxy=yinput(i1)
        imaxy=i1
     endif
     meany=meany+yinput(i1)
  enddo
  if (nPts .gt. 0) then
     meany=meany/real(nPts)
  else
     write(*,*) "Error (no input points for Gaussian approximation"
  endif
  !write(*,*) " GAUSSIANA L1105: meany=", meany, " maxy=", maxy, " imaxy=", imaxy

  !Initial values for the input parameters
!  param(1)=0.0 ! 2.0
!  param(2)=meany ! 6.0
!  param(3)=nPts/2.0 ! 10.0
!  param(4)=param(3)/3.0 ! 5.0

  signature=1.0
  amplitude=1.0
  amplitude0=0.0
  do i=1,numparam-1,3
     signature=(-1.0)*signature
     if (i .gt. 3) then 
        amplitude=2.0
        amplitude0=1.0
     endif
     param(i)=maxy/(amplitude+amplitude0*5.0) 
     param(i+1)=real(imaxy)+amplitude0*signature*nPts/4.0 
!     param(i+1)=real(nPts)/2+amplitude0*signature*real(nPts)/4.0 
!     param(i+2)=nPts/2.0
     param(i+2)=2.5 ! nPts/amplitude
  enddo
  !write(*,*) "******************000000000000000000"
  !write(*,*) "             Initial parameter values:"
  do i=1, numparam-1,3
     !write(*,*) " L1103 i=", i, " param(i+1)=", param(i), &
     !  " param(i+2)=", param(i+1), " param(i+3)=", param(i+2)
  enddo
  siginput=16.0
  aLambda=-10
  
  call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
       covMatrix, alphaMatrix, chiSqr, fgaussup, aLambda, atry, beta, da, oneda, ochisqr)
!  VNY 14.01.2004
!  chi0=chiSqr
  chi0=chiSqr*1.2
  ochiSqr=chi0
  !write(*,*) "  First chiSqr1=", chiSqr, " old chiSqr=", ochiSqr, " aLambda1=", aLambda 
  !write(*,*) "  First   p1=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i1=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "*********************1111111111111111"

  !write(*,*) " Consecutive steps :"
  istep=0
  fchi: do 
     istep=istep+1
     call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
          covMatrix, alphaMatrix, chiSqr, fgaussup, aLambda, atry, beta, da, oneda, ochiSqr)
     !write(*,*) "1) istep=", istep, "    chi0=", chi0, " chiSqr=", chiSqr, " ochiSqr=", ochiSqr
     if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi
     chi0=chiSqr
  enddo fchi
  !write (*,*) " Intermediate chiSqr=", chiSqr, " istep=", istep
  !write (*,*) " Intermediate parameters=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i1=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************2222222222222222222"
  finchi0=chiSqr
  !write(*,*) " steps :"
  jstep=0
  fchi0 : do
     jstep=jstep+1
     istep=0
     aLambda=-10
     fchi1: do 
        istep=istep+1
        call Mrqmin(xinput, yinput, siginput, nPts, param, fitList, numparam, &
             covMatrix, alphaMatrix, chiSqr, fgaussup, aLambda, atry, beta, da, oneda, ochiSqr)
        if (abs(chiSqr-chi0) .lt. 1.E-6 .or. istep .gt. 80) exit fchi1
        chi0=chiSqr
     enddo fchi1
     if (abs(chiSqr-finchi0) .lt. 1.E-6 .or. jstep .gt. 10) exit fchi0
     finchi0=chiSqr
  enddo fchi0
  !write (*,*) " Final chiSqr=", chiSqr, " jstep=", jstep
  !write (*,*) " Final parameters2=", param
  do i1=1, nPts
     xtest=real(i1)
     call fgaussup(xtest,param,ytest,dydptest,numparam)
     !write(*,*) "     i=", i1, " x(i)=", xtest, " y(i)=",yinput(i1), " ytest=", ytest
  enddo
  !write(*,*) "**************3333333333333333333"
end subroutine gaussiana2


!---exponential integral function 27.02.03
real(double) function eiFunc(x)
! using normal function: seems to be valid for -2<x<0
   real(kind=double), intent(in)   :: x
   real(kind=double)    :: smallgamma,x0, x1, xd, ei, ei0, nfact
   real(kind=double) :: sum,sig
   integer    :: i
   smallgamma=0.5772156649015328606
   x0=x
   ei=0.0
   ei0=0.0
   if (x0 .lt. 0) then
      x1=abs(x0)
      if (x1 .lt. 4.885) then
         ei0=0.5772156649015328606+log(x1)
         nfact=1.0
         xd=1.0
         do i=1,20
            nfact=nfact*real(i)
            xd=x0**real(i)
            if (xd .gt. 1.0E10) exit
            ei=ei0+xd/(real(i)*nfact)
            if (abs(ei-ei0)  .lt. 1.0E-8) exit
            ei0=ei
         enddo
      endif
      if (x1 .ge. 4.885) then
         sum=1.0
         nfact=1.0
         sig=1.0
         xd=1.0
         do i=1,5
            xd=xd*x1
            nfact=nfact*real(i)
            sig=-1.0*sig
            sum=sum+sig*nfact/xd
         enddo
         ei0=exp(-x1)/x1
         ei0=ei0*sum
         ei0=-ei0
      endif
   endif
   eiFunc=ei0
end function eiFunc
!--
!---exponential integral function 27.02.03
real(double) function eiFunc2(x)
! using formula for x>>1
   real(kind=double), intent(in)   :: x
   real(kind=double)    :: smallgamma,x0, x1, xd, ei, ei0, nfact
   real(kind=double) :: sum,sig
   integer    :: i
   smallgamma=0.5772156649015328606
   ei=0.0
   ei0=0.0
   x0=x
   if (x0 .lt. 0) then
      x1=abs(x0)
!      if (x1 .lt. 5.0) then
!         ei0=0.5772156649015328606+log(x1)
!         nfact=1.0
!         xd=1.0
!         do i=1,20
!            nfact=nfact*real(i)
!            xd=x**real(i)
!            if (xd .gt. 1.0E10) exit
!            ei=ei0+xd/(real(i)*nfact)
!            if (abs(ei-ei0)  .lt. 1.0E-8) exit
!            ei0=ei
!         enddo
!!         write (*,*) " eiFunc: i=",i, " ei0=", ei0
!      endif
!      if (x1 .ge. 5.0) then
         sum=1.0
         nfact=1.0
         sig=1.0
         xd=1.0
         do i=1,5
            !xd=x1**real(i)
            xd=xd*x1
            nfact=nfact*real(i)
            sig=-1.0*sig
            sum=sum+sig*nfact/xd
!            write (*,*) " i=", i, " xd=", xd, " nfact=", nfact, &
!                 " t=", sig*nfact/xd, " sum=", sum
         enddo
         ei0=exp(-x1)/x1
         ei0=ei0*sum
!         write (*,*) "x1=", x1, " sum=", sum, " e1func=", ei0
         ei0=-ei0
!      endif
   endif
   eiFunc2=ei0
!   eiFunc2=sum
!   write (*,*) "x1=", x1, " sum=", sum, " e1func=", ei0, " f2=", eiFunc2
end function eiFunc2


subroutine polyfit(x,y,sigmay,npts,nterms,mode,a,chisqr)

   real(kind=double) x(:)
   real(kind=double) y(:)
   real(kind=double) sigmay(:)
   integer(kind=int32) npts
   integer(kind=int32) nterms
   integer(kind=int32) mode
   real(kind=double) a(:)
   real(kind=double) chisqr

   integer(kind=int32) nmax
   integer(kind=int32) n
   integer(kind=int32) j
   integer(kind=int32) i
   integer(kind=int32) k
   integer(kind=int32) l
   real(kind=double) chisq
   real(kind=double) weight
   real(kind=double) xterm
   real(kind=double) yterm
   real(kind=double) delta
   real(kind=double) free
   real(kind=double) xi
   real(kind=double) yi
   real(kind=double), allocatable :: array(:,:) 
   real(kind=double), allocatable :: sumx(:)
   real(kind=double), allocatable :: sumy(:)

   allocate(array(nterms, nterms))
   allocate(sumx(npts))
   allocate(sumy(npts))

   nmax  =  2 * nterms - 1
   sumx  =  0.0d0
   sumy  =  0.0d0
   chisq  =  0.0d0
   do i  =  1, npts
      xi  =  x(i)
      yi  =  y(i)
      if (mode < 0) then
         if (yi < 0) then
            weight  =  -1.0d0 / yi
         else if (yi == 0) then
            weight = 1.d0
         else
            weight = 1.0d0 / yi
         end if
      else if (mode == 0) then
         weight = 1.d0
      else
         weight = 1.0d0 / sigmay(i)**2
      end if

      xterm = weight
      do n = 1, nmax
         sumx(n) = sumx(n) + xterm
         xterm = xterm * xi
      end do

      yterm = weight * yi
      do n  =  1, nterms
         sumy(n) = sumy(n) + yterm
         yterm = yterm * xi
      end do

      chisq = chisq + weight * yi**2
   end do

   do j = 1, nterms
      do k = 1, nterms
         n = j + k - 1
         array(j, k) = sumx(n)
      end do
   end do

   delta = determ(array, nterms)
   if (delta /= 0) then
      do l = 1, nterms
         do j = 1, nterms
            do k = 1, nterms
               n  = j + k - 1
               array(j, k) = sumx(n)
            end do

            array(j, l) = sumy(j)
         end do

         a(l) = determ(array, nterms) / delta
      end do

      do j = 1, nterms
         chisq = chisq - 2.0d0 * a(j) * sumy(j)
         do k = 1, nterms
            n = j + k - 1
            chisq = chisq + a(j) * a(k) * sumx(n)
         end do
      end do

      free = dble(npts - nterms)
      chisqr = chisq / free
   else
      chisqr = 0.0d0
      do j = 1, nterms
         a(j) = 0.0d0
      end do
   end if

   deallocate(array)
   deallocate(sumx)
   deallocate(sumy)

end subroutine polyfit

function determ(array, norder) result(res)

   integer(kind=int32) :: norder
   real(kind=double) array(norder, norder)

   real(kind=double) :: res

   integer(kind=int32) :: i
   integer(kind=int32) :: j
   integer(kind=int32) :: k
   integer(kind=int32) :: k1
   real(kind=double) :: save

   res = 1.d0

   do k = 1, norder
      if (array(k, k) /= 0) goto 100

      do j = k, norder
         if (array(k,j) /= 0) goto 50
      end do

      res = 0.00
      goto 200

50    continue

      do i = k, norder
         save = array(i, j)
         array(i, j) = array(i, k)
         array(i, k) = save
      end do

      res = -res

100   continue

      res = res * array(k, k)

      if (k < norder) then
         k1 = k + 1

         do i = k1, norder
            do j = k1, norder
               array(i, j) = array(i, j) - &
                             array(i, k) * &
			     array(k, j) / &
			     array(k, k)
            end do
         end do
      end if
   end do

200 continue

end function determ
  
subroutine locate(xx, n, x, j)

   real(kind=double) :: xx(:)
   integer(kind=int32) n
   real(kind=double) :: x
   integer(kind=int32) j

   integer(kind=int32) ju
   integer(kind=int32) jl
   integer(kind=int32) jm
   
   j = 0

   if (x < xx(1)) then
      j = 1
   endif

   if (x > xx(n)) then
      j = n
   endif


   if ( j == 0) then

      if (x .eq. xx(n)) then
         ju = n-1
         jl = n-1
      endif

      if (x .eq. xx(1)) then
         jl = 1
         ju = 1 
      endif

      jl = 0
      ju = n + 1

      do while(ju - jl > 1)
         jm = (ju + jl) / 2
 
         if ((xx(n) > xx(1)) .eqv. (x > xx(jm))) then
            jl = jm
         else
            ju = jm
         end if
      end do

      j = jl
   endif

end subroutine locate



!*************************************************************
! subroutine dlocate
! VNY 25.01.2002; If the dimension of the array xx is 1 or 0
! this subroutin can crash due to subscript. Protection 
! against this crash has been added
! Modified 12.4.2002 to use arrays starting at subscript 0
!**************************************************************
subroutine dlocate(xx, n, x, j)
   integer (kind=int32), intent(in) :: n
   real(kind=double)   :: xx(0 : n - 1)
   
   real(kind=double)   :: x
   integer(kind=int32) :: j

   integer(kind=int32) :: ju
   integer(kind=int32) :: jl
   integer(kind=int32) :: jm

   jl = 0
   ju = n + 1

! VNY 25.01.2002 Protection against one-element array
   if (n .le. 1) then
      j = 0
      return
   endif

   if (x .ge. xx(n - 1)) then
       j = n - 1
       return
   end if 
      

   if (x .le. xx(0)) then
      j = 0
      return
   endif

   do while(ju - jl > 1)
      jm = (ju + jl) / 2

      if (x >= xx(jm - 1)) then
         jl = jm
      else
         ju = jm
      end if
   end do

   j = jl - 1

end subroutine dlocate

subroutine Mrqmin(x, y, sig, nData, a, ia, ma, covMatrix, &
                    alphaMatrix, chiSqr, funcs, aLambda, atry, &
                    beta, da, oneda, ochiSqr)
! Leveenberg-Marquardt method, attempting to reduce the value 
! of chi**2 of a fit between a set of data points x[1..nData],
! y[1..nData] with individual standard deviations sig[1..nData],
! and a nonlinear function dependent on ma coefficients a[1..ma].
! The input array ia[1..ma] indicates by nonzero entries those components
! of a that should be fitted for, and by zero entries those compoents
! that should be held fixed at their input values. The program 
! returns current best-fit values for the parameters a[1..ma],
! and chi**2=chiSqr. The arrays covar[1..ma][1..ma], alpha[1..ma][1..ma]
! are used as working space during most iterations. Supply a routine
! funcs(x,a,yfit,dyda,ma) that evaluates the fitting function yfit,
! and its derivatives dyda[1..ma] with respect to the fitting parameters
! a at x. On the first call provide an initial guess for the parameters
! a, and set aLambda<0 for initialization (which then sets aLambda=.001).
! If a step succeeds chiSqr becomes smaller and aLambda decreases 
! by a factor of 10. If a step fails aLambda grows by a factor of 10.
! You must call this routine repeatedly until convergence is achieved.
! Then, make one final call with aLambda=0, so that covar[1..ma][1.ma]
! returns the covariance matrix, and alpha the curvature matrix.
! (Parameters held fixed will return zero covariances.)
!  
   integer(kind=int32) :: nData
   real(kind=double)   :: x(nData)
   real(kind=double)   :: y(nData)
   real(kind=double)   :: sig(nData)
   integer(kind=int32) :: ma
   real(kind=double)   :: a(ma)
   integer(kind=int32) :: ia(ma)
!   integer(kind=int32) :: fitList(*)
   integer(kind=int32) :: mFit
!   integer(kind=int32) :: nca
   real(kind=double)   :: covMatrix(ma, ma)
   real(kind=double) :: chiSqr
   real(kind=double) :: aLambda
   real(kind=double) :: alphaMatrix(ma, ma)
   integer(kind=int32), parameter :: mmax = 20
!   real                :: atry(mmax)
!   real                :: beta(mmax)
!   real                :: da(mmax)
!   real                :: oneda(mmax,mmax)
   real(kind=double) :: atry(ma)
   real(kind=double) :: beta(ma)
   real(kind=double) :: da(ma)
   real(kind=double) :: oneda(ma,ma)
!   integer(kind=int32) :: ihit
   integer(kind=int32) :: j
   integer(kind=int32) :: k
   integer(kind=int32) :: l
!   integer(kind=int32) :: kk
   real(kind=double) :: ochiSqr

   external funcs

   !write(*,*) " OMGNUMERICAL L1873 Marquardt input aLambda=", aLambda
   
   ! Initialisation
   if (aLambda < 0.0) then
      mFit=0
      do j=1,ma
         if (ia(j) .ne. 0) mFit=mFit+1
      enddo

      aLambda = 0.001
      !write (*,*) " Mrqmin: L1882, aLambda=", aLambda,  " a=", a, " beta=", beta, " mFit=", mFit
      call MrqCof(x, y, sig, nData, a, ia, ma, alphaMatrix, beta, &
                  chiSqr, funcs)
      !write(*,*) " Mrqmin: L1885, a=", a, " chiSqr=", chiSqr, " beta=", beta
      ochiSqr = chiSqr
      atry(1:ma) = a(1:ma)
   end if

   ! Alter linearised fitting matrix, by augmenting
   ! diagonal elements
   mFit=0
   do j=1,ma
      if (ia(j) .ne. 0) mFit=mFit+1
   enddo
   !write (*,*) " Mrqmin: L1896, mFit=", mFit, " at the entrance atry=", atry 
   !write(*,*) " alphaMatrix="
   
   !do j=1, mFit
   !   write (*,*) alphaMatrix(j,1:mFit)
   !enddo
   !write(*,*) "---"
   !write(*,*) "L1903  beta=", beta
   !write(*,*) "---"
   do j = 1, mFit
      do k = 1, mFit
         CovMatrix(j, k) = alphaMatrix(j, k)
      end do
      CovMatrix(j, j) = alphaMatrix(j, j) * (1.0 + aLambda)
      oneda(j,1) = beta(j)
   end do
   
   !Test VNY 30.01.03
   !write (*,*) " Mqtmin: L1914, da=", da
   ! Matrix solution
   call gaussj(covMatrix, mFit, da, 1)
   !write (*,*) " Mqtmin: L1917, da=", da

   da(1:mFit) = oneda(1:mFit,1)

   if (aLambda == 0.0) then
      !write (*,*) " Mqtmin: L1922, before call CovarianceSort-1"
      ! Once converged, evaluate covariance matrix
      call CovarianceSort(covMatrix, ma, ia, mFit)
      !write (*,*) " Mqtmin: Line 1925, before call CovarianceSort-2"
      ! Spread out alpha to its full size too
      call CovarianceSort(alphaMatrix, ma, ia, mFit)
      !write (*,*) " Mqtmin: Line 1928, after call Covariance Sort"
      return
   end if
   
   ! Did the trial succeed?
   !write(*,*) "((((((((((( before the modification: atry=a+da )))))))))))"
   !write(*,*) " L1934: atry=", atry
   !write(*,*) " L1935: a=", a
   !write(*,*) " L1936: da=", da
   j=0
   do l = 1, ma
      if (ia(l) .ne. 0) then
         j=j+1
         ! VNY 13.01.2004
         atry(l) = a(l) + da(j)
         !atry(l) = atry(l) + da(j)
      endif
   end do

   !write (*,*) " L1947: MODIFIED atry=", atry
   !write (*,*) "         da=", da
   call MrqCof(x, y, sig, nData, atry, ia, ma, covMatrix, da, &
               chiSqr, funcs)
   !write (*,*) " L1951: atry=", atry
   !write (*,*) " chiSqr=", chiSqr, " old chiSqr=", ochiSqr
   if (chiSqr < ochiSqr) then
      ! Success, accept the new solution
      !write(*,*) " The trial has succeed: decreasing aLambda"
      aLambda = 0.1 * aLambda
      ochiSqr = chiSqr

      do j = 1, mFit
         do k = 1, mFit
            alphaMatrix(j, k) = CovMatrix(j,k)
         end do
         beta(j) = da(j)
      end do
      !write (*,*) " Mqtmin: Line 1965, after do j=1,mFit"
      a(1:ma)=atry(1:ma)
   else
      ! Failure, increase aLambda and return
      !write(*,*) " The trial is not successful: chiSqr=", chiSqr, &
      !     " ochiSqr=", ochiSqr
      aLambda = 10.0 * aLambda
      chiSqr = ochiSqr
   end if
   !write (*,*) " Mqtmin: Line 1974, before return from Mrqmin"
end subroutine Mrqmin

subroutine MrqCof(x, y, sig, nData, a, ia, ma, alphaMatrix, beta, &
                 chiSqr, funcs)
! Used by Mrqmin to evaluate the linearized fitting matrix alphaMatrix,
! and vector beta, and calculate chi**2
   integer(kind=int32) nData
   real(kind=double) x(nData)
   real(kind=double) y(nData)
   real(kind=double) sig(nData)
   integer(kind=int32) ma
   real(kind=double) a(ma)
   integer(kind=int32) ia(*)
!   integer(kind=int32) fitList(*)
   integer(kind=int32) mFit
!   integer(kind=int32) nalp
   real(kind=double) alphaMatrix(ma, ma)
   real(kind=double) beta(ma)
   real(kind=double) chiSqr
   real(kind=double) sig2i,dy,ymod,wt

   integer(kind=int32) i
   integer(kind=int32) j
   integer(kind=int32) k
   integer(kind=int32) l
   integer(kind=int32) m

   external funcs

   integer(kind=int32), parameter :: mmax = 20
!   real dyda(mmax)
   real(kind=double) dyda(ma)

   mFit=0
   do j=1,ma
      if (ia(j) .ne. 0) mFit=mFit+1
   enddo
!   write(*,*) " MRQCOF L2012 ---"
!   write(*,*) "L2013: mFit=", mFit
   do j = 1, mFit
      ! Initialise (symmetric) alpha, beta
      do k = 1, j
         alphaMatrix(j, k) = 0.0
      end do
      beta(j) = 0.0
   end do
!---VNY 03.02.03, commented 13.01.2004
!   alphaMatrix=0.0
!   beta=0.0
!---
!   write(*,*) "L493  alphaMatrix="
!   do j=1, mFit
!      write (*,*) alphaMatrix(j,1:mFit)
!   enddo
!   write(*,*) "L497: beta=", beta

   chiSqr = 0.0
!   write(*,*) "....................................................................."
!   write(*,*) " MRQCOF L2033: Computing the function values for the given parameters" 
   do i = 1, nData
      ! summation loop over all data
      
      !write(*,*) " before ymod=", ymod, " dyda=", dyda
      call funcs(x(i), a, ymod, dyda, ma)
      !write(*,*) " i=", i, " x(i)=", x(i), " y(i)=", ymod, " sig(i)=", sig(i)

      sig2i = 1.0 / (sig(i) * sig(i))
      dy = y(i)- ymod
      ! VNY 13.01.2004
      !sig2i=sig2i/abs(dy)
!      write(*,*) " i=", i, "x(i)=", x(i), " y(i)=", y(i), " ymod=", ymod, " dy=", dy
!      write(*,*) " i=", i, " dyda=", dyda
      j=0
      do l = 1, ma
         if (ia(l) .ne. 0) then
            j=j+1
! VNY 28.01.2004
!            wt = dyda(l) * sig2i
            wt = 0.01*dyda(l) * sig2i
!
!            write(*,*) " l=",l, " dyda=", dyda(l), " sig2i=", sig2i
            k=0
            do m = 1, l
               if (ia(m) .ne. 0) then
                  k=k+1
                  alphaMatrix(j, k) = alphaMatrix(j, k) + wt * dyda(m)
               endif
            enddo
            beta(j)=beta(j)+dy*wt
!
!            write (*,*) "L1613 j=", j, " dy=", dy, " wt=", wt, & 
!                 " beta(j)=", beta(j)
         endif
      end do

      chiSqr = chiSqr + dy * dy * sig2i ! and find chi**2.
   end do ! i=1, nData
!   write(*,*) "....................................................................."

   do j = 2, mFit ! Fill in the synnetric side
      do k = 1, j - 1
         alphaMatrix(k, j) = alphaMatrix(j, k)
      end do
   end do
   return
end subroutine MrqCof

subroutine GaussJ(a, n, b, m)
! Linear equation solution by Gaus-Jordan elimination.
! a[1..n][1..n] is the input matrix. b[1..n][1..m] is
! input containing the m right-hand side vectors. On output,
! a is replaced by its matrix inverse, and b is replaced by the 
! corresponding set of solution vectors
   integer(kind=int32) n
   real(kind=double) :: a(n, n)
   integer(kind=int32) m
   real(kind=double) ::  b(n, m)
   integer i,j,k
   integer, parameter :: nmax=50
   real(kind=double):: big,pivinv,dum !, temp
   integer irow, icol, l, ll
   integer ipiv(nmax),indxr(nmax),indxc(nmax)

   ! The integer arrays ipiv, indxr, and indxc are used 
   ! for bookkeeping on the pivoting.

   ipiv(1:n) = 0

   do i = 1, n
      !This is the main loop over the columns to be reduced
      big = 0.0

      do j = 1, n
         if (ipiv(j) /= 1) then
            do k = 1, n
               if (ipiv(k) == 0) then
                  if (abs(a(j, k)) >= big) then
                     big = abs(a(j, k))
                     irow = j
                     icol = k
                  end if
               else if (ipiv(k) > 1) then
                  call fatal("sigMat", "Singular matrix in Marquardt minimization")
               end if
            end do
         end if
      end do

      ipiv(icol) = ipiv(icol) + 1
      ! We now have the pivot element, so we interchange rows,
      ! if needed, to put the pivot element on the diagonal.
      ! The column are not physically interchanged, only relabeled:
      ! indxc[i], the column of the ith pivot element, is the ith column
      ! that is reduced, while indxr[i] is the row i which that pivot
      ! element was originallylocated. If indxr[i] .ne. indxc[i] there
      ! is an implied column interchange. With this form of bookkeeping,
      ! the solution b's will end up in the correct order, and the inverse
      ! matrix will be scrambled by columns.
      if (irow /= icol) then
         do l = 1, n
            dum = a(irow, l)
            a(irow, l) = a(icol, l)
            a(icol, l) = dum
         end do

         do l = 1, m
            dum = b(irow, l)
            b(irow, l) = b(icol, l)
            b(icol, l) = dum
         end do
      end if
      
      ! we are now ready to divide the pivot row by the 
      ! pivot element, located at irow and icol.
      indxr(i) = irow
      indxc(i) = icol

      if (a(icol, icol) == 0.0) then
         call fatal("sigMat", "Singular matrix in Marquardt minimization")
      end if

      pivinv = 1.0 / a(icol, icol)
      a(icol, icol) = 1.0

      do l = 1, n
         a(icol, l) = a(icol, l) * pivinv
      end do

      do l = 1, m
         b(icol, l) = b(icol, l) * pivinv
      end do

      do ll = 1, n
         if (ll /= icol) then
            ! Next, we reduce the rows except for the pivot one
            dum = a(ll, icol)
            a(ll, icol) = 0.0

            do l = 1, n
               a(ll, l) = a(ll, l) - a(icol, l) * dum
            end do

            do l = 1, m
               b(ll, l) = b(ll, l) - b(icol, l) * dum
            end do
         end if
      end do
      !This this the end of the main loop over columns of the 
      ! reduction. It only remains to unscramble the solution
      ! in view of the column interchanges. We do this by 
      ! interchanging pairs of columns in the reverse order
      ! that the permutation was build up.
   end do

   do  l = n, 1, -1
      if (indxr(l) /= indxc(l)) then
         do k = 1, n
            dum = a(k, indxr(l))
            a(k, indxr(l)) = a(k, indxc(l))
            a(k, indxc(l)) = dum
         end do
      end if
   end do

end subroutine GaussJ

subroutine CovarianceSort(covMatrix, ma, ia, mFit)
! expand in storage the covariance matrix covMatrix, so as to take
! into account parameters that are being held fixed. (For the latter,
! return zero covariances)
  integer :: mFit, i, j, k, ma
  real(kind=double) :: CovMatrix(ma,ma)
  integer :: ia(*)
  real(kind=double) ::  dum
  do i=mFit+1, ma
     do j=1,i
        covMatrix(i,j)=0.
        covMatrix(j,i)=0
     enddo
  enddo
  k=mFit
  do j=ma,1,-1
     if (ia(j) .gt. 0)  then
        do i=1,ma
           dum=covMatrix(i,k)
           covMatrix(i,k)=covMatrix(i,j)
           covMatrix(i,j)=dum
        enddo
        do i=1,ma
           dum=covMatrix(k,i)
           covMatrix(k,i)=covMatrix(j,i)
           covMatrix(j,i)=dum
        enddo
        k=k-1
     endif
  enddo
end subroutine CovarianceSort

real function computeMedian(image, nx, ny, i1, i2, j1, j2) result (median)

   ! computes the median of an image over the specific limits

   integer :: nx, ny
   real(kind=double) :: image(0:nx-1, 0:ny-1)
   integer :: i1, i2, j1, j2
   integer :: i, j
   real(kind=single), allocatable :: array(:)
   integer :: npts
   
   npts = (i2-i1+1)*(j2-j1+1)
   allocate(array(npts))

   npts = 0
   do i = i1, i2
      do j = j1, j2
         npts = npts + 1
         array(npts) = image(i,j)
      enddo
   enddo

   call sort(npts, array)

   if (NINT(real(npts)/2.) /= INT(npts/2)) then
      median = array(NINT(real(npts)/2.))
       else
      median = 0.5*(array(npts/2) + array(npts/2+1))
   endif

end function computeMedian


real function arrayMedian(array, npts) result (median)

   ! computes the median of an array

   integer :: npts
   real(kind=single) :: array(*)
   
!   integer i

   if (npts .eq. 1) then
      median = array(1)
    else
      call sort(npts, array)
      if (NINT(real(npts)/2.) /= INT(npts/2)) then
        median = array(NINT(real(npts)/2.))
         else
        median = 0.5*(array(npts/2) + array(npts/2+1))
      endif
   endif

!   write (*, *) "npts : ", npts
!   do i = 1, npts
!      write (*, *) i, array(i)
!   end do

end function arrayMedian

!------ Subroutines related to fitpack
subroutine curv1 (n,x,y,slp1,slpn,islpsw,yp,temp,sigma,ierr)
!
      integer(kind=int32) :: n,islpsw,ierr
      real(kind=double) :: x(n),y(n),slp1,slpn,yp(n),temp(n),sigma
      integer(kind=double) :: nm1, np1
      real(kind=double) :: sigmap, delx1, delx2, slpp1, c1, c2, c3
      real(kind=double) :: delxn, delxnm, slppn, dx1, diag1, sdiag1
      real(kind=double) :: diag, dx2, diag2, sdiag2
      integer(kind=int32) :: i, ibak
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute an interpolatory spline under tension through
! a sequence of functional values. the slopes at the two
! ends of the curve may be specified or omitted.  for actual
! computation of points on the curve it is necessary to call
! the function curv2.
!
! on input--
!
!   n is the number of values to be interpolated (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   functional values.
!
!   y is an array of the n ordinates of the values, (i. e.
!   y(k) is the functional value corresponding to x(k) ).
!
!   slp1 and slpn contain the desired values for the first
!   derivative of the curve at x(1) and x(n), respectively.
!   the user may omit values for either or both of these
!   parameters and signal this with islpsw.

!   islpsw contains a switch indicating which slope data
!   should be used and which should be estimated by this
!   subroutine,
!          = 0 if slp1 and slpn are to be used,
!          = 1 if slp1 is to be used but not slpn,
!          = 2 if slpn is to be used but not slp1,
!          = 3 if both slp1 and slpn are to be estimated
!              internally.
!
!   yp is an array of length at least n.
!
!   temp is an array of length at least n which is used for
!   scratch storage.
!
! and
!
!   sigma contains the tension factor. this value indicates
!   the curviness desired. if abs(sigma) is nearly zero
!   (e.g. .001) the resulting curve is approximately a
!   cubic spline. if abs(sigma) is large (e.g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results.  a standard value
!   for sigma is approximately 1. in absolute value.
!
! on output--
!
!   yp contains the values of the second derivative of the
!   curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if x-values are not strictly increasing.
!
! and
!
!   n, x, y, slp1, slpn, islpsw and sigma are unaltered.

! this subroutine references package modules ceez, terms,
! and snhcsh.
!
!-----------------------------------------------------------
!
      nm1 = n-1
      np1 = n+1
      ierr = 0
      if (n .le. 1) then
         !
         ! too few points
         !
         ierr = 1
         return
      endif

      if (x(n) .le. x(1)) then
         !
         ! x-values not strictly increasing
         !
         ierr = 2
         return
      endif

      !
      ! denormalize tension factor
      !
      sigmap = abs(sigma)*real(n-1)/(x(n)-x(1))
      !
      ! approximate end slopes
      !
      if (islpsw .ge. 2) then
         delx1 = x(2)-x(1)
         delx2 = delx1+delx1
         if (n .gt. 2) delx2 = x(3)-x(1)
         if (delx1 .le. 0. .or. delx2 .le. delx1) then ! go to 9
            ierr = 2
            return
         endif
         call ceez (delx1,delx2,sigmap,c1,c2,c3,n)
         slpp1 = c1*y(1)+c2*y(2)
         if (n .gt. 2) slpp1 = slpp1+c3*y(3)
      else
         slpp1 = slp1
      endif
      if (islpsw .eq. 1 .or. islpsw .eq. 3) then
         delxn = x(n)-x(nm1)
         delxnm = delxn+delxn
         if (n .gt. 2) delxnm = x(n)-x(n-2)
         if (delxn .le. 0. .or. delxnm .le. delxn) then !go to 9
            ierr = 2
            return
         endif
         call ceez (-delxn,-delxnm,sigmap,c1,c2,c3,n)
         slppn = c1*y(n)+c2*y(nm1)
         if (n .gt. 2) slppn = slppn+c3*y(n-2)
      else
         slppn = slpn
      endif

      !
      ! set up right hand side and tridiagonal system for yp and
      ! perform forward elimination
      !
      delx1 = x(2)-x(1)
      if (delx1 .le. 0.) then
         ierr = 2
         return
      endif
      dx1 = (y(2)-y(1))/delx1
      call terms (diag1,sdiag1,sigmap,delx1)
      yp(1) = (dx1-slpp1)/diag1
      temp(1) = sdiag1/diag1
      if (n .eq. 2) then
         diag = diag1-sdiag1*temp(nm1)
         yp(n) = (slppn-dx1-sdiag1*yp(nm1))/diag
      else
         do i = 2,nm1
            delx2 = x(i+1)-x(i)
            if (delx2 .le. 0.) then
               ierr = 2
               return
            endif
            dx2 = (y(i+1)-y(i))/delx2
            call terms (diag2,sdiag2,sigmap,delx2)
            diag = diag1+diag2-sdiag1*temp(i-1)
            yp(i) = (dx2-dx1-sdiag1*yp(i-1))/diag
            temp(i) = sdiag2/diag
            dx1 = dx2
            diag1 = diag2
            sdiag1 = sdiag2
         enddo
         diag = diag1-sdiag1*temp(nm1)
         yp(n) = (slppn-dx1-sdiag1*yp(nm1))/diag
      endif
      !
      ! perform back substitution
      !
      do  i = 2,n
         ibak = np1-i
         yp(ibak) = yp(ibak)-temp(ibak)*yp(ibak+1)
      enddo
      return
end subroutine curv1

!****************************************************
!  subroutine CEEZ
!****************************************************
      subroutine ceez (del1,del2,sigma,c1,c2,c3,n)
!
      real(kind=double) :: del1,del2,sigma,c1,c2,c3
      integer(kind=int32) :: n
      real(kind=double) :: del,delp, delm, denom, coshm1, sinhmp
      real(kind=double) :: sinhmm, coshm2, dummy
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the coefficients c1, c2, and c3
! used to determine endpoint slopes. specifically, if
! function values y1, y2, and y3 are given at points x1, x2,
! and x3, respectively, the quantity c1*y1 + c2*y2 + c3*y3
! is the value of the derivative at x1 of a spline under
! tension (with tension factor sigma) passing through the
! three points and having third derivative equal to zero at
! x1. optionally, only two values, c1 and c2 are determined.
!
! on input--
!
!   del1 is x2-x1 (.gt. 0.).
!
!   del2 is x3-x1 (.gt. 0.). if n .eq. 2, this parameter is
!   ignored.
!
!   sigma is the tension factor.
!
! and
!
!   n is a switch indicating the number of coefficients to
!   be returned. if n .eq. 2 only two coefficients are
!   returned. otherwise all three are returned.
!
! on output--
!
!   c1, c2, and c3 contain the coefficients.
!
! none of the input parameters are altered.
!
! this subroutine references package module snhcsh.
!
!-----------------------------------------------------------
!
      if (n .eq. 2) go to 2
      if (sigma .ne. 0.) go to 1
      del = del2-del1
!
! tension .eq. 0.
!
      c1 = -(del1+del2)/(del1*del2)
      c2 = del2/(del1*del)
      c3 = -del1/(del2*del)
      return
!
! tension .ne. 0.
!
    1 call snhcsh (dummy,coshm1,sigma*del1,1)
      call snhcsh (dummy,coshm2,sigma*del2,1)
      delp = sigma*(del2+del1)/2.
      delm = sigma*(del2-del1)/2.
      call snhcsh (sinhmp,dummy,delp,-1)
      call snhcsh (sinhmm,dummy,delm,-1)
      denom = coshm1*(del2-del1)-2.*del1*delp*delm* &
           (1.+sinhmp)*(1.+sinhmm)
      c1 = 2.*delp*delm*(1.+sinhmp)*(1.+sinhmm)/denom
      c2 = -coshm2/denom
      c3 = coshm1/denom
      return
!
! two coefficients
!
    2 c1 = -1./del1
      c2 = -c1
      return
    end subroutine ceez
!*************************************************
! subroutine TERMS
!*************************************************
      subroutine terms (diag,sdiag,sigma,del)
!
      real(kind=double) :: diag,sdiag,sigma,del
      real(kind=double) :: sigdel, denom, sinhm, coshm
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine computes the diagonal and superdiagonal
! terms of the tridiagonal linear system associated with
! spline under tension interpolation.
!
! on input--
!
!   sigma contains the tension factor.
!
! and
!
!   del contains the step size.
!
! on output--
!
!                sigma*del*cosh(sigma*del) - sinh(sigma*del)
!   diag = del*--------------------------------------------.
!                     (sigma*del)**2 * sinh(sigma*del)
!
!                   sinh(sigma*del) - sigma*del
!   sdiag = del*----------------------------------.
!                (sigma*del)**2 * sinh(sigma*del)
!
! and
!
!   sigma and del are unaltered.
!
! this subroutine references package module snhcsh.
!
!-----------------------------------------------------------
!
      if (sigma .ne. 0.) go to 1
      diag = del/3.
      sdiag = del/6.
      return
    1 sigdel = sigma*del
      call snhcsh (sinhm,coshm,sigdel,0)
      denom = sigma*sigdel*(1.+sinhm)
      diag = (coshm-sinhm)/denom
      sdiag = sinhm/denom
      return
    end subroutine terms
!--------------------------

!*************************************************
! subroutine SNHCSH
!*************************************************
      subroutine snhcsh (sinhm,coshm,x,isw)
!
      integer(kind=int32) :: isw
      real(kind=double) :: sinhm,coshm,x
      real(kind=double) :: ax, xs, sp14, sp13, sp12, sp11, sq12, sq11
      real(kind=double) :: sq10, sp25, sp24, sp23, sp22, sp21, sq22
      real(kind=double) :: sq21, sq20, sp35, sp34, sp33, sp32, sp31
      real(kind=double) :: sq32, sq31, sq30, sp45, sp44, sp43, sp42
      real(kind=double) :: sp41, sq42, sq41, sq40, expx, cp5, cp4
      real(kind=double) :: cp3, cp2, cp1, cq2, cq1, cq0, zp4, zp3
      real(kind=double) :: zp2, zp1, zq2, zq1, zq0
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine returns approximations to
!       sinhm(x) = sinh(x)/x-1
!       coshm(x) = cosh(x)-1
! and
!       coshmm(x) = (cosh(x)-1-x*x/2)/(x*x)
! with relative error less than 4.0e-14.
!
! on input--
!
!   x contains the value of the independent variable.
!
!   isw indicates the function desired
!           = -1 if only sinhm is desired,
!           =  0 if both sinhm and coshm are desired,
!           =  1 if only coshm is desired,
!           =  2 if only coshmm is desired,
!           =  3 if both sinhm and coshmm are desired.
!
! on output--
!
!   sinhm contains the value of sinhm(x) if isw .le. 0 or
!   isw .eq. 3 (sinhm is unaltered if isw .eq.1 or isw .eq.
!   2).
!
!   coshm contains the value of coshm(x) if isw .eq. 0 or
!   isw .eq. 1 and contains the value of coshmm(x) if isw
!   .ge. 2 (coshm is unaltered if isw .eq. -1).
!
! and
!
!   x and isw are unaltered.
!
!-----------------------------------------------------------
!
      data sp14/.227581660976348e-7/, &
           sp13/.612189863171694e-5/, &
           sp12/.715314759211209e-3/, &
           sp11/.398088289992973e-1/, &
           sq12/.206382701413725e-3/, &
           sq11/-.611470260009508e-1/, &
           sq10/.599999999999986e+1/ 
      data sp25/.129094158037272e-9/, &
           sp24/.473731823101666e-7/, &
           sp23/.849213455598455e-5/, &
           sp22/.833264803327242e-3/, &
           sp21/.425024142813226e-1/, &
           sq22/.106008515744821e-3/, &
           sq21/-.449855169512505e-1/, &
           sq20/.600000000268619e+1/ 
      data sp35/.155193945864942e-9/, &
           sp34/.511529451668737e-7/, &
           sp33/.884775635776784e-5/, &
           sp32/.850447617691392e-3/, &
           sp31/.428888148791777e-1/, &
           sq32/.933128831061610e-4/, &
           sq31/-.426677570538507e-1/, &
           sq30/.600000145086489e+1/
      data sp45/.188070632058331e-9/, &
           sp44/.545792817714192e-7/, &
           sp43/.920119535795222e-5/, &
           sp42/.866559391672985e-3/, &
           sp41/.432535234960858e-1/, &
           sq42/.824891748820670e-4/, &
           sq41/-.404938841672262e-1/, &
           sq40/.600005006283834e+1/
      data cp5/.552200614584744e-9/, &
           cp4/.181666923620944e-6/, &
           cp3/.270540125846525e-4/, &
           cp2/.206270719503934e-2/, &
           cp1/.744437205569040e-1/, &
           cq2/.514609638642689e-4/, &
           cq1/-.177792255528382e-1/, &
           cq0/.200000000000000e+1/
      data zp4/.664418805876835e-8/, &
           zp3/.218274535686385e-5/, &
           zp2/.324851059327161e-3/, &
           zp1/.244515150174258e-1/, &
           zq2/.616165782306621e-3/, &
           zq1/-.213163639579425e0/, &
           zq0/.240000000000000e+2/
!
      ax = abs(x)
      if (isw .ge. 0) go to 5
!
! sinhm approximation
!
      if (ax .gt. 3.9) go to 2
      xs = ax*ax
      if (ax .gt. 2.2) go to 1
!
! sinhm approximation on (0.,2.2)
!
      sinhm = xs*((((sp14*xs+sp13)*xs+sp12)*xs+sp11)*xs+1.)/ &
                   ((sq12*xs+sq11)*xs+sq10)
      return
!
! sinhm approximation on (2.2,3.9)
!
    1 sinhm = xs*(((((sp25*xs+sp24)*xs+sp23)*xs+sp22)*xs+sp21) &
              *xs+1.)/((sq22*xs+sq21)*xs+sq20)
      return
    2 if (ax .gt. 5.1) go to 3
!
! sinhm approximation on (3.9,5.1)
!
      xs = ax*ax
      sinhm = xs*(((((sp35*xs+sp34)*xs+sp33)*xs+sp32)*xs+sp31) &
              *xs+1.)/((sq32*xs+sq31)*xs+sq30)
      return
    3 if (ax .gt. 6.1) go to 4
!
! sinhm approximation on (5.1,6.1)
!
      xs = ax*ax
      sinhm = xs*(((((sp45*xs+sp44)*xs+sp43)*xs+sp42)*xs+sp41) &
              *xs+1.)/((sq42*xs+sq41)*xs+sq40)
      return
!
! sinhm approximation above 6.1
!
    4 expx = exp(ax)
      sinhm = (expx-1./expx)/(ax+ax)-1.
      return
!
! coshm and (possibly) sinhm approximation
!
    5 if (isw .ge. 2) go to 7
      if (ax .gt. 2.2) go to 6
      xs = ax*ax
      coshm = xs*(((((cp5*xs+cp4)*xs+cp3)*xs+cp2)*xs+cp1) &
              *xs+1.)/((cq2*xs+cq1)*xs+cq0)
      if (isw .eq. 0) sinhm = xs*((((sp14*xs+sp13)*xs+sp12) &
                *xs+sp11)*xs+1.)/((sq12*xs+sq11)*xs+sq10)
      return
    6 expx = exp(ax)
      coshm = (expx+1./expx)/2.-1.
      if (isw .eq. 0) sinhm = (expx-1./expx)/(ax+ax)-1.
      return
!
! coshmm and (possibly) sinhm approximation
!
    7 xs = ax*ax
      if (ax .gt. 2.2) go to 8
      coshm = xs*((((zp4*xs+zp3)*xs+zp2)*xs+zp1)*xs+1.)/ &
                   ((zq2*xs+zq1)*xs+zq0)
      if (isw .eq. 3) sinhm = xs*((((sp14*xs+sp13)*xs+sp12) &
                *xs+sp11)*xs+1.)/((sq12*xs+sq11)*xs+sq10)
      return
    8 expx = exp(ax)
      coshm = ((expx+1./expx-xs)/2.-1.)/xs
      if (isw .eq. 3) sinhm = (expx-1./expx)/(ax+ax)-1.
      return
    end subroutine snhcsh
!----------------

!*********************************************************
! function CURV2
!*********************************************************

real  function curv2 (t,n,x,y,yp,sigma) result (curv2result)
!
      integer(kind=int32) ::  n
      real(kind=double) t,x(n),y(n),yp(n),sigma
      integer(kind=int32) :: im1, i
      real(kind=double) :: sigmap, del1, del2, dels, sum, sigdel, s1, s2, ss
      real(kind=double) :: dummy
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this function interpolates a curve at a given point
! using a spline under tension. the subroutine curv1 should
! be called earlier to determine certain necessary
! parameters.
!
! on input--
!
!   t contains a real value to be mapped onto the interpo-
!   lating curve.
!
!   n contains the number of points which were specified to
!   determine the curve.
!
!   x and y are arrays containing the abscissae and
!   ordinates, respectively, of the specified points.
!
!   yp is an array of second derivative values of the curve
!   at the nodes.
!
! and
!
!   sigma contains the tension factor (its sign is ignored).
!
! the parameters n, x, y, yp, and sigma should be input
! unaltered from the output of curv1.
!
! on output--
!
!   curv2 contains the interpolated value.
!
! none of the input parameters are altered.
!
! this function references package modules intrvl and
! snhcsh.
!
!-----------------------------------------------------------
!
! determine interval
!
      im1 = intrvl(t,x,n)
      i = im1+1
!
! denormalize tension factor
!
      sigmap = abs(sigma)*float(n-1)/(x(n)-x(1))
!
! set up and perform interpolation
!
      del1 = t-x(im1)
      del2 = x(i)-t
      dels = x(i)-x(im1)
      sum = (y(i)*del1+y(im1)*del2)/dels
      if (sigmap .ne. 0.) go to 1
      curv2result=sum-del1*del2*(yp(i)*(del1+dels)+yp(im1)* &
              (del2+dels))/(6.*dels)
      return
    1 sigdel = sigmap*dels
      call snhcsh (ss,dummy,sigdel,-1)
      call snhcsh (s1,dummy,sigmap*del1,-1)
      call snhcsh (s2,dummy,sigmap*del2,-1)
      curv2result = sum+(yp(i)*del1*(s1-ss)+yp(im1)*del2*(s2-ss))/ &
                  (sigdel*sigmap*(1.+ss))
      return
    end function curv2
!-------------------------------------------
integer function intrvl (t,x,n)
!
      integer(kind=int32) ::  n
      real(kind=double) ::  t,x(n)
      real(kind=double) :: tt
      integer(kind=int32) :: i, il, ih
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this function determines the index of the interval
! (determined by a given increasing sequence) in which
! a given value lies.
!
! on input--
!
!   t is the given value.
!
!   x is a vector of strictly increasing values.
!
! and
!
!   n is the length of x (n .ge. 2).
!
! on output--
!
!   intrvl returns an integer i such that
!
!          i =  1       if         e   t .lt. x(2)  ,
!          i =  n-1     if x(n-1) .le. t            ,
!          otherwise       x(i)  .le. t .le. x(i+1),
!
! none of the input parameters are altered.
!
!-----------------------------------------------------------
!
      save i
      data i /1/
!
      tt = t
!
! check for illegal i
!
      if (i .ge. n) i = n/2
!
! check old interval and extremes
!
      if (tt .lt. x(i)) then
        if (tt .le. x(2)) then
          i = 1
          intrvl = 1
          return
        else
          il = 2
          ih = i
        end if
      else if (tt .le. x(i+1)) then
        intrvl = i
        return
      else if (tt .ge. x(n-1)) then
        i = n-1
        intrvl = n-1
        return
      else
        il = i+1
        ih = n-1
      end if
!
! binary search loop
!
    1 i = (il+ih)/2
      if (tt .lt. x(i)) then
         ih = i
      else if (tt .gt. x(i+1)) then
         il = i+1
      else
         intrvl = i
         return
      end if
      go to 1
    end function intrvl

!---------------------

 subroutine curvs (n,x,y,d,isw,s,eps,ys,ysp,sigma,temp,ierr)
!
      integer(kind=int32) ::  n,isw,ierr
      real(kind=double) :: x(n),y(n),d(n),s,eps,ys(n),ysp(n),sigma,temp(n,9)

!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute a smoothing spline under tension. for a given
! increasing sequence of abscissae (x(i)), i = 1,..., n and
! associated ordinates (y(i)), i = 1,..., n, the function
! determined minimizes the summation from i = 1 to n-1 of
! the square of the second derivative of f plus sigma
! squared times the difference of the first derivative of f
! and (f(x(i+1))-f(x(i)))/(x(i+1)-x(i)) squared, over all
! functions f with two continuous derivatives such that the
! summation of the square of (f(x(i))-y(i))/d(i) is less
! than or equal to a given constant s, where (d(i)), i = 1,
! ..., n are a given set of observation weights. the
! function determined is a spline under tension with third
! derivative discontinuities at (x(i)), i = 2,..., n-1. for
! actual computation of points on the curve it is necessary
! to call the function curv2. the determination of the curve
! is performed by subroutine curvss, the subroutine curvs
! only decomposes the workspace for curvss.
!
! on input--
!
!   n is the number of values to be smoothed (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   values to be smoothed.
!
!   y is an array of the n ordinates of the values to be
!   smoothed, (i. e. y(k) is the functional value
!   corresponding to x(k) ).
!
!   d is a parameter containing the observation weights.
!   this may either be an array of length n or a scalar
!   (interpreted as a constant). the value of d
!   corresponding to the observation (x(k),y(k)) should
!   be an approximation to the standard deviation of error.
!
!   isw contains a switch indicating whether the parameter
!   d is to be considered a vector or a scalar,
!          = 0 if d is an array of length n,
!          = 1 if d is a scalar.
!
!   s contains the value controlling the smoothing. this
!   must be non-negative. for s equal to zero, the
!   subroutine does interpolation, larger values lead to
!   smoother funtions. if parameter d contains standard
!   deviation estimates, a reasonable value for s is
!   float(n).
!
!   eps contains a tolerance on the relative precision to
!   which s is to be interpreted. this must be greater than
!   or equal to zero and less than or equal to one. a
!   reasonable value for eps is sqrt(2./float(n)).
!
!   ys is an array of length at least n.
!
!   ysp is an array of length at least n.
!
!   sigma contains the tension factor. this value indicates
!   the degree to which the first derivative part of the
!   smoothing functional is emphasized. if sigma is nearly
!   zero (e. g. .001) the resulting curve is approximately a
!   cubic spline. if sigma is large (e. g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results. a standard value for
!   sigma is approximately 1.
!
! and
!
!   temp is an array of length at least 9*n which is used
!   for scratch storage.
!
! on output--
!
!   ys contains the smoothed ordinate values.
!
!   ysp contains the values of the second derivative of the
!   smoothed curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if s is negative,
!        = 3 if eps is negative or greater than one,
!        = 4 if x-values are not strictly increasing,
!        = 5 if a d-value is non-positive.
!
! and
!
!   n, x, y, d, isw, s, eps, and sigma are unaltered.
!
! this subroutine references package modules curvss, terms,
! and snhcsh.
!
!-----------------------------------------------------------
!
! decompose temp into nine arrays and call curvss
!
      call curvss (n,x,y,d,isw,s,eps,ys,ysp,sigma,temp(1,1), &
                   temp(1,2),temp(1,3),temp(1,4),temp(1,5), &
                   temp(1,6),temp(1,7),temp(1,8),temp(1,9), &
                   ierr)
      return
    end subroutine curvs
!--------------------
!**********************************************************
! subroutine CURVSS
!**********************************************************
      subroutine curvss (n,x,y,d,isw,s,eps,ys,ysp,sigma,td, &
                         tsd1,hd,hsd1,hsd2,rd,rsd1,rsd2,v, &
                         ierr)
!
      integer(kind=int32) ::  n,isw,ierr
      real(kind=double) ::  x(n),y(n),d(n),s,eps,ys(n),ysp(n),sigma,td(n) 
      real(kind=double) ::  tsd1(n),hd(n),hsd1(n),hsd2(n),rd(n),rsd1(n)
      real(kind=double) ::  rsd2(n),v(n)
!
      real(kind=double) :: p, rdim1, yspim2, sigmap, delxi1, delyi1
      real(kind=double) :: dim1, delxi, delyi, di, sl, su, betapp, betap
      real(kind=double) :: alphap, alpha, beta, hsd1p, hdim1, hdi, rsd2i
      real(kind=double) :: rsd1i, sum, f, g, wim2, wim1, tui, wi, h, step
      integer(kind=int32) :: nm1, nm3, i, ibak
!
!                                 coded by alan kaylor cline
!                           from fitpack -- january 26, 1987
!                        a curve and surface fitting package
!                      a product of pleasant valley software
!                  8603 altus cove, austin, texas 78759, usa
!
! this subroutine determines the parameters necessary to
! compute a smoothing spline under tension. for a given
! increasing sequence of abscissae (x(i)), i = 1,..., n and
! associated ordinates (y(i)), i = 1,..., n, the function
! determined minimizes the summation from i = 1 to n-1 of
! the square of the second derivative of f plus sigma
! squared times the difference of the first derivative of f
! and (f(x(i+1))-f(x(i)))/(x(i+1)-x(i)) squared, over all
! functions f with two continuous derivatives such that the
! summation of the square of (f(x(i))-y(i))/d(i) is less
! than or equal to a given constant s, where (d(i)), i = 1,
! ..., n are a given set of observation weights. the
! function determined is a spline under tension with third
! derivative discontinuities at (x(i)), i = 2,..., n-1. for
! actual computation of points on the curve it is necessary
! to call the function curv2.
!
! on input--
!
!   n is the number of values to be smoothed (n.ge.2).
!
!   x is an array of the n increasing abscissae of the
!   values to be smoothed.
!
!   y is an array of the n ordinates of the values to be
!   smoothed, (i. e. y(k) is the functional value
!   corresponding to x(k) ).
!
!   d is a parameter containing the observation weights.
!   this may either be an array of length n or a scalar
!   (interpreted as a constant). the value of d
!   corresponding to the observation (x(k),y(k)) should
!   be an approximation to the standard deviation of error.
!
!   isw contains a switch indicating whether the parameter
!   d is to be considered a vector or a scalar,
!          = 0 if d is an array of length n,
!          = 1 if d is a scalar.
!
!   s contains the value controlling the smoothing. this
!   must be non-negative. for s equal to zero, the
!   subroutine does interpolation, larger values lead to
!   smoother funtions. if parameter d contains standard
!   deviation estimates, a reasonable value for s is
!   float(n).
!
!   eps contains a tolerance on the relative precision to
!   which s is to be interpreted. this must be greater than
!   or equal to zero and less than equal or equal to one. a
!   reasonable value for eps is sqrt(2./float(n)).
!
!   ys is an array of length at least n.
!
!   ysp is an array of length at least n.
!
!   sigma contains the tension factor. this value indicates
!   the degree to which the first derivative part of the
!   smoothing functional is emphasized. if sigma is nearly
!   zero (e. g. .001) the resulting curve is approximately a
!   cubic spline. if sigma is large (e. g. 50.) the
!   resulting curve is nearly a polygonal line. if sigma
!   equals zero a cubic spline results. a standard value for
!   sigma is approximately 1.
!
! and
!
!   td, tsd1, hd, hsd1, hsd2, rd, rsd1, rsd2, and v are
!   arrays of length at least n which are used for scratch
!   storage.
!
! on output--
!
!   ys contains the smoothed ordinate values.
!
!   ysp contains the values of the second derivative of the
!   smoothed curve at the given nodes.
!
!   ierr contains an error flag,
!        = 0 for normal return,
!        = 1 if n is less than 2,
!        = 2 if s is negative,
!        = 3 if eps is negative or greater than one,
!        = 4 if x-values are not strictly increasing,
!        = 5 if a d-value is non-positive.
!
! and
!
!   n, x, y, d, isw, s, eps, and sigma are unaltered.
!
! this subroutine references package modules terms and
! snhcsh.
!
!-----------------------------------------------------------
! 
      if (n .lt. 2) go to 16
      if (s .lt. 0.) go to 17
      if (eps .lt. 0. .or. eps .gt. 1.) go to 18
      ierr = 0
      p = 0.
      v(1) = 0.
      v(n) = 0.
      ysp(1) = 0.
      ysp(n) = 0.
      if (n .eq. 2) go to 14
      rsd1(1) = 0.
      rd(1) = 0.
      rsd2(n) = 0.
      rdim1 = 0.
      yspim2 = 0.
!
! denormalize tension factor
!
      sigmap = abs(sigma)*float(n-1)/(x(n)-x(1))
!
! form t matrix and second differences of y into ys
!
      nm1 = n-1
      nm3 = n-3
      delxi1 = 1.
      delyi1 = 0.
      dim1 = 0.
      do  i = 1,nm1
        delxi = x(i+1)-x(i)
        if (delxi .le. 0.) go to 19
        delyi = (y(i+1)-y(i))/delxi
        ys(i) = delyi-delyi1
        call terms (di,tsd1(i+1),sigmap,delxi)
        td(i) = di+dim1
        hd(i) = -(1./delxi+1./delxi1)
        hsd1(i+1) = 1./delxi
        delxi1 = delxi
        delyi1 = delyi
        dim1 = di
     enddo
!
! calculate lower and upper tolerances
!
      sl = s*(1.-eps)
      su = s*(1.+eps)
      if (isw .eq. 1) go to 3
!
! form h matrix - d array
!
      if (d(1) .le. 0. .or. d(2) .le. 0.) go to 20
      betapp = 0.
      betap = 0.
      alphap = 0.
      do  i = 2,nm1
        alpha = hd(i)*d(i)*d(i)
        if (d(i+1) .le. 0.) go to 20
        beta = hsd1(i+1)*d(i+1)*d(i+1)
        hd(i) = (hsd1(i)*d(i-1))**2+alpha*hd(i) &
                                   +beta*hsd1(i+1)
        hsd2(i) = hsd1(i)*betapp
        hsd1(i) = hsd1(i)*(alpha+alphap)
        alphap = alpha
        betapp = betap
        betap = beta
     enddo
      go to 5
!
! form h matrix - d constant
!
    3 if (d(1) .le. 0.) go to 20
      sl = d(1)*d(1)*sl
      su = d(1)*d(1)*su
      hsd1p = 0.
      hdim1 = 0.
      do  i = 2,nm1
        hdi = hd(i)
        hd(i) = hsd1(i)*hsd1(i)+hdi*hdi+hsd1(i+1)*hsd1(i+1)
        hsd2(i) = hsd1(i)*hsd1p
        hsd1p = hsd1(i)
        hsd1(i) = hsd1p*(hdi+hdim1)
        hdim1 = hdi
     enddo
!
! top of iteration
! cholesky factorization of p*t+h into r
!
    5 do i = 2,nm1
        rsd2i = hsd2(i)
        rsd1i = p*tsd1(i)+hsd1(i)-rsd2i*rsd1(i-1)
        rsd2(i) = rsd2i*rdim1
        rdim1 = rd(i-1)
        rsd1(i) = rsd1i*rdim1
        rd(i) = 1./(p*td(i)+hd(i)-rsd1i*rsd1(i) &
                                 -rsd2i*rsd2(i))
        ysp(i) = ys(i)-rsd1(i)*ysp(i-1)-rsd2(i)*yspim2
        yspim2 = ysp(i-1)
     enddo
!
! back solve of r(transpose)* r * ysp = ys
!
      ysp(nm1) = rd(nm1)*ysp(nm1)
      if (n .eq. 3) go to 8
      do  ibak = 1,nm3
        i = nm1-ibak
        ysp(i) = rd(i)*ysp(i)-rsd1(i+1)*ysp(i+1) &
                             -rsd2(i+2)*ysp(i+2)
     enddo
    8 sum = 0.
      delyi1 = 0.
      if (isw .eq. 1) go to 10
!
! calculation of residual norm
!  - d array
!
      do  i = 1,nm1
        delyi = (ysp(i+1)-ysp(i))/(x(i+1)-x(i))
        v(i) = (delyi-delyi1)*d(i)*d(i)
        sum = sum+v(i)*(delyi-delyi1)
        delyi1 = delyi
     enddo
      v(n) = -delyi1*d(n)*d(n)
      go to 12
!
! calculation of residual norm
!  - d constant
!
   10 do  i = 1,nm1
        delyi = (ysp(i+1)-ysp(i))/(x(i+1)-x(i))
        v(i) = delyi-delyi1
        sum = sum+v(i)*(delyi-delyi1)
        delyi1 = delyi
     enddo
      v(n) = -delyi1
   12 sum = sum-v(n)*delyi1
!
! test for convergence
!
      if (sum .le. su) go to 14
!
! calculation of newton correction
!
      f = 0.
      g = 0.
      wim2 = 0.
      wim1 = 0.
      do  i = 2,nm1
        tui = tsd1(i)*ysp(i-1)+td(i)*ysp(i) &
                              +tsd1(i+1)*ysp(i+1)
        wi = tui-rsd1(i)*wim1-rsd2(i)*wim2
        f = f+tui*ysp(i)
        g = g+wi*wi*rd(i)
        wim2 = wim1
        wim1 = wi
     enddo
      h = f-p*g
      if (h .le. 0.) go to 14
!
! update p - newton step
!
      step = (sum-sqrt(sum*sl))/h
      if (sl .ne. 0.) step = step*sqrt(sum/sl)
      p = p+step
      go to 5
!
! store smoothed y-values and second derivatives
!
   14 do  i = 1,n
        ys(i) = y(i)-v(i)
        ysp(i) = p*ysp(i)
     enddo
      return
!
! n less than 2
!
   16 ierr = 1
      return
!
! s negative
!
   17 ierr = 2
      return
!
! eps negative or greater than 1
!
   18 ierr = 3
      return
!
! x-values not strictly increasing
!
   19 ierr = 4
      return
!
! weight non-positive
!
   20 ierr = 5
      return
    end subroutine curvss
!--------------------

! wsc@research.bell-labs.com Mon Dec 30 16:55 EST 1985
! W. S. Cleveland
! Bell Laboratories
! Murray Hill NJ 07974
! 
! outline of this file:
!    lines 1-72   introduction
!        73-177   documentation for lowess
!       178-238   ratfor version of lowess
!       239-301   documentation for lowest
!       302-350   ratfor version of lowest
!       351-end   test driver and fortran version of lowess and lowest
! 
!   a multivariate version is available by "send dloess from a"
! 
!              COMPUTER PROGRAMS FOR LOCALLY WEIGHTED REGRESSION
! 
!             This package consists  of  two  FORTRAN  programs  for
!        smoothing    scatterplots   by   robust   locally   weighted
!        regression, or lowess.   The  principal  routine  is  LOWESS
!        which   computes   the  smoothed  values  using  the  method
!        described in The Elements of Graphing Data, by William S.
!        Cleveland    (Wadsworth,    555 Morego   Street,   Monterey,
!        California 93940).
! 
!             LOWESS calls a support routine, LOWEST, the code for
!        which is included. LOWESS also calls a routine  SORT,  which
!        the user must provide.
! 
!             To reduce the computations, LOWESS  requires  that  the
!        arrays  X  and  Y,  which  are  the  horizontal and vertical
!        coordinates, respectively, of the scatterplot, be such  that
!        X  is  sorted  from  smallest  to  largest.   The  user must
!        therefore use another sort routine which will sort X  and  Y
!        according  to X.
!             To summarize the scatterplot, YS,  the  fitted  values,
!        should  be  plotted  against X.   No  graphics  routines are
!        available in the package and must be supplied by the user.
! 
!             The FORTRAN code for the routines LOWESS and LOWEST has
!        been   generated   from   higher   level   RATFOR   programs
!        (B. W. Kernighan, ``RATFOR:  A Preprocessor for  a  Rational
!        Fortran,''  Software Practice and Experience, Vol. 5 (1975),
!        which are also included.
! 
!             The following are data and output from LOWESS that  can
!        be  used  to check your implementation of the routines.  The
!        notation (10)v means 10 values of v.
! 
! 
! 
! 
!        X values:
!          1  2  3  4  5  (10)6  8  10  12  14  50
! 
!        Y values:
!           18  2  15  6  10  4  16  11  7  3  14  17  20  12  9  13  1  8  5  19
! 
! 
!        YS values with F = .25, NSTEPS = 0, DELTA = 0.0
!         13.659  11.145  8.701  9.722  10.000  (10)11.300  13.000  6.440  5.596
!           5.456  18.998
! 
!        YS values with F = .25, NSTEPS = 0 ,  DELTA = 3.0
!          13.659  12.347  11.034  9.722  10.511  (10)11.300  13.000  6.440  5.596
!            5.456  18.998
! 
!        YS values with F = .25, NSTEPS = 2, DELTA = 0.0
!          14.811  12.115  8.984  9.676  10.000  (10)11.346  13.000  6.734  5.744
!            5.415  18.998
! 
! 
! 
! 
!                                   LOWESS
! 
! 
! 
!        Calling sequence
! 
!        CALL LOWESS(X,Y,N,F,NSTEPS,DELTA,YS,RW,RES)
! 
!        Purpose
! 
!        LOWESS computes the smooth of a scatterplot of Y  against  X
!        using  robust  locally  weighted regression.  Fitted values,
!        YS, are computed at each of the  values  of  the  horizontal
!        axis in X.
! 
!        Argument description
! 
!              X = Input; abscissas of the points on the
!                  scatterplot; the values in X must be ordered
!                  from smallest to largest.
!              Y = Input; ordinates of the points on the
!                  scatterplot.
!              N = Input; dimension of X,Y,YS,RW, and RES.
!              F = Input; specifies the amount of smoothing; F is
!                  the fraction of points used to compute each
!                  fitted value; as F increases the smoothed values
!                  become smoother; choosing F in the range .2 to
!                  .8 usually results in a good fit; if you have no
!                  idea which value to use, try F = .5.
!         NSTEPS = Input; the number of iterations in the robust
!                  fit; if NSTEPS = 0, the nonrobust fit is
!                  returned; setting NSTEPS equal to 2 should serve
!                  most purposes.
!          DELTA = input; nonnegative parameter which may be used
!                  to save computations; if N is less than 100, set
!                  DELTA equal to 0.0; if N is greater than 100 you
!                  should find out how DELTA works by reading the
!                  additional instructions section.
!             YS = Output; fitted values; YS(I) is the fitted value
!                  at X(I); to summarize the scatterplot, YS(I)
!                  should be plotted against X(I).
!             RW = Output; robustness weights; RW(I) is the weight
!                  given to the point (X(I),Y(I)); if NSTEPS = 0,
!                  RW is not used.
!            RES = Output; residuals; RES(I) = Y(I)-YS(I).
! 
! 
!        Other programs called
! 
!               LOWEST
!               SSORT
! 
!        Additional instructions
! 
!        DELTA can be used to save computations.   Very  roughly  the
!        algorithm  is  this:   on the initial fit and on each of the
!        NSTEPS iterations locally weighted regression fitted  values
!        are computed at points in X which are spaced, roughly, DELTA
!        apart; then the fitted values at the  remaining  points  are
!        computed  using  linear  interpolation.   The  first locally
!        weighted regression (l.w.r.) computation is carried  out  at
!        X(1)  and  the  last  is  carried  out at X(N).  Suppose the
!        l.w.r. computation is carried out at  X(I).   If  X(I+1)  is
!        greater  than  or  equal  to  X(I)+DELTA,  the  next  l.w.r.
!        computation is carried out at X(I+1).   If  X(I+1)  is  less
!        than X(I)+DELTA, the next l.w.r.  computation is carried out
!        at the largest X(J) which is greater than or equal  to  X(I)
!        but  is not greater than X(I)+DELTA.  Then the fitted values
!        for X(K) between X(I)  and  X(J),  if  there  are  any,  are
!        computed  by  linear  interpolation  of the fitted values at
!        X(I) and X(J).  If N is less than 100 then DELTA can be  set
!        to  0.0  since  the  computation time will not be too great.
!        For larger N it is typically not necessary to carry out  the
!        l.w.r.  computation for all points, so that much computation
!        time can be saved by taking DELTA to be  greater  than  0.0.
!        If  DELTA =  Range  (X)/k  then,  if  the  values  in X were
!        uniformly  scattered  over  the  range,  the   full   l.w.r.
!        computation  would be carried out at approximately k points.
!        Taking k to be 50 often works well.
! 
!        Method
! 
!        The fitted values are computed by using the nearest neighbor
!        routine  and  robust locally weighted regression of degree 1
!        with the tricube weight function.  A few additional features
!        have  been  added.  Suppose r is FN truncated to an integer.
!        Let  h  be  the  distance  to  the  r-th  nearest   neighbor
!        from X(I).   All  points within h of X(I) are used.  Thus if
!        the r-th nearest neighbor is exactly the  same  distance  as
!        other  points,  more  than r points can possibly be used for
!        the smooth at  X(I).   There  are  two  cases  where  robust
!        locally  weighted regression of degree 0 is actually used at
!        X(I).  One case occurs when  h  is  0.0.   The  second  case
!        occurs  when  the  weighted  standard error of the X(I) with
!        respect to the weights w(j) is  less  than  .001  times  the
!        range  of the X(I), where w(j) is the weight assigned to the
!        j-th point of X (the tricube  weight  times  the  robustness
!        weight)  divided by the sum of all of the weights.  Finally,
!        if the w(j) are all zero for the smooth at X(I), the  fitted
!        value is taken to be Y(I).
! 
! 
! 
! 
!  subroutine lowess(x,y,n,f,nsteps,delta,ys,rw,res)
!  real x(n),y(n),ys(n),rw(n),res(n)
!  logical ok
!  if (n<2){ ys(1) = y(1); return }
!  ns = max0(min0(ifix(f*float(n)),n),2)  # at least two, at most n points
!  for(iter=1; iter<=nsteps+1; iter=iter+1){      # robustness iterations
!         nleft = 1; nright = ns
!         last = 0        # index of prev estimated point
!         i = 1   # index of current point
!         repeat{
!                 while(nright<n){
!  # move nleft, nright to right if radius decreases
!                         d1 = x(i)-x(nleft)
!                         d2 = x(nright+1)-x(i)
!  # if d1<=d2 with x(nright+1)==x(nright), lowest fixes
!                         if (d1<=d2) break
!  # radius will not decrease by move right
!                         nleft = nleft+1
!                         nright = nright+1
!                         }
!                 call lowest(x,y,n,x(i),ys(i),nleft,nright,res,iter>1,rw,ok)
!  # fitted value at x(i)
!                 if (!ok) ys(i) = y(i)
!  # all weights zero - copy over value (all rw==0)
!                 if (last<i-1) { # skipped points -- interpolate
!                         denom = x(i)-x(last)    # non-zero - proof?
!                         for(j=last+1; j<i; j=j+1){
!                                 alpha = (x(j)-x(last))/denom
!                                 ys(j) = alpha*ys(i)+(1.0-alpha)*ys(last)
!                                 }
!                         }
!                 last = i        # last point actually estimated
!                 cut = x(last)+delta     # x coord of close points
!                 for(i=last+1; i<=n; i=i+1){     # find close points
!                         if (x(i)>cut) break     # i one beyond last pt within cut
!                         if(x(i)==x(last)){      # exact match in x
!                                 ys(i) = ys(last)
!                                 last = i
!                                 }
!                         }
!                 i=max0(last+1,i-1)
!  # back 1 point so interpolation within delta, but always go forward
!                 } until(last>=n)
!         do i = 1,n      # residuals
!                 res(i) = y(i)-ys(i)
!         if (iter>nsteps) break  # compute robustness weights except last time
!         do i = 1,n
!                 rw(i) = abs(res(i))
!         call sort(rw,n)
!         m1 = 1+n/2; m2 = n-m1+1
!         cmad = 3.0*(rw(m1)+rw(m2))      # 6 median abs resid
!         c9 = .999*cmad; c1 = .001*cmad
!         do i = 1,n {
!                 r = abs(res(i))
!                 if(r<=c1) rw(i)=1.      # near 0, avoid underflow
!                 else if(r>c9) rw(i)=0.  # near 1, avoid underflow
!                 else rw(i) = (1.0-(r/cmad)**2)**2
!                 }
!         }
!  return
!  end
! 
! 
! 
! 
!                                   LOWEST
! 
! 
! 
!        Calling sequence
! 
!        CALL LOWEST(X,Y,N,XS,YS,NLEFT,NRIGHT,W,USERW,RW,OK)
! 
!        Purpose
! 
!        LOWEST is a support routine for LOWESS and  ordinarily  will
!        not  be  called  by  the  user.   The  fitted  value, YS, is
!        computed  at  the  value,  XS,  of  the   horizontal   axis.
!        Robustness  weights,  RW,  can  be employed in computing the
!        fit.
! 
!        Argument description
! 
! 
!              X = Input; abscissas of the points on the
!                  scatterplot; the values in X must be ordered
!                  from smallest to largest.
!              Y = Input; ordinates of the points on the
!                  scatterplot.
!              N = Input; dimension of X,Y,W, and RW.
!             XS = Input; value of the horizontal axis at which the
!                  smooth is computed.
!             YS = Output; fitted value at XS.
!          NLEFT = Input; index of the first point which should be
!                  considered in computing the fitted value.
!         NRIGHT = Input; index of the last point which should be
!                  considered in computing the fitted value.
!              W = Output; W(I) is the weight for Y(I) used in the
!                  expression for YS, which is the sum from
!                  I = NLEFT to NRIGHT of W(I)*Y(I); W(I) is
!                  defined only at locations NLEFT to NRIGHT.
!          USERW = Input; logical variable; if USERW is .TRUE., a
!                  robust fit is carried out using the weights in
!                  RW; if USERW is .FALSE., the values in RW are
!                  not used.
!             RW = Input; robustness weights.
!             OK = Output; logical variable; if the weights for the
!                  smooth are all 0.0, the fitted value, YS, is not
!                  computed and OK is set equal to .FALSE.; if the
!                  fitted value is computed OK is set equal to
! 
! 
!        Method
! 
!        The smooth at XS is computed using (robust) locally weighted
!        regression of degree 1.  The tricube weight function is used
!        with h equal to the maximum of XS-X(NLEFT) and X(NRIGHT)-XS.
!        Two  cases  where  the  program  reverts to locally weighted
!        regression of degree 0 are described  in  the  documentation
!        for LOWESS.
! 
! 
! 
! 
!  subroutine lowest(x,y,n,xs,ys,nleft,nright,w,userw,rw,ok)
!  real x(n),y(n),w(n),rw(n)
!  logical userw,ok
!  range = x(n)-x(1)
!  h = amax1(xs-x(nleft),x(nright)-xs)
!  h9 = .999*h
!  h1 = .001*h
!  a = 0.0        # sum of weights
!  for(j=nleft; j<=n; j=j+1){     # compute weights (pick up all ties on right)
!         w(j)=0.
!         r = abs(x(j)-xs)
!         if (r<=h9) {    # small enough for non-zero weight
!                 if (r>h1) w(j) = (1.0-(r/h)**3)**3
!                 else      w(j) = 1.
!                 if (userw) w(j) = rw(j)*w(j)
!                 a = a+w(j)
!                 }
!         else if(x(j)>xs)break   # get out at first zero wt on right
!         }
!  nrt=j-1        # rightmost pt (may be greater than nright because of ties)
!  if (a<=0.0) ok = FALSE
!  else { # weighted least squares
!         ok = TRUE
!         do j = nleft,nrt
!                 w(j) = w(j)/a   # make sum of w(j) == 1
!         if (h>0.) {     # use linear fit
!                 a = 0.0
!                 do j = nleft,nrt
!                         a = a+w(j)*x(j) # weighted center of x values
!                 b = xs-a
!                 c = 0.0
!                 do j = nleft,nrt
!                         c = c+w(j)*(x(j)-a)**2
!                 if(sqrt(c)>.001*range) {
!  # points are spread out enough to compute slope
!                         b = b/c
!                         do j = nleft,nrt
!                                 w(j) = w(j)*(1.0+b*(x(j)-a))
!                         }
!                 }
!         ys = 0.0
!         do j = nleft,nrt
!                 ys = ys+w(j)*y(j)
!         }
!  return
!  end
! 
! 
! 
!  test driver for lowess
!  for expected output, see introduction
!     real x(20), y(20), ys(20), rw(20), res(20)
!     data x /1,2,3,4,5,10*6,8,10,12,14,50/
!     data y /18,2,15,6,10,4,16,11,7,3,14,17,20,12,9,13,1,8,5,19/
!     call lowess(x,y,20,.25,0,0.,ys,rw,res)
!     write(6,*) ys
!     call lowess(x,y,20,.25,0,3.,ys,rw,res)
!     write(6,*) ys
!     call lowess(x,y,20,.25,2,0.,ys,rw,res)
!     write(6,*) ys
!     end
!**************************************************************
!  Fortran output from ratfor
!
      subroutine lowess(x, y, n, f, nsteps, delta, ys, rw, res)
      integer(kind=int32) ::  n
      integer(kind=int32) ::  nsteps
      real(kind=double) ::  x(n), y(n), f, delta, ys(n), rw(n)
      real(kind=double) ::  res(n)
      integer(kind=int32) ::  nright, min0, max0, i, j, ifix
      integer(kind=int32) ::  iter, last, m1, m2, ns, nleft
      real(kind=double) :: abs, cut, cmad, r, d1, d2
      real(kind=double) ::  c1, c9, alpha, denom, float
      logical ok
      if (n .ge. 2) goto 1
         ys(1) = y(1)
         return
! at least two, at most n points
!   1  ns = max(min(ifix(f*float(n))), n), 2)
   1  ns = max(min(int(f*float(n)), n), 2)
      iter = 1
         goto  3
   2     iter = iter+1
   3     if (iter .gt. nsteps+1) goto  22
! robustness iterations
         nleft = 1
         nright = ns
! index of prev estimated point
         last = 0
! index of current point
         i = 1
   4        if (nright .ge. n) goto  5
! move nleft, nright to right if radius decreases
               d1 = x(i)-x(nleft)
! if d1<=d2 with x(nright+1)==x(nright), lowest fixes
               d2 = x(nright+1)-x(i)
               if (d1 .le. d2) goto  5
! radius will not decrease by move right
               nleft = nleft+1
               nright = nright+1
               goto  4
! fitted value at x(i)
   5        call lowest(x, y, n, x(i), ys(i), nleft, nright, res, iter &
                 .gt. 1, rw, ok)
            if (.not. ok) ys(i) = y(i)
! all weights zero - copy over value (all rw==0)
            if (last .ge. i-1) goto 9
               denom = x(i)-x(last)
! skipped points -- interpolate
! non-zero - proof?
               j = last+1
                  goto  7
   6              j = j+1
   7              if (j .ge. i) goto  8
                  alpha = (x(j)-x(last))/denom
                  ys(j) = alpha*ys(i)+(1.0-alpha)*ys(last)
                  goto  6
   8           continue
! last point actually estimated
   9        last = i
! x coord of close points
            cut = x(last)+delta
            i = last+1
               goto  11
  10           i = i+1
  11           if (i .gt. n) goto  13
! find close points
               if (x(i) .gt. cut) goto  13
! i one beyond last pt within cut
               if (x(i) .ne. x(last)) goto 12
                  ys(i) = ys(last)
! exact match in x
                  last = i
  12           continue
               goto  10
! back 1 point so interpolation within delta, but always go forward
  13        i = max(last+1, i-1)
  14        if (last .lt. n) goto  4
! residuals
         do  15 i = 1, n
            res(i) = y(i)-ys(i)
  15        continue
         if (iter .gt. nsteps) goto  22
! compute robustness weights except last time
         do  16 i = 1, n
            rw(i) = abs(res(i))
  16        continue
!         call sort(rw, n)
         call ssort(n,rw)
         m1 = n/2+1
         m2 = n-m1+1
! 6 median abs resid
         cmad = 3.0*(rw(m1)+rw(m2))
         c9 = .999*cmad
         c1 = .001*cmad
         do  21 i = 1, n
            r = abs(res(i))
            if (r .gt. c1) goto 17
               rw(i) = 1.
! near 0, avoid underflow
               goto  20
  17           if (r .le. c9) goto 18
                  rw(i) = 0.
! near 1, avoid underflow
                  goto  19
  18              rw(i) = (1.0-(r/cmad)**2)**2
  19        continue
  20        continue
  21        continue
         goto  2
  22  return
      end subroutine lowess


      subroutine lowest(x, y, n, xs, ys, nleft, nright, w, userw &
           , rw, ok)
      integer(kind=int32) ::  n
      integer(kind=int32) ::  nleft, nright
      real(kind=double) :: x(n), y(n), xs, ys, w(n), rw(n)
      logical :: userw, ok
      integer(kind=int32) ::  nrt, j
      real(kind=double) ::  abs, a, b, c, h, r
      real(kind=double) ::  h1, sqrt, h9, amax1, range
      range = x(n)-x(1)
      h = max(xs-x(nleft), x(nright)-xs)
      h9 = .999*h
      h1 = .001*h
! sum of weights
      a = 0.0
      j = nleft
         goto  2
   1     j = j+1
   2     if (j .gt. n) goto  7
! compute weights (pick up all ties on right)
         w(j) = 0.
         r = abs(x(j)-xs)
         if (r .gt. h9) goto 5
            if (r .le. h1) goto 3
               w(j) = (1.0-(r/h)**3)**3
! small enough for non-zero weight
               goto  4
   3           w(j) = 1.
   4        if (userw) w(j) = rw(j)*w(j)
            a = a+w(j)
            goto  6
   5        if (x(j) .gt. xs) goto  7
! get out at first zero wt on right
   6     continue
         goto  1
! rightmost pt (may be greater than nright because of ties)
   7  nrt = j-1
      if (a .gt. 0.0) goto 8
         ok = .false.
         goto  16
   8     ok = .true.
! weighted least squares
         do  9 j = nleft, nrt
! make sum of w(j) == 1
            w(j) = w(j)/a
   9        continue
         if (h .le. 0.) goto 14
            a = 0.0
! use linear fit
            do  10 j = nleft, nrt
! weighted center of x values
               a = a+w(j)*x(j)
  10           continue
            b = xs-a
            c = 0.0
            do  11 j = nleft, nrt
               c = c+w(j)*(x(j)-a)**2
  11           continue
            if (sqrt(c) .le. .001*range) goto 13
               b = b/c
! points are spread out enough to compute slope
               do  12 j = nleft, nrt
                  w(j) = w(j)*(b*(x(j)-a)+1.0)
  12              continue
  13        continue
  14     ys = 0.0
         do  15 j = nleft, nrt
            ys = ys+w(j)*y(j)
  15        continue
  16  return
      end subroutine lowest

subroutine ssort(n,ra)
  integer(kind=int32) :: n
  real(kind=double) :: ra(*)
  integer(kind=int32) ::  i,j,l,ir
  real(kind=double) :: rra

  l=n/2+1
  ir=n

do 
  if(l.gt.1)then
     l=l-1
     rra=ra(l)
  else
     rra=ra(ir)
     ra(ir)=ra(1)
     ir=ir-1
     if(ir.le.1)then
        ra(1)=rra
        EXIT
     endif
!     ir=ir-1
  endif
  i=l
  j=l+l
  do while (j.le.ir)
     if(j.lt.ir)then
        if(ra(j).lt.ra(j+1))j=j+1
     endif
     if(rra.lt.ra(j))then
        ra(i)=ra(j)
        i=j
        j=j+j
     else
        j=ir+1
     endif
  enddo
  ra(i)=rra
enddo
end subroutine ssort

end module omgextraction_mod








