!+
! Module: omgrism_mod
!
! Authors:  Vladimir Yershov - MSSL
!          
! Description: Constructs an OM OSW FITS source spectra from the grism 
! image and source list  files produced by using omdetect
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!
! History:
! 
!    2003.10.27 VNY -  First version
!    2003.10.31, 2003.11.07 VNY - V0.2 Parameters for spectra regions 
!    are passed from omgchain; parameters for image smoothing removed
!    2003.11.10 V0.3 two lines exceeding 132 characters are shorten
!    2003.11.11 V0.4 a keyword (COMMENT) with the spectrum filename is added
!                    to the source-list table (of the input source list file) 
!    12.11.2003 VNY Checking the headers of the output spectrum file 
!               Applying the input parameters for the spectrum and background 
!               extraction regions
!    17.11.2003 VNY making extraction with a separated routine
!    28.11.2003 VNY set EFLAG to 2 if it corresponds to a genuine spectrum
!    02.12.2003 VNY the parameters src2spectrum and spectrumlength are removed as misleading
!    04.12.2003 VNY extraction of the undetected spectra is included
!    09.12.2003 VNY The spectra labels in the source-list file are changed: 
!                   now they correspod to the zero-order spectra 
!    07.01.2004 VNY Checks for optimal extraction 
!    19.01.2004 VNY Fixing the mistaken coordinates of zero- and first-order 
!               images in the headers of the spectra tables (SSC-SPR-3220)
!    26.01.2004 VNY extractionmode parameter is activated for the values 0 and 2
!    28.01.2004 - 01.02.2004 VNY adding a new output file - SPECLI - 
!                 list of the extracted spectra 
!    02.02.2004 VNY removing the column ORIGIN_ID from the table SPECLI
!    04.02.2004 VNY Checking the possibility of flagging those sources 
!    that fall inside of the spectra areas
!    06.02.2004 VNY Creating quality flags for the spectra based on 
!    solely zero-orders
!    12.02.2004 VNY Selectrion criteria for spectra are improved 
!    19.02.2004 VNY Protected from crushing when called by omgsource
!    23.02.2004 VNY a small bug removed (the task was always in
!                   the regime for omgsource)
!    03.03.2004 VNY Rejection of spurious spectra improved
!
module omgrism_mod

  use dal
  use cal
  use param  
  use errorHandling
  use frameTime_mod
  use datafilter
  use sasdal
!  use minim_mod
!  use mssllib
!  use omgnumerical_mod
  use omgextraction_mod

  implicit none 
  
  private 

  public omgrism

contains 
  
  !---------------------------------------------------------------------
  subroutine omgrism
  !---------------------------------------------------------------------

   ! variables
   ! a temp DAL dataset to find area of regions

!   type(DataSetT) :: tempSet !+
!   type(TableT) :: tempTable !+
!   type(ColumnT) :: tempColX, tempColY !+
   !type(DataSetT) :: inDataSetH                                       ! Input dataset handle.
   type(ArrayT) :: inArrayH                                          ! Input array handle.

!   integer(Kind=int32), pointer :: xArray(:), yArray(:) !+
!   integer(kind=int32) :: nPts !, nPix !+
   integer(kind=int32) :: dim(2),nx, ny !+
   integer(kind=int32) :: i,j ,k !+
   real(kind=single), pointer :: oswImage(:, :)
   real(kind=single)          :: exposureTime         ! The exposure time for an input image
   !integer             :: x0, y0             ! offset of the image
   !integer             :: xsize , ysize       ! size of the image
   integer :: ysize, xsize

   !character(len=1024) :: sourceRegFile !+
!   character(len=1024) :: expression !+
   character(len=1024) :: sourceListFile !+
   real(kind=single) :: sourceArea !+
!   type(DataSetT) :: sourceRegSetH !+
   type(DataSetT) :: grismSetH !+ 
   ! Input table handle
   type(TableT) :: inTableH !+
   integer, parameter :: noColumns = 20 !+
   type(ColumnT), dimension(noColumns) :: inColumnH !+
   integer(kind=int8), dimension(noColumns) :: inColumnFlag !+
   ! Input dataset handle
   type(DataSetT) :: sourceListSetH !+
   real(kind=double)              :: detectionLimitCountRate !+         ! The detection limit count rate (counts/second)
!   real(kind=double)              :: detectionLimitMagnitude         ! The detection limit magnitude
!   real(kind=double) :: error = 0.0 
!   real(kind=single) :: calApertureRadius = 12.0 !+         ! Radius used for coincidence losses   
!   real(kind=single) :: photometryApertureRadius=36.0 !+          ! Radius used for converting counts to magnitudes
 
   character(len=8) :: dataMode !+                                      ! Input file data mode.
   character(len=9) :: filter !+                                        ! Keywords.
   real(kind=single) :: exposure !+
   real(kind=double) :: frameTime !+                                    ! OM CCD frame time.
   real(kind=double) :: deadFrac  !+                                   ! Dead time in units of CCD frame time.
   integer(kind=int32) ccdMode !+                                       ! OM CCD data mode.
   integer(kind=int32) :: filterIdInt !+
!
   integer(kind=int32) :: tmexpId, frmTime
   character(len=80) :: dateObs, dateEnd, dataType, obsId, expId

   logical :: vignApp !+p
   logical :: backApp !+
   logical :: deadApp !+
   logical :: psfApp !+
   real(kind=single), dimension(:), pointer :: rate !+                  ! Rates pointer.
   real(kind=single), dimension(:), pointer :: rateErr !+               ! Rates error pointer.
   real(kind=double), dimension(:), pointer :: sourceRA !+              ! Source Right Ascension
   real(kind=double), dimension(:), pointer :: sourceDEC !+             ! Source Declination
   real(kind=double), dimension(:), pointer :: sourceLII !+             ! Source Galactic Longitude
   real(kind=double), dimension(:), pointer :: sourceBII !+             ! Source Galactic Lattitude
   integer(kind=int8), dimension(:), pointer :: sourceQF !+             ! Source quality flag
   integer(kind=int8), dimension(:), pointer :: sourceCF !+             ! Source confusion flag
   real(kind=single), dimension(:), pointer :: sourcePA !+              ! Source positional angle
   real(kind=single), dimension(:), pointer :: sourceFWHMaj !+              ! Source major semiaxis
   real(kind=single), dimension(:), pointer :: sourceFWHMin !+              ! Source minor semiaxis

   real(kind=single), dimension(:), pointer :: sourceFWHMaj2 !+              ! Source major semiaxis
   real(kind=single), dimension(:), pointer :: sourceFWHMin2 !+              ! Source minor semiaxis

   integer(kind=int32) :: iSource
   real(kind=single) :: sourceWx, sourceWy

   real(kind=single), dimension(:), pointer :: limitLeftMax
   real(kind=single), dimension(:), pointer :: limitRightMax

   integer(kind=int8), dimension(:), pointer :: sourceShape  ! Source shape flag (point/spectrum)
   integer(kind=int8), dimension(:), pointer :: sourceRelation ! Source relation flag

   real(kind = double), parameter :: PI = 3.141592653589793_double  
   real(kind = double), parameter :: degreesToRadians = PI / 180.0_double
   real(kind = double), parameter :: radiansToDegrees = 180.0_double / PI
   real(kind = double), parameter :: radiansToArcSecs = radiansToDegrees * 3600.0_double 

   real(kind=single), dimension(:), pointer :: sourceXCoord0 !+          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord0 !+         ! Source y-coordinate. 

   real(kind=single), dimension(:), pointer :: sourceXCoord !+          ! Source x-coordinate.
   real(kind=single), dimension(:), pointer :: sourceYCoord !+         ! Source y-coordinate. 
!---
   real(kind=single), dimension(:), pointer :: sourceApertureRadius !+  ! Source aperture radius
   real(kind=single), dimension(:), pointer :: backGroundRate !+        ! Source background rate pointer.
   real(kind=single), dimension(:), pointer :: mag !+                   ! Magnitude pointer.
   real(kind=single), dimension(:), pointer :: magErr !+                ! Magnitude error pointer.
   integer(kind=int8), dimension(:), pointer :: extended !+

   real(kind=double) :: xZero, yZero ! XY-coordinates of zero-order image
   real(kind=double) :: xZeroA, yZeroA ! XY-coordinates of zero-order image
   real(kind=double) :: xFirst, yFirst ! XY-coordinates of the first-order image
   real(kind=double) :: zeroMaxValue
   real(kind=double) :: spectrumHalfLength

   real(kind=single), dimension(:), pointer :: x0aux, x1aux, y0aux, y1aux
   logical :: extend
!   integer :: iLoop !+
   integer(kind=int32) :: nSources, nSpectraSources !+  ! Number of sources.
   integer(kind=int32) :: nBins !+


   ! output variables
   type(DataSetT) :: outSet0H ! output file including all the spectra
   !type(TableT) :: outTableH !+
   type(TableT) :: outTableLH !+
   type(TableT) :: outTableRH !+
   type(TableT) :: outTable0H
! VNY 28.01.2004
   type(DataSetT) :: outSpectraListSetH
   type(TableT) :: outSpectraTableH
!
   character(len=1024) :: outFile !+
   character(len=1024) :: grismFile !+
   character(len=1024) :: outFile2, outFile3 !+
   character(len=7) :: blockName                ! name for the current 
   character(len=8) :: blockNameL, blockNameR   ! block in the output file with tables 
                                                ! corresponding to the spectra
   character(len=1024) :: outFile4 !+
   !character(len=1024) :: outFileL, outFileR !+
   character(len=1024) :: outFileShortenName
   character(len=80) :: spectrumFileComment
   character(len=3) :: sourceIndex
   integer(kind=int16) :: posPoint

   ! rates file columns
   type(ColumnT) :: rateSourceOutCol0 !+
   type(ColumnT) :: rateSourceErrOutCol0 !+
   type(ColumnT) :: rateBackOutCol0 !+
   type(ColumnT) :: rateBackErrOutCol0 !+
!
   type(ColumnT) :: rateSourceOutColL !+
   type(ColumnT) :: rateSourceErrOutColL !+
   type(ColumnT) :: rateBackOutColL !+
   type(ColumnT) :: rateBackErrOutColL !+
!   type(ColumnT) :: timeSourceOutColL !+
   type(ColumnT) :: rateSourceOutColR !+
   type(ColumnT) :: rateSourceErrOutColR !+
   type(ColumnT) :: rateBackOutColR !+
   type(ColumnT) :: rateBackErrOutColR !+
!   type(ColumnT) :: timeSourceOutColR !+

   type(ColumnT) :: fluxOutCol0 !+
   type(ColumnT) :: fluxErrOutCol0 !+
   type(ColumnT) :: fluxOutColL !+
   type(ColumnT) :: fluxErrOutColL !+
   type(ColumnT) :: fluxOutColR !+
   type(ColumnT) :: fluxErrOutColR !+

   type(ColumnT) :: waveSourceOutColL
   type(ColumnT) :: waveSourceOutColR

   type(ColumnT) :: spectrumNoCol
   type(ColumnT) :: spectrumOriginCol
   integer(kind=int32), dimension(:), pointer :: spectrumNoArray
   integer(kind=int32), dimension(:), pointer :: spectrumOriginArray
! VNY 28.01.2004
   type(ColumnT) :: specNoListCol
   type(ColumnT) :: specX0Col
   type(ColumnT) :: specY0Col
   type(ColumnT) :: specX1Col
   type(ColumnT) :: specY1Col
   type(ColumnT) :: specIdCol
   !type(ColumnT) :: specOriginListCol
   integer(kind=int32), dimension(:), pointer :: specNoListArray
   real(kind=single), dimension(:), pointer :: specX0Array
   real(kind=single), dimension(:), pointer :: specY0Array
   real(kind=single), dimension(:), pointer :: specX1Array
   real(kind=single), dimension(:), pointer :: specY1Array
   integer(kind=int32), dimension(:), pointer :: specIdArray
   !integer(kind=int32), dimension(:), pointer :: specOriginListArray
!

   type(ColumnT) :: waveSourceOutCol0
   real(kind=Single), pointer :: waveSourceOutArray0(:)
!--- 
   real(kind=Single), pointer :: waveSourceOutArrayL(:) 
   real(kind=Single), pointer :: waveSourceOutArrayR(:) 

   ! rates file pointers
   real(kind=Single), pointer :: rateSourceOutArray0(:) !+
   real(kind=Single), pointer :: rateSourceErrOutArray0(:) !+
   real(kind=Single), pointer :: rateBackOutArray0(:) !+
   real(kind=Single), pointer :: rateBackErrOutArray0(:) !+
!
   real(kind=Single), pointer :: fluxOutArray0(:) !+
   real(kind=Single), pointer :: fluxErrOutArray0(:) !+
   real(kind=Single), pointer :: fluxOutArrayL(:) !+
   real(kind=Single), pointer :: fluxErrOutArrayL(:) !+
   real(kind=Single), pointer :: fluxOutArrayR(:) !+
   real(kind=Single), pointer :: fluxErrOutArrayR(:) !+
!

   real(kind=Single), pointer :: rateSourceOutArrayL(:) !+
   real(kind=Single), pointer :: rateSourceErrOutArrayL(:) !+
   real(kind=Single), pointer :: rateBackOutArrayL(:) !+
   real(kind=Single), pointer :: rateBackErrOutArrayL(:) !+
!   real(kind=Double), pointer :: timeSourceOutArrayL(:) !+
   real(kind=Single), pointer :: rateSourceOutArrayR(:) !+
   real(kind=Single), pointer :: rateSourceErrOutArrayR(:) !+
   real(kind=Single), pointer :: rateBackOutArrayR(:) !+
   real(kind=Single), pointer :: rateBackErrOutArrayR(:) !+
!   real(kind=Double), pointer :: timeSourceOutArrayR(:) !+
!
   real(kind=double) :: timeExposure !+
   integer(kind=int32) :: srcNo !+
   integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1 !+   ! Coordinates of the OSW origin
   integer(kind=int32) :: windowDX, windowDY !+
   logical :: binBpe !+
   real(kind=single),  dimension(:), pointer :: psf1Correction !+          ! psf1 correction
   real(kind=single),  dimension(:), pointer :: psf2Correction !+         ! psf2 correction
   real(kind=single) :: nullValue = -999.999  !+
   integer :: columnDim(1) =(/1/) !+ 
!   real(kind=double) :: lambda
   real(kind=double), dimension(:), pointer :: calLambda
   integer(kind=int32) :: lambdaSize
   real(kind=double) :: plateScale
   integer(kind=int32) :: binAx1                                     ! DPU x-axis binning factors.
   integer(kind=int32) :: binAx2                                     ! DPU y-axis binning factors.
   integer(kind=int32) :: binX, xbin                                       ! Bin x.
   integer(kind=int32) :: binY, ybin                                       ! Bin y.
   !integer(kind=int32) :: srcNumber                                ! Number of source in source list.
   real(kind=double) :: xs0,x1,x2,ys0,y1,y2
   real(kind=double) :: mx1,mx2,mx3,px1,px2,px3
!    real(kind=single) :: averAngle, averAngle2, sumAngle2
    real(kind=single) :: averImage
    real(kind=single) :: sum, cosPA, sinPA, black, white !, xMax, yMax
    integer(kind=int32) :: jup, jdown, i1, j1, limitUp, limitDown,limitLeft,limitRight, sourceDeformation
    integer(kind=int32) :: jRight, jLeft
    real(kind=single) :: level
    real(kind=single) :: aL0,aR0,aL1,aR1
!    integer(kind=int32) :: iMax
!    real(kind=single) :: sMax, dist, sinMax, cosMax , angleMax
    real(kind=single) :: averLeft, averRight, numLeft, numRight

!    real(kind=single) :: averBkgL, averBkgR, numBkgL, numBkgR 
!    real(kind=single) :: averSpec, numSpec, averStrip, numStrip

!    integer :: jcross, jbkg1, jbkg2, jspc1, jspc2, jyspc1, jyspc2 

!    real :: nLevel, wSpec, dSpec

   real(kind=double), dimension(4) :: param

!   real(kind=double)   :: gres
   integer(kind=int32) :: numparam, numpoints
!   real  :: varparam

!   integer(kind=int32) :: sizeOfData
!   real, dimension (:), pointer :: xinput
!   real, dimension (:), pointer :: yinput
!---
!   real  :: xtest , ytest ! this must be uncomm. for opt. extr. 
!   real, dimension(4) ::dydptest ! this must be uncomm. for opt. extr. 
!   real :: sigm ! this must be uncomm. for opt. extr. 
!---
!   real :: sum1, sum2, sumTest, sumGauss, sumWeight
   real(kind=double), dimension(:), pointer :: bkgVect, bkgVect2, bkgVect3
   real(kind=double), dimension(:), pointer :: bkgVectR, bkgVectL, specVect, smoothBkgVectR, smoothBkgVectL

   ! VNY 17.11.2003
   real(kind=double), dimension(:), pointer :: wavelengthArray, spectrumArray, bkgArray
   real(kind=double), dimension(:), pointer :: spectrumErrArray, bkgErrArray
   real(kind=double), dimension(:), pointer :: fluxArray, fluxErrArray

   real(kind=double), dimension(:), pointer :: countsScaleFactor
   real(kind=double) :: deltaX, deltaY, deltaCriterium, predictedSpectrumPosition, blueEdge, redEdge
   real(kind=double) :: deltaXY
   real(kind=double) :: brightness1, brightness2, brightness3, meanBr3
   integer :: ii,jj,jx, iy
   integer :: iblueEdge, iredEdge
   real(kind=double) :: sbRatio

   real(kind=double) :: shapeCriterium1, shapeCriterium2 ! criteria for assigning a specific shape to the source 
   real(kind=double) :: separation1, separation2 ! separation between the point source and its spectrum (in pixels)
   real(kind=double) :: source2spectrum, spectrumLength, checkWidth, checkLength 

   real(kind=double) :: nbr1, nbr2, nbr3
   integer :: jxR
   real(kind=double) :: mn1a,mn2a,mn3a,mn1b,mn2b,mn3b, mnab, critMin, critMax
   real(kind=double) :: mnx1a,mnx2a,mnx3a,mnx1b,mnx2b,mnx3b, mnxab, mxaver
   integer :: imn
   logical :: lateralSpectrumL, lateralSpectrumR
   real(kind=double) :: critMin0, critMax0
   real(kind=double) :: latShift, latWidth ! Parameters for the lateral spectra extraction in cases of saturated images
   integer :: blkNum
!---
   real(kind=double) :: bkgOffsetLeft, bkgWidthLeft, bkgOffsetRight, bkgWidthRight
   real(kind=double) :: spectrumHalfWidth !, imageSmoothBoxWidth
   integer :: spectrumSmoothLength
   real(kind=double) :: spectrumLengthParU, spectrumLengthParV
   real(kind=double) :: spectrumToleranceU, spectrumToleranceV, spectrumTolerance
   real(kind=double) :: src2spectrumU, src2spectrumV
   real(kind=double) :: binningInterval
!   integer :: nPixSpec, nPixBkgL, nPixBkgR
   integer :: extractionMode
   logical :: subtractBkg
   integer :: iyUp, iyDown, iyUp2
   real(kind=double) :: availableSpectrumFraction, availableSpectrumLength
   real(kind=double) :: xFirstOrder, yFirstOrder, xZeroOrder, yZeroOrder
! VNY 28.01.2004
   character(len=1024) :: spectraListFile !+
   integer(kind=int32) :: nAux
   real(kind=double) :: regionU1, regionU2, regionV1, regionV2, region1, region2
   character(len=1024) :: tempFileName
   logical :: omgsourceRegime
   integer :: verbFlag
   real(kind=double) :: maxValue2, meanValue2, yZero2
   real(kind=double) :: searchLength
   omgsourceRegime=.false.

   param(1)=2.0 ! The original values of the parameters
   param(2)=6.0
   param(3)=10.0
   param(4)=5.0
   numparam=4
   numpoints=21

   critMin0=0.5
   critMax0=0.97
   ! get parameters
   grismFile = stringParameter("set")
   sourceListFile = stringParameter("sourcelistset")
   !srcNumber = intParameter("srcnumber")
   outFile = stringParameter("outset")
   posPoint=index(trim(outFile), 'SPECTR', .true.)
   spectraListFile=outFile
   if (posPoint .gt. 0) then
      spectraListFile(posPoint:posPoint+5)="SPECLI"
   else
      spectraListFile="pSPECLI.fit"
   endif
   posPoint=index(trim(outFile), '/p', .true.)
   outFileShortenName=outFile(posPoint+1:1024)
   write(spectrumFileComment, '(a, a)') "spectrumFile=", &
        trim(outFileShortenName)
 
   bkgOffsetLeft = realParameter("bkgoffsetleft")
   bkgWidthLeft = realParameter("bkgwidthleft")
   bkgOffsetRight = realParameter("bkgoffsetright")
   bkgWidthRight = realParameter("bkgwidthright")
   extractionMode= intParameter("extractionmode")

   spectrumLengthParU = 1100.0
   spectrumLengthParV = 600.0
! VNY 04.02.2004
!   spectrumToleranceU= 300.0
!   spectrumToleranceV=150.0

   src2spectrumU = 900.0 ! 1100.0
!   src2spectrumU = 1000.0
! VNY 05.02.2004
!   src2spectrumU = 675.0
   spectrumToleranceU= 600.0
   spectrumToleranceV=300.0

!
   src2spectrumV = 785.0
!
   spectrumHalfWidth = realParameter("spectrumhalfwidth")

   spectrumSmoothLength = intParameter("spectrumsmoothlength")

   bkgOffsetLeft=abs(bkgOffsetLeft)
   bkgWidthLeft=abs(bkgWidthLeft)
   bkgOffsetRight=abs(bkgOffsetRight)
   bkgWidthRight=abs(bkgWidthRight)
   spectrumHalfWidth=abs(spectrumHalfWidth)


   grismSetH = dataSet(TRIM(grismFile), read)
   sourceListSetH=dataSet(TRIM(sourceListFile), modify)


   !---Protection against the absence of the SRCLIST block
   if (hasBlock(sourceListSetH, "SRCLIST")) then

      !**************************************************************
      ! Read in the detection limit count rate from the PRIMARY header
      !**************************************************************
      detectionLimitCountRate = real(real32Attribute(sourceListSetH, "CRLIMIT"))


      ! Get a handle on the source list binary table extension.
      inTableH = table(sourceListSetH, "SRCLIST")

      tempFileName = " "
      if (hasAttribute(inTableH, "FILENAME")) then
         tempFileName= stringAttribute(inTableH, "FILENAME")
         posPoint=index(trim(tempFileName), 'OMGSOURCE')
         if (posPoint .gt. 0) omgsourceRegime=.true.
      else
         call setAttribute(inTableH, "FILENAME", trim(outFileShortenName), "Name of the spectrum file")
      endif
      if (omgsourceRegime) then
         write(errstr,*) "Working in the omgsource task regime"
         call message(AppMsg, 2)
      endif

      ! Read in the source list table extension keywords.
      dataMode = stringAttribute(sourceListSetH, "DATAMODE")
      filter = stringAttribute(sourceListSetH, "FILTER")
      exposure = real32Attribute(sourceListSetH, "EXPOSURE") 
      binBpe= booleanAttribute(sourceListSetH, "BINBPE")
      frameTime=real64Attribute(sourceListSetH, "FRAMTIME")
      frameTime=frameTime/1000.0 ! convert frametaime into seconds
      deadFrac=real64Attribute(sourceListSetH, "DEADFRAC")

      binAx1 = int32Attribute(sourceListSetH, "BINAX1")
      binAx2 = int32Attribute(sourceListSetH, "BINAX2")
!
      tmExpId = int32Attribute(grismSetH, "TMEXP_ID")
      frmTime = int32Attribute(grismSetH, "FRMTIME")
      dataType = stringAttribute(grismSetH, "DATATYPE")
      obsId = stringAttribute(grismSetH, "OBS_ID")
      expId = stringAttribute(grismSetH, "EXP_ID")
      dateObs = stringAttribute(grismSetH, "DATE-OBS")
      dateEnd = stringAttribute(grismSetH, "DATE-END")

      ! Calculate image binning.          
      binX = binAx1
      if (binX .lt. 1.0) binX = 1.0
      binY = binAx2
      if (binY .lt. 1.0) binY = 1.0
      
      if (binBPE) then
         binX = 2.0
         binY = 2.0
      end if

      xbin=binX
      ybin=binY
      !Criteria for extraction 
      if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (visual grism)
         
         ! VNY 02.03.2004
         !shapeCriterium1=80.0 ! shape criteria for assigning the shape to the source
         shapeCriterium1=30.0
!         shapeCriterium2=50.0 
         shapeCriterium2=30.0 

         separation1=290.0
         separation2=450.0
         checkWidth=10.0
         checkLength=100.0
         ! Checking the specrum regions for both filters
         source2spectrum=src2spectrumV ! 800.0
         spectrumLength=spectrumLengthParV !700.0 !400.0
         spectrumTolerance=spectrumToleranceV
         if (binY .gt. 1.0) then 
            source2spectrum=source2spectrum/binY
            spectrumLength=spectrumLength/binY
            spectrumTolerance=spectrumTolerance/binY
            shapeCriterium1=shapeCriterium1/binY
            shapeCriterium2=shapeCriterium2/binY
            checkLength=checkLength/binY
         endif
         if (binX .gt. 1.0) then
            checkWidth=checkWidth/binX
         endif
      endif
!---
      if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
         shapeCriterium1=200.0 ! shape criteria for assigning the shape to the source
         shapeCriterium2=100.0 
         separation1=290.0
         separation2=450.0
         checkWidth=15.0 ! criterium for the x-separation of the zero and first orders
         checkLength=200.0
         ! Checking the same specrum regions for both filters
         source2spectrum=src2spectrumU ! 800.0 ! 1100.0 
         spectrumLength=spectrumLengthParU ! 700.0 !1000.0
         spectrumTolerance=spectrumToleranceU
         if (binY .gt. 1.0) then 
            source2spectrum=source2spectrum/binY
            spectrumLength=spectrumLength/binY
            shapeCriterium1=shapeCriterium1/binY
            shapeCriterium2=shapeCriterium2/binY
            spectrumTolerance=spectrumTolerance/binY
            checkLength=checkLength/binY
         endif
         if (binX .gt. 1.0) then
            checkWidth=checkWidth/binX
         endif
      endif

! Checking different shifts (these shifts are for the lateral wings spectra)
      latWidth=10.0
      latShift=20.0
      if (binX .gt. 1.0) then
         latShift=latShift/binX
         latWidth=latWidth/binX
         checkWidth=checkWidth/binX
      endif
!---

      !********************
      ! Get the ccd mode.
      !********************
      if (trim(dataMode) .eq. "IMAGING" .or. trim(dataMode) .eq. "GRISM") then
         ccdMode = IMAGE
      else
         ccdMode = FAST
      end if

      !**************************************
      ! Get the filter Id integer value
      !**************************************
      filterIdInt = filterIdToInt(filter)
      !*******************
      ! Set the CAL state
      !********************
      call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)
      !************************************* 
      ! Get the Plate scale (arcsecs/pixel)
      !*************************************
! VNY 12.11.2003 - removing plateScale, since the accuracy doesn't matter
!      call CAL_omGetPlateScale(lambda, plateScale)
      plateScale=0.476513
! VNY 05.02.2004
    allocate(calLambda(0:2048))
    calLambda=0.0
    call CAL_omGetGrismWavelength(calLambda)
    lambdaSize=size(calLambda)
    regionU1=-1.0
    regionU2=-1.0
    regionV1=-1.0
    regionV2=-1.0
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       do iSource=1, lambdaSize-1
          if (calLambda(iSource) .ge. 1800.0 .and. calLambda(iSource-1) .lt. 1800.0) then
             regionU1=iSource
          endif
          if (calLambda(iSource) .ge. 3600.0 .and. calLambda(iSource-1) .lt. 3600.0) then
             regionU2=iSource
          endif
       enddo
       if (binY .gt. 1.0) then 
          regionU1=regionU1/binY
          regionU2=regionU2/binY
       endif
    endif

    if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (Vis grism)
       do iSource=1, lambdaSize-1
          if (calLambda(iSource) .ge. 3000.0 .and. calLambda(iSource-1) .lt. 3000.0) then
             regionV1=iSource
          endif
          if (calLambda(iSource) .ge. 6000.0 .and. calLambda(iSource-1) .lt. 6000.0) then
             regionV2=iSource
          endif
       enddo
       if (binY .gt. 1.0) then 
          regionV1=regionV1/binY
          regionV2=regionV2/binY
       endif
    endif

    if (filter .eq. "Grism2  " .or. filter .eq. "GRISM2  ")  then ! (Vis grism)
       region1=regionV1
       region2=regionV2
    endif
    if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
       region1=regionU1
       region2=regionU2
    endif
    
      !******************************************************************
      ! Get the number of rows in the source list binary table extension.
      !******************************************************************
      nSources = numberOfRows(inTableH)
      !VNY 01.02.2004 
      nSpectraSources=nSources

      !--- Setting a limit to the input rows to 1000 sources
      if (nSources .gt. 0 .and. nSources .lt. 1000) then
         allocate(limitLeftMax(0:nSources-1))
         allocate(limitRightMax(0:nSources-1))
         limitLeftMax=0.0
         limitRightMax=0.0
         !  Allocate space for the source relation and shape flags
         allocate(sourceShape(0:nSources-1))
         allocate(sourceRelation(0:nSources-1))
!
         allocate(x0aux(0:nSources-1))
         allocate(x1aux(0:nSources-1))
         allocate(y0aux(0:nSources-1))
         allocate(y1aux(0:nSources-1))
!
         sourceShape=-1
         sourceRelation=-1
         x0aux=-1.0
         x1aux=-1.0
         y1aux=-1.0
         y0aux=-1.0
!
         ! Get handles and pointers on the source list binary table extension colums.
         inColumnFlag=1

         if (hasColumn(inTableH, "RATE")) then
            inColumnH(1) = column(inTableH, "RATE", modify)
            rate => real32Data(inColumnH(1))
         else
            inColumnFlag(1)=0
            allocate(rate(0:nSources-1))
         endif

         if (hasColumn(inTableH, "RATE_ERR")) then
            inColumnH(2) = column(inTableH, "RATE_ERR", read)
            rateErr => real32Data(inColumnH(2))
         else
            inColumnFlag(2)=0
            allocate(rateErr(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAG")) then
            inColumnH(3) = column(inTableH, "MAG", modify)
            mag => real32Data(inColumnH(3))
         else
            inColumnFlag(3)=0
            allocate(mag(0:nSources-1))
         endif

         if (hasColumn(inTableH, "MAGERR")) then
            inColumnH(4) = column(inTableH, "MAGERR", modify)
            magErr => real32Data(inColumnH(4))
         else
            inColumnFlag(4)=0
            allocate(magErr(0:nSources-1))
         endif

         if (hasColumn(inTableH, "CORR_BACKGROUND_RATE")) then
            inColumnH(17)=column(inTableH, "CORR_BACKGROUND_RATE", modify)
         else
            inColumnFlag(17)=0
         endif

         if (hasColumn(inTableH, "EFLAG")) then
            inColumnH(5) = column(inTableH, "EFLAG", modify)
            extended => int8Data(inColumnH(5))
         else
            inColumnFlag(5)=0
            allocate(extended(0:nSources-1))
         endif

         if (hasColumn(inTableH, "BACKGROUND_RATE")) then
            inColumnH(6) = column(inTableH, "BACKGROUND_RATE", read)
            backGroundRate => real32Data(inColumnH(6))
         else
            inColumnFlag(6)=0
            allocate(backGroundRate(0:nSources-1))
         endif

         inColumnH(7) = column(inTableH, "XPOS", read)
         sourceXCoord0 => real32Data(inColumnH(7))

         allocate(sourceXCoord(0:nSources-1))
         sourceXCoord=sourceXCoord0

         inColumnH(8) = column(inTableH, "YPOS", read)
         sourceYCoord0 => real32Data(inColumnH(8))

         allocate(sourceYCoord(0:nSources-1))
         sourceYCoord=sourceYCoord0

         if (hasColumn(inTableH, "APERTURE")) then
            inColumnH(9) = column(inTableH, "APERTURE", modify)
            sourceApertureRadius => real32Data(inColumnH(9))
         else
            inColumnFlag(9)=0
           allocate(sourceApertureRadius(0:nSources-1))
         endif

         inColumnH(18) = column(inTableH, "FWHM_MAJ", read)
         sourceFWHMaj => real32Data(inColumnH(18))

         allocate(sourceFWHMaj2(0:nSources-1))
         sourceFWHMaj2=sourceFWHMaj

         inColumnH(19) = column(inTableH, "FWHM_MIN", read)
         sourceFWHMin => real32Data(inColumnH(19))

         allocate(sourceFWHMin2(0:nSources-1))
         sourceFWHMin2=sourceFWHMin

         inColumnH(20) = column(inTableH, "PA", read)
         sourcePA => real32Data(inColumnH(20))

         !write(*,*) "********************************************************************"
         !do iSource=0, nSources-1
         !   sourceWx=sourceFWHMin2(iSource)*3.0/(2.354*binX*plateScale)
         !   sourceWy=sourceFWHMaj2(iSource)*3.0/(2.354*binY*plateScale)
         !   write(*,'(i3,a,f6.1,a,f6.1,a,f6.1,a, f6.1,a,f6.1,a,f6.1,a,f6.1)') &
         !        iSource, " X=", sourceXCoord(iSource), & 
         !        " Y=", sourceYCoord(iSource), & 
         !        " Maj=", sourceFWHMaj2(iSource), " Min=", sourceFWHMin2(iSource), &
         !        " Wy=", sourceWy, " Wx=", sourceWx, &
         !        " PA=", sourcePA(iSource)
         !   sourceFWHMaj2(iSource)=sourceWy
         !   sourceFWHMin2(iSource)=sourceWx
         !enddo
         !write(*,*) "******************************************************************"

         ! the image contains the information on the science window
         inArrayH = array(grismSetH, "PRIMARY", read)
         exposureTime = real32Attribute(grismSetH, "EXPOSURE") ! Image exposure time
         oswImage => real32array2data(inArrayH)
         dim = shape(oswImage)
         nx = dim(1)
         ny = dim(2) 
         ! Assign flags for the extended and point-like sources
         do iSource=0, nSources-1
            sourceShape(iSource)=0
            if (sourceFWHMaj2(iSource) .gt. shapeCriterium1 .and. & 
                 sourceFWHMin2(iSource) .lt. shapeCriterium2 .and. &
                 ! VNY 12.02.2004
                 abs(sourcePA(iSource)) .gt. 82.0) & 
                 sourceShape(iSource)=1
         enddo
         !   Search for the contaminating sources (point-like sources at the vicinity of a primary source).

         deltaCriterium=24.0/xBin!12.0 ! 12-pixel minimal distance between the sources 
         do iSource=0, nSources-1
            ! Do it only for the point-like sources
            if (sourceShape(iSource) .eq. 0) then
               do i=0, nSources-1
                  if (i .ne. iSource) then ! calculate distance between the sources 
                     deltaX=abs(sourceXCoord(i)-sourceXCoord(iSource))
                     deltaY=abs(sourceYCoord(i)-sourceYCoord(iSource))
                     deltaXY=sqrt(deltaX*deltaX+deltaY*deltaY)
                     if (deltaXY .lt. deltaCriterium .and. .not. omgsourceRegime) then ! assign -1 to the source shape 
                        ! compare brightnesses of these sources within 3x3 pixels
                        brightness1=0.0 ! Brightness of the first source
                        do ii=-2,2
                           iy=sourceYCoord(iSource)+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceXCoord(iSource)+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness1=brightness1+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2

                        brightness2=0.0 ! Brightness of the second source
                        do ii=-2,2
                           iy=sourceYCoord(i)+ii
                           if (iy .ge. 0 .and. iy .lt. ny) then
                              do jj=-2,2
                                 jx=sourceXCoord(i)+jj
                                 if (jx .ge. 0 .and. jx .lt. nx) then
                                    brightness2=brightness2+oswImage(jx,iy)
                                 endif
                              enddo
                           endif
                        enddo ! ii=-2,2

                        if (brightness1 .ge. brightness2) then ! eliminating the second source
                           sourceShape(i)=-1
                        else ! eliminate the first source
                           sourceShape(iSource)=-1
                        endif
                     endif ! deltaX .lt. ...
                     ! VNY 04.02.04 Check if the source fall inside of a spectrum
                     if (sourceShape(i) .eq. 1) then
!                        if (deltaX .le. sourceFWHMin(i) .and. deltaY .le. sourceFWHMaj(i)) then
                        if (deltaX .le. sourceFWHMin2(i) .and. deltaY .le. sourceFWHMaj2(i) &
                             .and. .not. omgsourceRegime) then
                           sourceShape(iSource)=-2
                        endif ! (deltaX .le. sourceFWHMin
                     endif
                  endif ! i .ne. iSource
               enddo
            endif ! sourceShape
         enddo ! iSource-0, nSources-1
         !--- Analysis of the source interconnections
         do iSource=0, nSources-1
            if (sourceRelation(iSource) .lt. 0) then
               !  should analyse distances only for point-like sources
               if (sourceShape(iSource) .eq. 0) then
                  !
                  !
                  do i=0, nSources-1
                     if (sourceShape(i) .eq. 1) then 
                        if (i .ne. iSource .and. sourceRelation(i) .lt. 0) then
                           deltaX=abs(sourceXCoord(i)-sourceXCoord(iSource))
                           if (omgsourceRegime) deltaX=0.0
                           deltaY=(sourceYCoord(i)-sourceYCoord(iSource))
                           predictedSpectrumPosition=sourceYCoord(iSource)+region2
                           blueEdge=sourceYCoord(i)-spectrumTolerance/2.0
                           redEdge=sourceYCoord(i)+spectrumTolerance/2.0
                           if (omgsourceRegime) predictedSpectrumPosition=(blueEdge+redEdge)/2.0
                           if (deltaX .lt. checkWidth &
                                .and. predictedSpectrumPosition .gt. blueEdge &
                                .and. predictedSpectrumPosition .lt. redEdge) then
                              if (sourceRelation(iSource) .lt. 0) then
                                 sourceRelation(i)=iSource
                                 sourcerelation(iSource)=i
                              endif
                           endif
                        endif
                     endif
                  enddo 
               endif ! sourceShape .eq. 0
            endif ! sourceRelation .lt. 0
         enddo
!--
!  Checking if there are some undetected extended sources   
!                related to some of the rest of the point-like sources 
!                by using source2spectrum and spectrumLength parameters
         do iSource=0, nSources-1
            if (sourceShape(iSource) .eq. 0 .and. sourceRelation(iSource) .lt. 0) then ! Check only the point-like sources
            ! calculate flux within the expected area of    
               brightness1=0.0
               nbr1=0.0
               brightness2=0.0
               nbr2=0.0
               brightness3=0.0 ! the maximal value of the strip
               meanBr3=0.0
               nbr3=0.0
               iy=sourceYCoord(iSource)+source2spectrum
               if (iy .ge. 0 .and. iy .lt. ny) then                                    
                  do ii=iy-int(spectrumLength/2), iy+int(spectrumLength/2)
                     if (ii .ge. 0 .and. ii .lt. ny) then
                        brightness3=0.0
                        do jj=-int(checkWidth),int(checkWidth)
                           jx=sourceXCoord(iSource)+jj
                           jxR=jx+checkWidth*3+1
                           if (jx .ge. 0 .and. jx .lt. nx) then
                              brightness1=brightness1+oswImage(jx,ii)
                              nbr1=nbr1+1.0
                              ! compute the maximal value on the strip
                              if (oswImage(jx,ii) .gt. brightness3) brightness3=oswImage(jx,ii)
                           endif
                           if (jxR .ge. 0 .and. jxR .lt. nx) then
                              brightness2=brightness2+oswImage(jxR,ii)
                              nbr2=nbr2+1.0
                           endif
                        enddo
                        meanBr3=meanBr3+brightness3
                        nbr3=nbr3+1.0
                     endif
                  enddo
                  if (nbr1 .gt. 0.0) brightness1=brightness1/nbr1
                  if (nbr2 .gt. 0.0) brightness2=brightness2/nbr2
                  if (nbr3 .gt. 0.0) meanBr3=meanBr3/nbr3
               endif
            endif ! sourceShape .eq. 0
         enddo ! iSource=0, nSource-1

         if (hasColumn(inTableH, "RA")) then
            inColumnH(10) = column(inTableH, "RA", modify)
            sourceRA => real64Data(inColumnH(10))
         else
            inColumnFlag(10)=0
            allocate(sourceRA(0:nSources-1))
         endif

         if (hasColumn(inTableH, "DEC")) then
            inColumnH(11) = column(inTableH, "DEC", modify)
            sourceDEC => real64Data(inColumnH(11))
         else
            inColumnFlag(11)=0
            allocate(sourceDEC(0:nSources-1))
         endif

         if (hasColumn(inTableH, "LII")) then
            inColumnH(12) = column(inTableH, "LII", modify)
            sourceLII => real64Data(inColumnH(12))
         else
            inColumnFlag(12)=0
            allocate(sourceLII(0:nSources-1))
         endif

         if (hasColumn(inTableH, "BII")) then
            inColumnH(13) = column(inTableH, "BII", modify)
            sourceBII => real64Data(inColumnH(13))
         else
            inColumnFlag(13)=0
            allocate(sourceBII(0:nSources-1))
         endif

         if (hasColumn(inTableH, "QFLAG")) then
            inColumnH(14) = column(inTableH, "QFLAG", read)
            sourceQF => int8Data(inColumnH(14))
         else
            inColumnFlag(14)=0
            allocate(sourceQF(0:nSources-1))
         endif

         if (hasColumn(inTableH, "CFLAG")) then
            inColumnH(15) = column(inTableH, "CFLAG", read)
            sourceCF => int8Data(inColumnH(14))
         else
            inColumnFlag(15)=0
            allocate(sourceCF(0:nSources-1))
         endif

         if (hasColumn(inTableH, "PSF1_CORR")) then
            inColumnH(16) = column(inTableH, "PSF1_CORR", modify)
            psf1Correction => real32Data(inColumnH(16))
         else
            inColumnFlag(16)=0
            allocate(psf1Correction(0:nSources-1))
            psf1Correction = nullValue
         endif
         allocate(psf2Correction(0:nSources-1))    
         psf2Correction = nullValue

         ! Add columns with the spectrum identification number to the (input) source list table 
         if (.not. hasColumn(inTableH, "SPECTR_ID")) then
            spectrumNoCol = addColumn(inTableH, "SPECTR_ID", integer32, "", columnDim, "spectrum identifier")
         else
            spectrumNoCol = column(inTableH, "SPECTR_ID", modify)
         endif
         spectrumNoArray => int32data(spectrumNoCol)
         call setNullValue(spectrumNoCol, 999999 )
 
         do iSource=0, nSources-1
            call setToNull(spectrumNoCol, iSource)
         enddo

         ! The column with the spectrum origin (the number of the corresponding point-like source)
         if (.not. hasColumn(inTableH, "ORIGIN")) then
            spectrumOriginCol = addColumn(inTableH, "ORIGIN", integer32, "", columnDim, "spectrum origin number ")
         else
            spectrumOriginCol = column(inTableH, "ORIGIN", modify)
         endif
         spectrumOriginArray => int32data(spectrumOriginCol)
         call setNullValue(spectrumOriginCol, 999999 )
         do iSource=0, nSources-1
            call setToNull(spectrumOriginCol, iSource)
         enddo

         !-(origin of the OSW)
         windowX0 = int16Attribute(sourceListSetH, "WINDOWX0")
         windowY0 = int16Attribute(sourceListSetH, "WINDOWY0")
         windowDX = int16Attribute(sourceListSetH, "WINDOWDX")
         windowDY = int16Attribute(sourceListSetH, "WINDOWDY")
         windowX1=windowX0+windowDX-1
         windowY1=windowY0+windowDY-1

         allocate(bkgVect(0:ny-1))
         allocate(bkgVect2(0:ny-1))
         allocate(bkgVect3(0:ny-1))
         allocate(bkgVectR(0:ny-1))
         allocate(bkgVectL(0:ny-1))
         allocate(specVect(0:ny-1))
         allocate(smoothBkgVectR(0:ny-1))
         allocate(smoothBkgVectL(0:ny-1))
         allocate(countsScaleFactor(0:ny-1))

         ! VNY 17.11.2003
         ysize=ny
         xsize=nx

         allocate(wavelengthArray(0:ny-1))
         allocate(spectrumArray(0:ny-1))
         allocate(spectrumErrArray(0:ny-1))
         allocate(bkgArray(0:ny-1))
         allocate(bkgErrArray(0:ny-1))
         allocate(fluxArray(0:ny-1))
         allocate(fluxErrArray(0:ny-1))

         !***********************************************************
         ! Compute the average value of the image
         !***********************************************************
         black = minval(oswImage)
         white = maxval(oswImage)
         averImage=0.0
         sum=0.0
         do i=0,nx-1
            do j=0,ny-1
               averImage=averImage+oswImage(i,j)
               sum=sum+1.0
            enddo
         enddo

         if (sum .gt. 0.0) then
            averImage=averImage/sum
         endif
!**********************************************************************
         !  Create a single file containing spectra for all the sources            
!**********************************************************************   
         outFile4=outFile
         outSet0H = dataSet(outFile4, create)
!         call copyAllAttributes(attributable(sourceListSetH), attributable(outSet0H))
         call setAttribute(outSet0H, "CONTENT", "OM GRISM SPECTRA", &
              "File contains OM grsim spectrum/spectra")
         call setAttribute(outSet0H, "HDUCLASS", "OGIP    ", &
              "Format conforms to OGIP/GSFC conventions")
         call setAttribute(outSet0H, "HDUCLAS1", "SPECTRUM", &
              "File contains spectral data")
         call setAttribute(outSet0H, "HDUVERS", "1.2.0    ", &
              "The version number of the format")
         call setAttribute(outSet0H, "TELESCOP", "XMM     ", &
              "XMM mission")
         call setAttribute(outSet0H, "INSTRUME", "OM      ", &
              "Instrument name")

!      ! Read in the source list table extension keywords.
!      dataMode = stringAttribute(sourceListSetH, "DATAMODE")
!      filter = stringAttribute(sourceListSetH, "FILTER")
!      exposure = real32Attribute(sourceListSetH, "EXPOSURE") 
!      binBpe= booleanAttribute(sourceListSetH, "BINBPE")
!      frameTime=real64Attribute(sourceListSetH, "FRAMTIME")
!      frameTime=frameTime/1000.0 ! convert frametaime into seconds
!      deadFrac=real64Attribute(sourceListSetH, "DEADFRAC")

!      binAx1 = int32Attribute(sourceListSetH, "BINAX1")
!      binAx2 = int32Attribute(sourceListSetH, "BINAX2")

         call setAttribute(outSet0H, "DATAMODE", dataMode, &
              "Data mode")
         call setAttribute(outSet0H, "FILTER", filter, &
              "Filter used for obtaining the data")
         call setAttribute(outSet0H, "FRAMTIME", frameTime, &
              "Frame time")
         call setAttribute(outSet0H, "DEADFRAC", deadFrac, &
              "Frame dead fraction")
         call setAttribute(outSet0h, "TMEXP_ID", tmExpId, &
              "Exposure Id from telemetry")
         call setAttribute(outSet0H, "FRMTIME", frmTime, &
              "Frametime from telemetry")
         call setAttribute(outSet0H, "DATATYPE", dataType, &
              "Data type")
         call setAttribute(outSet0H, "OBS_ID", obsId, &
              "Observation identifier")
         call setAttribute(outSet0H, "EXP_ID", expId, &
              "Exposure identifier")
         call setAttribute(outSet0H, "DATE-OBS", dateObs, &
              "Observation start time")
         call setAttribute(outSet0H, "DATE-END", dateEnd, &
              "Observation end time")
!***********************************************************************


         !*********************************************************
         ! Processing source by source 
         !*********************************************************
         do iSource=0, nSources-1
            lateralSpectrumL = .false.
            lateralSpectrumR = .false.
            if (extended(iSource) .eq. 0.0) then
               extend = .false.
            else
               extend = .true.
            end if

            xZeroOrder=0.0
            yZeroOrder=0.0
            xFirstOrder=0.0
            yFirstOrder=0.0

            ! Set the EXTENDED parameter to 2 in it is a genuine spectrum
            if (sourceShape(iSource) .eq. 1) then
               extended(iSource)=2
            endif
            !  Restrict the analysis only for the extended sources related to their zero-orders
            if (sourceShape(iSource) .eq. 1 .and. sourceRelation(iSource) .ge. 0) then

               !  Calculation of the spectrum edge coordinates
               sourceWx=sourceFWHMin2(iSource)
               sourceWy=sourceFWHMaj2(iSource)

               xs0=sourceXCoord(iSource)
               x1=xs0
               x2=x1
               xFirstOrder=xs0
               ys0=sourceYCoord(iSource)
               y1=ys0
               y2=y1
               yFirstOrder=ys0
! VNY 06.02.2004
               blueEdge=sourceYCoord(iSource)+region1
               redEdge=sourceYCoord(iSource)+region2
               iblueEdge=int(blueEdge)
               iredEdge=int(redEdge)
!              
               if (sourcePA(iSource) .ne. 0.) then
                  x1=x1+0.5*sourceWy/tan(-sourcePA(iSource)*PI/180.0_double)
                  x2=x2-0.5*sourceWy/tan(-sourcePA(iSource)*PI/180.0_double)
               else
                  x1=x1+0.5*sourceWy
                  x2=x2-0.5*sourceWy
               endif

               y1=y1-0.25*spectrumLength
               y2=y2+0.25*spectrumLength

               !  Check the spectrum profile for three strips
               j=int(y2)
               mx1=0.0
               px1=xs0

               if (j .ge. 0 .and. j .lt. ny) then
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then
                        if (oswImage(i1,j) .gt. mx1) then
                           mx1=oswImage(i1,j)
                           px1=i1
                        endif
                     endif
                  enddo
                  !...................1
                  ! Test the left part of the strip for the presence of a minimum
                  mn1a=mx1
                  imn=int(px1)-1
                  do i=int(xs0)-20,int(px1)-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn1a) then
                           mn1a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the left part of the strip for the second maximum
                  mnx1a=mn1a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx1a) then
                           mnx1a=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the minimum
                  mn1b=mx1
                  imn=int(px1)+1
                  do i=int(px1)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn1b) then
                           mn1b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the right part of the strip for the second maximum
                  mnx1b=mn1b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx1b) then
                           mnx1b=oswImage(i,j)
                        endif
                     endif
                  enddo
                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................1
                  ! VNY test 07.01.2004 

               endif
               j=int(ys0)
               mx2=0.0
               px2=xs0

               if (j .ge. 0 .and. j .lt. ny) then
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then
                        if (oswImage(i1,j) .gt. mx2) then
                           mx2=oswImage(i1,j)
                           px2=i1
                        endif
                     endif
                  enddo
                  !...................2
                  ! Test the left part of the strip for the minimum
                  mn2a=mx2
                  imn=int(px2)-1
                  do i=int(xs0)-20,int(px2)-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn2a) then
                           mn2a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the left part of the strip for the second maximum
                  mnx2a=mn2a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx2a) then
                           mnx2a=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the minimum
                  mn2b=mx2
                  imn=int(px2)+1
                  do i=int(px2)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn2b) then
                           mn2b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the right part of the strip for the second maximum
                  mnx2b=mn2b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx2b) then
                           mnx2b=oswImage(i,j)
                        endif
                     endif
                  enddo
                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................2
               endif

               j=int(y1)

               mx3=0.0
               px3=xs0
               if (j .ge. 0 .and. j .lt. ny) then 
                  do i=-20,20
                     i1=int(xs0)+i
                     if (i1 .ge. 0 .and. i1 .lt. nx) then

                        if (oswImage(i1,j) .gt. mx3) then
                           mx3=oswImage(i1,j)
                           px3=i1
                        endif
                     endif
                  enddo
                  !...................3
                  ! Test the left part of the strip for the minimum
                  mn3a=mx3
                  imn=int(px3)-1
                  do i=int(xs0)-20,int(px3)-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn3a) then
                           mn3a=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the left part of the strip for the second maximum
                  mnx3a=mn3a
                  do i=int(xs0)-20,imn-1
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx3a) then
                           mnx3a=oswImage(i,j)
                        endif
                     endif
                  enddo

                  ! Test the right part of the strip for the minimum
                  mn3b=mx3
                  imn=int(px3)+1
                  do i=int(px3)+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .lt. mn3b) then
                           mn3b=oswImage(i,j)
                           imn=i
                        endif
                     endif
                  enddo
                  ! Test the right part of the strip for the second maximum
                  mnx3b=mn3b
                  do i=imn+1,int(xs0)+20
                     if (i .ge. 0 .and. i .lt. nx) then
                        if (oswImage(i,j) .gt. mnx3b) then
                           mnx3b=oswImage(i,j)
                        endif
                     endif
                  enddo
                  ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
                  !.......................3
               endif

               ! Averaging secondary minima and maxima
               mnab=(mn1a+mn1b+mn2a+mn2b+mn3a+mn3b)/6.0
               mnxab=(mnx1a+mnx1b+mnx2a+mnx2b+mnx3a+mnx3b)/6.0

               ! mn1a and mn1b are the minima; mnx1a and mnx1b are the secondary maxima
               !.......................1
               

               mxaver=(mx1+mx2+mx3)/3.0

               critMax=(mxaver-mnab)
               critMin=(mnxab-mnab)
               if (mxaver .gt. 0.) critMax=critMax/mxaver
               if (mnxab .gt. 0.)  critMin=critMin/mnxab

               xs0=(px1+px2+px3)/3.0

               sourceXCoord(iSource)=xs0

               if (critMin .gt. critMin0 .and. critMax .gt. critMax0) then
                  ! Check if it is possible to build the lateral spectra within the given window
                  if (int(xs0-latShift-latWidth/2.0) .gt. 0) then
                     lateralSpectrumL= .true.
                  endif
                  if (int(xs0+latShift+latWidth/2.0) .lt. nx) then
                     lateralSpectrumR= .true.
                  endif
                  write(errstr,*) "The spectrum No. ", iSource, " is corrupted by Mod-8 noise"
                  call message(AppMsg, 2)

               endif

               !***********************************************************
               ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
               !***********************************************************
               !call readModes(grismSetH, x0, y0, xsize, ysize)  

               i=int(sourceXCoord(iSource)+0.5)
               j=int(sourceYCoord(iSource)+0.5)
               cosPA=cos(sourcePA(iSource)*degreesToRadians)
               sinPA=sin(sourcePA(iSource)*degreesToRadians)

               ! Find limits of the current source upwards and downwards
               sourceDeformation=0.0
               limitUp=0
               limitDown=0  
               level=averImage!*1.05

               do jup=0,500
                  j1=j+jup
                  i1=int(sourceXCoord(iSource)+0.5)
                  if (j1 .lt. 0 .or. j1 .ge. ny .or. i1 .ge. nx .or. i1 .lt. 0) exit
                  if (oswImage(i1,j1) .lt. level) exit
               enddo

               limitUp=jup
               do jdown=0,500
                  j1=j-jdown
                  i1=int(sourceXCoord(iSource)+0.5)
                  if (j1 .lt. 0 .or. j1 .ge. ny .or. i1 .ge. nx .or. i1 .lt. 0) exit
                  if (oswImage(i1,j1) .lt. level) exit
               enddo
               limitDown=jdown

               ! Calculate limits left and right
               limitLeft=0.0
               limitRight=0.0
               if (cosPA .ne. 0.0) then
                  !            if (cosPA .ge. 0.5) then
                  !
                  do jRight=0,100
                     i1=i+jRight
                     j1=int(sourceYCoord(iSource)+0.5)
                     if (i1 .lt. 0 .or. i1 .ge. nx .or. j1 .ge. ny .or. j1 .lt. 0) exit
                     if (oswImage(i1,j1) .lt. level) exit
                  enddo
                  limitRight=jRight 
                  do jLeft=0,100
                     i1=i-jLeft
                     j1=int(sourceYCoord(iSource)+0.5)
                     if (i1 .lt. 0 .or. i1 .ge. nx .or. j1 .ge. ny .or. j1 .lt. 0) exit
                     if (oswImage(i1,j1) .lt. level) exit
                  enddo
                  limitLeft=jLeft
               endif

               if (limitLeft .gt. limitLeftMax(iSource)) limitLeftMax(iSource)=limitLeft
               if (limitRight .gt. limitRightMax(iSource)) limitRightMax(iSource)=limitRight
               
               if ((limitUp+limitDown+1.0) .gt. (limitRight+limitLeft+1.0)) then
                  sourceDeformation=(limitUp+limitDown+1.0)/(limitRight+limitLeft+1.0)
               endif

               sourceArea=0.0
               srcNo=iSource
               timeExposure = exposure
               !--- Create output only for the sources with the proper shape and relation
               !            if (sourceDeformation .gt. 10.0) then
               ! Create the output file only for the extended sources
               !  Name the output block using the source number (iSource)
               posPoint=scan(trim(outFile), '.', .true.)
               ! VNY 10.12.2003
               i=sourceRelation(iSource)
               write(sourceIndex, '(i3.3)') i+1
               outFile2=outFile
               if (posPoint .ne. 0) then
                  outFile2(posPoint:1024)=' '
                  write(outFile3, '(a,a,a)') trim(outFile2), sourceIndex, ".FIT"
               else
                  write(outFile3, '(a,a,a)') trim(outFile2), sourceIndex, ".FIT"
               endif

               write(blockName, '(a,a)') "RATE", sourceIndex
               write(blockNameL, '(a,a,a)') "RATE", sourceIndex, "L"
               write(blockNameR, '(a,a,a)') "RATE", sourceIndex, "R"

               !
               !  Create an additional file (or two files) for the lateral spectra
               !outFileL=outFile3
               !outFileR=outFile3

               !outSetH = dataSet(outFile3, create)
               !--- VNY 01.09.03
               !if (lateralSpectrumL) outSetLH=dataSet(outFileL, create)
               !if (lateralSpectrumR) outSetRH=dataSet(outFileR, create)
               !---
               ! The extracted region will have the full length of the y-column
               nBins=ny

               i1=int(sourceXCoord(iSource)+0.5)
               averLeft=real(i1)
               averRight=averLeft
               numLeft=0.0
               numRight=0.0

               deltaX=0.0
               ! The position of the corresponding zero-order is taken as reference
               xZero=-1.0
               yZero=-1.0
               if (sourceRelation(iSource) .gt. -1) then
                  i=sourceRelation(iSource)
                  deltaX=sourceYCoord(i)
                  ! XY-coordinates of the zero-order image
                  xZero=sourceXCoord(i)
                  yZero=sourceYCoord(i)
                  xZeroOrder=xZero
                  yZeroOrder=yZero
                  x0aux(i)=xZeroOrder
                  y0aux(i)=yZeroOrder
                  x1aux(i)=sourceXCoord(iSource)
                  y1aux(i)=sourceYCoord(iSource)
               endif
               !extractionMode=1
               subtractBkg=.true.
               call ExtractSpectrum(grismSetH, xZero, yZero, &
                    spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                    bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                    spectrumArray, spectrumErrArray, bkgArray, &
                    bkgErrArray, fluxArray, fluxErrArray, ysize, &
                    extractionMode, subtractBkg)
               !
               call EvaluateSpectrum(spectrumArray, bkgArray, ysize, &
                    iblueEdge, iredEdge, sbRatio)
! VNY 06.02.2004 Add table only if the spectrum significance is larger than 2.0
                  if (sbRatio .gt. 2.0 .or. omgsourceRegime) then
                     outTable0H=addTable(outSet0H, blockName, nBins, &
                          "Table with the source spectrum")
                     if (lateralSpectrumL) outTableLH=addTable(outSet0H, & 
                          blockNameL, nBins, "Table with the left lateral source spectrum")
                     if (lateralSpectrumR) outTableRH=addTable(outSet0H, &
                          blockNameR, nBins, "Table with the right lateral source spectrum")
                     !  Adding the spectrum number to the source list file
                     blkNum=blockNumber(outSet0H, blockName)
               
                     !spectrumNoArray(iSource)=blkNum+1
                     ! VNY 09.12.2003
                     jj = sourceRelation(iSource)
                     spectrumNoArray(jj)=sourceRelation(iSource)+1
                     spectrumOriginArray(iSource)=sourceRelation(iSource)+1
                     ! 
                     ! Computing the binning interval
                     if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                        binningInterval=2.40774*binY ! in Angstroms
                     else
                        binningInterval=5.626*binY
                     endif

                     call setAttribute(outTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                     call setAttribute(outTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                     call setAttribute(outTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                     call setAttribute(outTable0H, "HDUCLAS2", "NET", "Background subtracted")
                     call setAttribute(outTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                     !               call setAttribute(outTable0H, "TIMEDEL", real(10), "Angstrom", "binning interval")
                     call setAttribute(outTable0H, "TIMEDEL", binningInterval, "Angstrom", "binning interval")
                     !--Adding columns to the output table
                     waveSourceOutCol0 = addColumn(outTable0H, "WAVELENGTH", real32, &
                          "ANGSTROM", columnDim, "wavelengths ")
                     waveSourceOutArray0 => real32data(waveSourceOutCol0)
                     !
                     rateSourceOutCol0 = addColumn(outTable0H,"RATE", real32, &
                          "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                     rateSourceOutArray0 => real32data(rateSourceOutCol0)
                     !
                     rateSourceErrOutCol0 = addColumn(outTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                          columnDim, "source counts rate error")
                     rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                     !
                     rateBackOutCol0 = addColumn(outTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate")
                     rateBackOutArray0 => real32data(rateBackOutCol0)
                     !
                     rateBackErrOutCol0 = addColumn(outTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate error")
                     rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                     !
                     fluxOutCol0 = addColumn(outTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate")
                     fluxOutArray0 => real32data(fluxOutCol0)
                     !
                     fluxErrOutCol0 = addColumn(outTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate error")
                     fluxErrOutArray0 => real32data(fluxErrOutCol0)
                     !
                     
                     if (lateralSpectrumL) then ! Create lateral left-side spectrum
                        call setAttribute(outTableLH, "TIMEZERO",  0.d0, "s", &
                             "obs.start time relative to MJDREF")
                        call setAttribute(outTableLH, "HDUCLASS", "OGIP    ", &
                             "Format conforms to OGIP/GSFC conventions")
                        call setAttribute(outTableLH, "HDUCLAS1", "SPECTRUM", "Spectral data")
                        call setAttribute(outTableLH, "HDUCLAS2", "NET", "Background subtracted")
                        call setAttribute(outTableLH, "HDUCLAS3", "RATE", "Exposure corrected")
                        !                  call setAttribute(outTableLH, "TIMEDEL", real(10), "Angstrom", &
                        call setAttribute(outTableLH, "TIMEDEL", binningInterval, "Angstrom", &
                             "binning interval")
                        !--Adding columns to the output table
                        waveSourceOutColL = addColumn(outTableLH, "WAVELENGTH", real32, &
                             "ANGSTROM", columnDim, "wavelengths ")
                        waveSourceOutArrayL => real32data(waveSourceOutColL)

                        rateSourceOutColL = addColumn(outTableLH,"RATE", real32, &
                             "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                        rateSourceOutArrayL => real32data(rateSourceOutColL)

                        rateSourceErrOutColL = addColumn(outTableLH,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                             columnDim, "source counts rate error")
                        rateSourceErrOutArrayL => real32data(rateSourceErrOutColL)

                        rateBackOutColL = addColumn(outTableLH,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate")
                        rateBackOutArrayL => real32data(rateBackOutColL)
                        
                        rateBackErrOutColL = addColumn(outTableLH,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate error")
                        rateBackErrOutArrayL => real32data(rateBackErrOutColL)
                        ! 
                        fluxOutColL = addColumn(outTableLH,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate")
                        fluxOutArrayL => real32data(fluxOutColL)
                        !
                        fluxErrOutColL = addColumn(outTableLH,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate error")
                        fluxErrOutArrayL => real32data(fluxErrOutColL)
                        !
                     endif
                     if (lateralSpectrumR) then ! Create lateral right-side spectrum
                        !call copyAllAttributes(attributable(sourceListSetH), attributable(outSetRH))
                        !call setAttribute(outSetRH, "CONTENT", "OM GRISM SPECTRUM")
                        !call setAttribute(outSetRH, "SRCNO", srcNo, "n", "Source number")
                        !outTableRH = addTable(outSetRH, "RATE", nBins, "Table with the source spectrum")
                        call setAttribute(outTableRH, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                        call setAttribute(outTableRH, "HDUCLASS", "OGIP    ", &
                             "Format conforms to OGIP/GSFC conventions")
                        call setAttribute(outTableRH, "HDUCLAS1", "SPECTRUM", "Spectral data")
                        call setAttribute(outTableRH, "HDUCLAS2", "NET", "Background subtracted")
                        call setAttribute(outTableRH, "HDUCLAS3", "RATE", "Exposure corrected")
                        !                  call setAttribute(outTableRH, "TIMEDEL", real(10), "Angsroem", "binning interval")
                        call setAttribute(outTableRH, "TIMEDEL", binningInterval, "Angstrom", "binning interval")
                        !--Adding columns to the output table
                        waveSourceOutColR = addColumn(outTableRH, "WAVELENGTH", real32, "ANGSTROM", &
                             columnDim, "wavelengths ")
                        waveSourceOutArrayR => real32data(waveSourceOutColR)
                        
                        rateSourceOutColR = addColumn(outTableRH,"RATE", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, "source counts rate")
                        rateSourceOutArrayR => real32data(rateSourceOutColR)
                        
                        rateSourceErrOutColR = addColumn(outTableRH,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                             columnDim, "source counts rate error")
                        rateSourceErrOutArrayR => real32data(rateSourceErrOutColR)

                        rateBackOutColR = addColumn(outTableRH,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate")
                        rateBackOutArrayR => real32data(rateBackOutColR)

                        rateBackErrOutColR = addColumn(outTableRH,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                             columnDim, " background counts rate error")
                        rateBackErrOutArrayR => real32data(rateBackErrOutColR)
                        !
                        fluxOutColR = addColumn(outTableRH,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate")
                        fluxOutArrayR => real32data(fluxOutColR)
                        !
                        fluxErrOutColR = addColumn(outTableRH,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                             columnDim, " background counts rate error")
                        fluxErrOutArrayR => real32data(fluxErrOutColR)
                        !
                     endif

                     waveSourceOutArray0=wavelengthArray
                     rateSourceOutArray0=spectrumArray
                     rateSourceErrOutArray0=spectrumErrArray
                     rateBackOutArray0=bkgArray
                     rateBackErrOutArray0=bkgErrArray
                     fluxOutArray0=fluxArray
                     fluxErrOutArray0=fluxErrArray
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

                     if (lateralSpectrumL) then
                        waveSourceOutArrayL=wavelengthArray
                        do j=0,ny-1 
                           ! extraction of the lateral strips
                           aL0=0.0
                           aL1=0.0

                           do i=i1-int(latShift)-int(latwidth/2.0), i1-int(latShift)+int(latWidth/2.0)
                              if (i .ge. 0 .and. i .lt. nx .and. j .gt. 0 .and. j .le. ny) then 
                                 aL0=aL0+oswImage(i,j)
                                 aL1=aL1+1.0
                              endif
                           enddo

                           rateSourceOutArrayL(j)=aL0
                           rateSourceErrOutArrayL(j)=0.0
                           if (rateSourceOutArrayL(j) .gt. 1.0) rateSourceErrOutArrayL(j)= &
                                sqrt(rateSourceOutArrayL(j))
                           if (aL1 .ge. 2.0) rateSourceErrOutArrayL(j)=rateSourceErrOutArrayL(j)/sqrt(aL1-1.0)
                           fluxOutArrayL(j)= nullValue
                           fluxErrOutArrayL(j)= nullValue
                           
                           rateBackOutArrayL(j)=bkgVect2(j)
                           rateBackErrOutArrayL(j)=0.0
                           if (rateBackOutArrayL(j) .gt. 0.0) rateBackErrOutArrayL(j)=sqrt(rateBackOutArrayL(j))
                           !  Scaling counts
                           if (timeExposure .gt. 0.0) then
                              rateSourceOutArrayL(j)=rateSourceOutArrayL(j)/timeExposure
                              rateSourceErrOutArrayL(j)=rateSourceErrOutArrayL(j)/timeExposure
                              rateBackOutArrayL(j)=rateBackOutArrayL(j)/timeExposure
                              rateBackErrOutArrayL(j)=rateBackErrOutArrayL(j)/timeExposure
                           endif
                           if (countsScaleFactor(j) .gt. 0.0) then
                              rateSourceOutArrayL(j)=rateSourceOutArrayL(j)/countsScaleFactor(j)
                              rateSourceErrOutArrayL(j)=rateSourceErrOutArrayL(j)/countsScaleFactor(j)
                              rateBackOutArrayL(j)=rateBackOutArrayL(j)/countsScaleFactor(j)
                              rateBackErrOutArrayL(j)=rateBackErrOutArrayL(j)/countsScaleFactor(j)
                           endif
                        enddo
                     endif ! lateralSpectrumL

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                     if (lateralSpectrumR) then
                        waveSourceOutArrayR=wavelengthArray
                        do j=0,ny-1 
                           ! extraction of the lateral strips
                           aR0=0.0
                           aR1=0.0
                           do i=i1+int(latShift-latwidth/2.0), i1+int(latShift+latWidth/2.0)
                              if (i .ge. 0 .and. i .lt. nx .and. j .gt. 0 .and. j .le. ny) then
                                 aR0=aR0+oswImage(i,j)
                                 aR1=aR1+1.0
                              endif
                           enddo
                           rateSourceOutArrayR(j)=aR0
                           rateSourceErrOutArrayR(j)=0.0
                           if (rateSourceOutArrayR(j) .gt. 1.0) rateSourceErrOutArrayR(j)= &
                                sqrt(rateSourceOutArrayR(j))
                           if (aR1 .ge. 2.0) rateSourceErrOutArrayR(j)=rateSourceErrOutArrayR(j)/sqrt(aL1-1.0)
                           fluxOutArrayR(j)= nullValue
                           fluxErrOutArrayR(j)= nullValue

                           rateBackOutArrayR(j)=bkgVect2(j)
                           rateBackErrOutArrayR(j)=0.0
                           if (rateBackOutArrayR(j) .gt. 0.0) rateBackErrOutArrayR(j)=sqrt(rateBackOutArrayR(j))
                           
                           !  Scaling the counts
                           if (timeExposure .gt. 0.0) then
                              rateSourceOutArrayR(j)=rateSourceOutArrayR(j)/timeExposure
                              rateSourceErrOutArrayR(j)=rateSourceErrOutArrayR(j)/timeExposure
                              rateBackOutArrayR(j)=rateBackOutArrayR(j)/timeExposure
                              rateBackErrOutArrayR(j)=rateBackErrOutArrayR(j)/timeExposure
                           endif
                           if (countsScaleFactor(j) .gt. 0.0) then
                              rateSourceOutArrayR(j)=rateSourceOutArrayR(j)/countsScaleFactor(j)
                              rateSourceErrOutArrayR(j)=rateSourceErrOutArrayR(j)/countsScaleFactor(j)
                              rateBackOutArrayR(j)=rateBackOutArrayR(j)/countsScaleFactor(j)
                              rateBackErrOutArrayR(j)=rateBackErrOutArrayR(j)/countsScaleFactor(j)
                           endif
                        enddo
                     endif ! lateralSpectrumR

                     !
                     vignApp=.false.
                     deadApp=.false.
                     backApp=.true.
                     psfApp=.false.
                     call setAttribute(outTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                     call setAttribute(outTable0H, "NPIXSOU", sourceArea, "pix", "Source area")
                     !call setAttribute(outTable0H, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
                     !call setAttribute(outTable0H, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
                     call setAttribute(outTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                     call setAttribute(outTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                     call setAttribute(outTable0H, "BACKAPP", backApp,  "Background is subtracted")
                     call setAttribute(outTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            

                     call setAttribute(outTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                     call setAttribute(outTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                     call setAttribute(outTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                     call setAttribute(outTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                     call setAttribute(outTable0H, "QFLAG", sourceQF(0), " ",  "Quality flag")
                     call setAttribute(outTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                     !               call setAttribute(outTable0H, "XFIRST", sourceXCoord(iSource), &
                     call setAttribute(outTable0H, "XFIRST", xFirstOrder, &
                          "pix", "First-order (spectrum) image X-coordiate")
                     !               call setAttribute(outTable0H, "YFIRST", sourceYCoord(iSource), &
                     call setAttribute(outTable0H, "YFIRST", yFirstOrder, &
                          "pix", "First-order (spectrum) image Y-coordiate")
                     !               call setAttribute(outTable0H, "XZERO", xZero, &
                     call setAttribute(outTable0H, "XZERO", xZeroOrder, &
                          "pix", "Zero-order (object) image X-coordiate")
                     !               call setAttribute(outTable0H, "YZERO", yZero, &
                     call setAttribute(outTable0H, "YZERO", yZeroOrder, &
                          "pix", "Zero-order (object) image Y-coordiate")
                     call setAttribute(outTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                          "pix", "Width of the spectrum extraction region")
                     call setAttribute(outTable0H, "BKGREGWL", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (left)")
                     call setAttribute(outTable0H, "BKGREGWR", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (right)")
                     call setAttribute(outTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                          "pix", "Spectrum to backgr.reg.separation (left)")
                     call setAttribute(outTable0H, "BKGSEPR", abs(bkgoffsetright), &
                          "pix", "Spectrum to backgr.reg.separation (right)")

                     !---
                     !
                     if (lateralSpectrumL) then
                        call setAttribute(outTableLH, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                        call setAttribute(outTableLH, "NPIXSOU", sourceArea, "pix", "Source area")
                        !call setAttribute(outTableLH, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
                        !call setAttribute(outTableLH, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
                        call setAttribute(outTableLH, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                        call setAttribute(outTableLH, "DEADAPP", deadApp, " Dead time correction applied")
                        call setAttribute(outTableLH, "BACKAPP", backApp,  "Background is subtracted")
                        call setAttribute(outTableLH, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
                        
                        call setAttribute(outTableLH, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                        call setAttribute(outTableLH, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                        call setAttribute(outTableLH, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                        call setAttribute(outTableLH, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                        call setAttribute(outTableLH, "QFLAG", sourceQF(0), " ",  "Quality flag")
                        call setAttribute(outTableLH, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                        call setAttribute(outTableLH, "XFIRST", sourceXCoord(iSource), &
                             "pix", "First-order (spectrum) image X-coordiate")
                        call setAttribute(outTableLH, "YFIRST", sourceYCoord(iSource), &
                             "pix", "First-order (spectrum) image Y-coordiate")
                        call setAttribute(outTableLH, "XZERO", xZero, &
                             "pix", "Zero-order (object) image X-coordiate")
                        call setAttribute(outTableLH, "YZERO", yZero, &
                             "pix", "Zero-order (object) image Y-coordiate")
                        call setAttribute(outTableLH, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                             "pix", "Width of the spectrum extraction region")
                        call setAttribute(outTableLH, "BKGREGWL", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (left)")
                        call setAttribute(outTableLH, "BKGREGWR", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (right)")
                        call setAttribute(outTableLH, "BKGSEPL", abs(bkgoffsetleft), &
                             "pix", "Spectrum to backgr.reg.separation (left)")
                        call setAttribute(outTableLH, "BKGSEPR", abs(bkgoffsetright), &
                             "pix", "Spectrum to backgr.reg.separation (right)")
                     endif
!
                     if (lateralSpectrumR) then
                        call setAttribute(outTableRH, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                        call setAttribute(outTableRH, "NPIXSOU", sourceArea, "pix", "Source area")
                        !call setAttribute(outTableRH, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
                        !call setAttribute(outTableRH, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
                        call setAttribute(outTableRH, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                        call setAttribute(outTableRH, "DEADAPP", deadApp, " Dead time correction applied")
                        call setAttribute(outTableRH, "BACKAPP", backApp,  "Background is subtracted")
                        call setAttribute(outTableRH, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
                        
                        call setAttribute(outTableRH, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                        call setAttribute(outTableRH, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                        call setAttribute(outTableRH, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                        call setAttribute(outTableRH, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                        call setAttribute(outTableRH, "QFLAG", sourceQF(0), " ",  "Quality flag")
                        call setAttribute(outTableRH, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                        call setAttribute(outTableRH, "XFIRST", sourceXCoord(iSource), &
                             "pix", "First-order (spectrum) image X-coordiate")
                        call setAttribute(outTableRH, "YFIRST", sourceYCoord(iSource), &
                             "pix", "First-order (spectrum) image Y-coordiate")
                        call setAttribute(outTableRH, "XZERO", xZero, &
                             "pix", "Zero-order (object) image X-coordiate")
                        call setAttribute(outTableRH, "YZERO", yZero, &
                             "pix", "Zero-order (object) image Y-coordiate")
                        call setAttribute(outTableRH, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                             "pix", "Width of the spectrum extraction region")
                        call setAttribute(outTableRH, "BKGREGWL", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (left)")
                        call setAttribute(outTableRH, "BKGREGWR", abs(bkgwidthleft), &
                             "pix", "Backgr.extraction region width (right)")
                        call setAttribute(outTableRH, "BKGSEPL", abs(bkgoffsetleft), &
                             "pix", "Spectrum to backgr.reg.separation (left)")
                        call setAttribute(outTableRH, "BKGSEPR", abs(bkgoffsetright), &
                             "pix", "Spectrum to backgr.reg.separation (right)")
                     endif

                     ! release the dal handles
!---------
                     !---VNY 30.08.03
                     call release(rateBackErrOutCol0)
                     call release(rateBackOutCol0)
                     call release(rateSourceErrOutCol0)
                     call release(rateSourceOutCol0)
                     call release(fluxOutCol0)
                     call release(fluxErrOutCol0)
                     call release(waveSourceOutCol0)
                     call release(outTable0H)
               !---
                     if (lateralSpectrumL) then 
                        call release(rateBackErrOutColL)
                        call release(rateBackOutColL)
                        call release(rateSourceErrOutColL)
                        call release(rateSourceOutColL)
                        call release(fluxOutColL)
                        call release(fluxErrOutColL)
                        call release(waveSourceOutColL)
                        call release(outTableLH)
                        !call release(outSetLH)
                     endif
                     if (lateralSpectrumR) then
                        call release(rateBackErrOutColR)
                        call release(rateBackOutColR)
                        call release(rateSourceErrOutColR)
                        call release(rateSourceOutColR)
                        call release(fluxOutColR)
                        call release(fluxErrOutColR)
                        call release(waveSourceOutColR)
                        call release(outTableRH)
                        !call release(outSetRH)
                     endif

                  endif ! sbRatio .gt. 2.0

            endif ! sourceDeformation .gt. 10.0
         enddo ! iSource=0, nSource-1

         !*********************************************************
         ! Processing source by source for those point source that were not related
         ! to their spectra
         !*********************************************************

         do iSource=0, nSources-1
            xZeroOrder=0.0
            yZeroOrder=0.0
            xFirstOrder=0.0
            yFirstOrder=0.0
            if (sourceShape(iSource) .eq. 0 .and. sourceRelation(iSource) .lt. 0 &
                 .and. sourceFWHMaj(iSource) .lt. checkLength &
                 .and. abs(sourcePA(iSource)) .gt. 82.0 ) then ! Check only the point-like sources
               i1=int(sourceXCoord(iSource)+0.5)
               ! The position of the corresponding zero-order is taken as reference
               xZeroA=sourceXCoord(iSource)
               yZeroA=sourceYCoord(iSource)
               xZeroOrder=xZeroA
               yZeroOrder=yZeroA
               yFirst=sourceYCoord(iSource)+source2spectrum ! position of the first order
               xFirstOrder=xZeroOrder
               yFirstOrder=yFirst
!
               iyUp=int(yFirst)
               iyDown=int(yFirst-spectrumLength/2.0)
               availableSpectrumLength=0.0
               availableSpectrumFraction=0.0

               if (iyDown .gt. int(yZeroA) .and. iyDown .lt. ny-1 .and. iyUp .lt. ny-1) then 
                  if (oswImage(i1,iyDown) .gt. -0.9 .and. oswImage(i1,iyUp) .gt. -0.9) then
                     ! determine the upper available limit
                     availableSpectrumLength=0.0
                     checkUp: do iyUp2=iyDown+1,iyDown+int(spectrumLength)
                        if (iyUp2 .lt. ny) then
                           if (oswImage(i1,iyUp2) .gt. -0.9) then
                              availableSpectrumLength=availableSpectrumLength+1.0
                           else
                              exit checkUp
                           endif
                        endif
                     enddo checkUp
                  endif
               endif
               availableSpectrumFraction=availableSpectrumLength/spectrumLength

               ! Extract the spectrum only if more than 70 percent of the spectrum is available
               if (availableSpectrumFraction .gt. 0.7) then !extract
                  sourceArea=0.0
                  srcNo=iSource
                  timeExposure = exposure
! VNY 09.12.2003

                  nBins=ny

! VNY 01.02.2004
                  x0aux(iSource)=xZeroOrder
                  y0aux(iSource)=yZeroOrder
                  x1aux(iSource)=xFirstOrder
                  y1aux(iSource)=yFirstOrder

                  !extractionMode=1
                  subtractBkg=.true.
                  call ExtractSpectrum(grismSetH, xZeroA, yZeroA, &
                       spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                       bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                       spectrumArray, spectrumErrArray, bkgArray, &
                       bkgErrArray, fluxArray, fluxErrArray, ysize, &
                       extractionMode, subtractBkg)
                  ! VNY 06.02.2004
                  blueEdge=yZeroA+region1
                  redEdge=yZeroA+region2
                  iblueEdge=int(blueEdge)
                  iredEdge=int(redEdge)
                  call EvaluateSpectrum(spectrumArray, bkgArray, ysize, &
                       iblueEdge, iredEdge, sbRatio)


                  if (sbRatio .gt. 2.0) then
                     ! First check if there was some first-order and mark this first-order as already identified
                     do i=0, nSources-1
                        if (sourceShape(i) .eq. 1) then 
                           if (i .ne. iSource .and. sourceRelation(i) .lt. 0) then
                              if (sourceXcoord(i) .gt. xZeroA-shapeCriterium2 &
                                   .and. sourceXcoord(i) .lt. xZeroA+shapeCriterium2) then
                                 if (sourceYcoord(i) .gt. blueEdge .and. &
                                      sourceYcoord(i) .lt. redEdge) then
                                    ! Label the source
                                    sourceRelation(i)=iSource
                                 endif
                              endif
                           endif ! i .ne. iSource
                        endif ! sourceShape(i) .eq. 1
                     enddo ! i=0,nSources-1
                     spectrumNoArray(iSource)=iSource+1
                     write(sourceIndex, '(i3.3)') iSource+1
                     write(blockName, '(a,a)') "RATE", sourceIndex
                     outTable0H=addTable(outSet0H, blockName, nBins, &
                          "Table with the source spectrum")
                     !  Adding the spectrum number to the source list file
                     blkNum=blockNumber(outSet0H, blockName)
                     !spectrumNoArray(iSource)=blkNum+1
                     !spectrumOriginArray(iSource)=sourceRelation(iSource)+1
                     ! 
                     ! Computing the binning interval
                     if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                        binningInterval=2.40774*binY ! in Angstroms
                     else
                        binningInterval=5.626*binY
                     endif
                     !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                     call setAttribute(outTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                     call setAttribute(outTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                     call setAttribute(outTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                     call setAttribute(outTable0H, "HDUCLAS2", "NET", "Background subtracted")
                     call setAttribute(outTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                     !               call setAttribute(outTable0H, "TIMEDEL", real(10), "Angstrom", "binning interval")
                     call setAttribute(outTable0H, "TIMEDEL", binningInterval, "Angstrom", "binning interval")
                     !--Adding columns to the output table
                     waveSourceOutCol0 = addColumn(outTable0H, "WAVELENGTH", real32, &
                          "ANGSTROM", columnDim, "wavelengths ")
                     waveSourceOutArray0 => real32data(waveSourceOutCol0)
                     !
                     rateSourceOutCol0 = addColumn(outTable0H,"RATE", real32, &
                          "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                     rateSourceOutArray0 => real32data(rateSourceOutCol0)
                     !
                     rateSourceErrOutCol0 = addColumn(outTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                          columnDim, "source counts rate error")
                     rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                     !
                     rateBackOutCol0 = addColumn(outTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate")
                     rateBackOutArray0 => real32data(rateBackOutCol0)
                     !
                     rateBackErrOutCol0 = addColumn(outTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate error")
                     rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                     !
                     fluxOutCol0 = addColumn(outTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate")
                     fluxOutArray0 => real32data(fluxOutCol0)
                     !
                     fluxErrOutCol0 = addColumn(outTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate error")
                     fluxErrOutArray0 => real32data(fluxErrOutCol0)
                     !

                     !
                     waveSourceOutArray0=wavelengthArray
                     rateSourceOutArray0=spectrumArray
                     rateSourceErrOutArray0=spectrumErrArray
                     rateBackOutArray0=bkgArray
                     rateBackErrOutArray0=bkgErrArray
                     fluxOutArray0=fluxArray
                     fluxErrOutArray0=fluxErrArray

                     !
                     vignApp=.false.
                     deadApp=.false.
                     backApp=.true.
                     psfApp=.false.
                     call setAttribute(outTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                     call setAttribute(outTable0H, "NPIXSOU", sourceArea, "pix", "Source area")
                     !call setAttribute(outTable0H, "CLRADIUS", calApertureRadius, "pix", "Coincidence loss correction radius")
                     !call setAttribute(outTable0H, "PHRADIUS", photometryApertureRadius,"pix", "Photometry aperture radius")
                     call setAttribute(outTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                     call setAttribute(outTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                     call setAttribute(outTable0H, "BACKAPP", backApp,  "Background is subtracted")
                     call setAttribute(outTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            

                     call setAttribute(outTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                     call setAttribute(outTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                     call setAttribute(outTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                     call setAttribute(outTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                     call setAttribute(outTable0H, "QFLAG", sourceQF(0), " ",  "Quality flag")
                     call setAttribute(outTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                     !                  call setAttribute(outTable0H, "XFIRST", sourceXCoord(iSource), &
                     call setAttribute(outTable0H, "XFIRST", xFirstOrder, &
                          "pix", "First-order (spectrum) image X-coordiate")
                     !                  call setAttribute(outTable0H, "YFIRST", sourceYCoord(iSource), &
                     call setAttribute(outTable0H, "YFIRST", yFirstOrder, &
                          "pix", "First-order (spectrum) image Y-coordiate")
                     call setAttribute(outTable0H, "XZERO", xZeroOrder, &
                          "pix", "Zero-order (object) image X-coordiate")
                     call setAttribute(outTable0H, "YZERO", yZeroOrder, &
                          "pix", "Zero-order (object) image Y-coordiate")
                     call setAttribute(outTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                          "pix", "Width of the spectrum extraction region")
                     call setAttribute(outTable0H, "BKGREGWL", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (left)")
                     call setAttribute(outTable0H, "BKGREGWR", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (right)")
                     call setAttribute(outTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                          "pix", "Spectrum to backgr.reg.separation (left)")
                     call setAttribute(outTable0H, "BKGSEPR", abs(bkgoffsetright), &
                          "pix", "Spectrum to backgr.reg.separation (right)")

                     ! release the dal handles
                     !---------
                     !---VNY 30.08.03
                     call release(rateBackErrOutCol0)
                     call release(rateBackOutCol0)
                     call release(rateSourceErrOutCol0)
                     call release(rateSourceOutCol0)
                     call release(fluxOutCol0)
                     call release(fluxErrOutCol0)
                     call release(waveSourceOutCol0)
                     call release(outTable0H)
                     !---

                  endif ! sbRatio .gt. 2.0

               endif ! if availableSpectrumFraction .gt. 0.6
            endif ! sourceShape .gt. 0
         enddo ! iSource=0, nSources-1

         !*********************************************************
         ! Processing source by source those extended sources (spectra) 
         ! that were not related to their zero-orders
         !*********************************************************
         do iSource=0, nSources-1
            !xZeroOrder=0.0
            !yZeroOrder=0.0
            !xFirstOrder=0.0
            !yFirstOrder=0.0
            if (sourceShape(iSource) .eq. 1 .and. sourceRelation(iSource) .lt. 0) then
               spectrumHalfLength=sourceFWHMaj2(iSource)
               if(sourceYCoord(iSource)-spectrumHalfLength-region1 .gt. 0) then 
                  ! Trying to find a zero-order
                  yZero=sourceYCoord(iSource)-sourceFWHMaj2(iSource)-region1
                  xFirst=sourceXCoord(iSource)
                  yFirst=sourceYCoord(iSource)
                  verbFlag=0
                  if (iSource .eq. 50) verbFlag=1
!
                  call FindZeroOrder(grismSetH, xFirst, yFirst, &
                       spectrumHalfLength, yZero, zeroMaxValue, &
                       spectrumHalfWidth, shapeCriterium1, verbFlag, &
                       maxValue2, meanValue2, yZero2)

                  if (yZero .gt. 0 .and. maxValue2/meanValue2 .gt. 2.0) then
                     ! extract the spectrum
                     subtractBkg=.true.
                     xZeroA=xFirst
                     yZeroA=yZero
                     x0aux(iSource)=xZeroA
                     y0aux(iSource)=yZeroA
                     x1aux(iSource)=xFirst
                     y1aux(iSource)=yFirst

                     call ExtractSpectrum(grismSetH, xZeroA, yZeroA, &
                          spectrumHalfWidth, bkgOffsetLeft, bkgWidthLeft, &
                          bkgOffsetRight, bkgWidthRight, wavelengthArray, &
                          spectrumArray, spectrumErrArray, bkgArray, &
                          bkgErrArray, fluxArray, fluxErrArray, ysize, &
                          extractionMode, subtractBkg)

                     ! Include the spectrum into the output file
                     spectrumNoArray(iSource)=iSource+1
                     write(sourceIndex, '(i3.3)') iSource+1
                     write(blockName, '(a,a)') "RATE", sourceIndex
                     outTable0H=addTable(outSet0H, blockName, nBins, &
                          "Table with the source spectrum")
                     !  Adding the spectrum number to the source list file
                     blkNum=blockNumber(outSet0H, blockName)
                     if (filter .eq. "Grism1  " .or. filter .eq. "GRISM1  ")  then ! (UV grism)
                        binningInterval=2.40774*binY ! in Angstroms
                     else
                        binningInterval=5.626*binY
                     endif
                     !^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                     call setAttribute(outTable0H, "TIMEZERO",  0.d0, "s", "obs.start time relative to MJDREF")
                     call setAttribute(outTable0H, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
                     call setAttribute(outTable0H, "HDUCLAS1", "SPECTRUM", "Spectral data")
                     call setAttribute(outTable0H, "HDUCLAS2", "NET", "Background subtracted")
                     call setAttribute(outTable0H, "HDUCLAS3", "RATE", "Exposure corrected")
                     !               call setAttribute(outTable0H, "TIMEDEL", real(10), "Angstrom", "binning interval")
                     call setAttribute(outTable0H, "TIMEDEL", binningInterval, "Angstrom", "binning interval")
                     !--Adding columns to the output table
                     waveSourceOutCol0 = addColumn(outTable0H, "WAVELENGTH", real32, &
                          "ANGSTROM", columnDim, "wavelengths ")
                     waveSourceOutArray0 => real32data(waveSourceOutCol0)
                     !
                     rateSourceOutCol0 = addColumn(outTable0H,"RATE", real32, &
                          "COUNTS/S/ANGSTROM", columnDim, "source counts rate")
                     rateSourceOutArray0 => real32data(rateSourceOutCol0)
                     !
                     rateSourceErrOutCol0 = addColumn(outTable0H,"ERROR", real32,"COUNTS/S/ANGSTROM", &
                          columnDim, "source counts rate error")
                     rateSourceErrOutArray0 => real32data(rateSourceErrOutCol0)
                     !
                     rateBackOutCol0 = addColumn(outTable0H,"BACKV", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate")
                     rateBackOutArray0 => real32data(rateBackOutCol0)
                     !
                     rateBackErrOutCol0 = addColumn(outTable0H,"BACKE", real32, "COUNTS/S/ANGSTROM", &
                          columnDim, " background counts rate error")
                     rateBackErrOutArray0 => real32data(rateBackErrOutCol0)
                     !
                     fluxOutCol0 = addColumn(outTable0H,"FLUX", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate")
                     fluxOutArray0 => real32data(fluxOutCol0)
                     !
                     fluxErrOutCol0 = addColumn(outTable0H,"FLUX_ERR", real32, "ERG/S/CM**2/ANGSTROM", &
                          columnDim, " background counts rate error")
                     fluxErrOutArray0 => real32data(fluxErrOutCol0)
                     !

                     !
                     waveSourceOutArray0=wavelengthArray
                     rateSourceOutArray0=spectrumArray
                     rateSourceErrOutArray0=spectrumErrArray
                     rateBackOutArray0=bkgArray
                     rateBackErrOutArray0=bkgErrArray
                     fluxOutArray0=fluxArray
                     fluxErrOutArray0=fluxErrArray

                     !
                     vignApp=.false.
                     deadApp=.false.
                     backApp=.true.
                     psfApp=.false.
                     call setAttribute(outTable0H, "EXPOSURE", timeExposure, "s", "Duration of the exposure")
                     call setAttribute(outTable0H, "NPIXSOU", sourceArea, "pix", "Source area")

                     call setAttribute(outTable0H, "VIGNAPP", vignApp,  "Rates corrected for vignetting")
                     call setAttribute(outTable0H, "DEADAPP", deadApp, " Dead time correction applied")
                     call setAttribute(outTable0H, "BACKAPP", backApp,  "Background is subtracted")
                     call setAttribute(outTable0H, "PSF_CAPP", psfApp, "PSF-fraction correction applied")            
                     
                     call setAttribute(outTable0H, "RA_OBJ", sourceRA(0), "deg", "Right Ascension of the source")
                     call setAttribute(outTable0H, "DEC_OBJ", sourceDEC(0), "deg",  "Declination of the source")
                     call setAttribute(outTable0H, "LII", sourceLII(0), "deg", "Galactic Longitude of the source")
                     call setAttribute(outTable0H, "BII", sourceBII(0), "deg",  "Galactic Lattitude of the source")
                     call setAttribute(outTable0H, "QFLAG", sourceQF(0), " ",  "Quality flag")
                     call setAttribute(outTable0H, "CFLAG", sourceCF(0), " ",  "Confusion flag")
                     !                  call setAttribute(outTable0H, "XFIRST", sourceXCoord(iSource), &
                     call setAttribute(outTable0H, "XFIRST", xFirstOrder, &
                          "pix", "First-order (spectrum) image X-coordiate")
                     !                  call setAttribute(outTable0H, "YFIRST", sourceYCoord(iSource), &
                     call setAttribute(outTable0H, "YFIRST", yFirstOrder, &
                          "pix", "First-order (spectrum) image Y-coordiate")
                     call setAttribute(outTable0H, "XZERO", xZeroOrder, &
                          "pix", "Zero-order (object) image X-coordiate")
                     call setAttribute(outTable0H, "YZERO", yZeroOrder, &
                          "pix", "Zero-order (object) image Y-coordiate")
                     call setAttribute(outTable0H, "SPECREGW", 2.0*abs(spectrumHalfWidth), &
                          "pix", "Width of the spectrum extraction region")
                     call setAttribute(outTable0H, "BKGREGWL", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (left)")
                     call setAttribute(outTable0H, "BKGREGWR", abs(bkgwidthleft), &
                          "pix", "Backgr.extraction region width (right)")
                     call setAttribute(outTable0H, "BKGSEPL", abs(bkgoffsetleft), &
                          "pix", "Spectrum to backgr.reg.separation (left)")
                     call setAttribute(outTable0H, "BKGSEPR", abs(bkgoffsetright), &
                          "pix", "Spectrum to backgr.reg.separation (right)")
                     
                     ! release the dal handles
                     !---------
                     !---VNY 30.08.03
                     call release(rateBackErrOutCol0)
                     call release(rateBackOutCol0)
                     call release(rateSourceErrOutCol0)
                     call release(rateSourceOutCol0)
                     call release(fluxOutCol0)
                     call release(fluxErrOutCol0)
                     call release(waveSourceOutCol0)
                     call release(outTable0H)
                     !---
                     
                  endif ! if yZero .gt. 0
               endif
            endif
         enddo ! iSource=0, nSources-1


         ! Check the source list
         nSpectraSources=0
         do iSource=0, nSources-1
            if (spectrumNoArray(iSource) .lt. 99999 .and. spectrumNoArray(iSource) .gt. 0) then
               nSpectraSources=nSpectraSources+1
            endif
         enddo ! iSource=0, nSources-1
!
!   VNY 28.01.2004
!**********************************************************************
         !  Create a file containing list of the extracted spectra for             
!**********************************************************************   
         outSpectraListSetH = dataSet(spectraListFile, create)
         call setAttribute(outSpectraListSetH, "CONTENT", "OM GRISM SPECTRA LIST", &
              "File contains the list of the extracted OM grsim spectra")
         call setAttribute(outSpectraListSetH, "HDUCLASS", "OGIP    ", &
              "Format conforms to OGIP/GSFC conventions")
         call setAttribute(outSpectraListSetH, "HDUCLAS1", "SRCLIST", &
              "File contains a source/spectra list")
         call setAttribute(outSpectraListSetH, "HDUVERS", "1.2.0    ", &
              "The version number of the format")
         call setAttribute(outSpectraListSetH, "TELESCOP", "XMM     ", &
              "XMM mission")
         call setAttribute(outSpectraListSetH, "INSTRUME", "OM      ", &
              "Instrument name")
         call setAttribute(outSpectraListSetH, "DATAMODE", dataMode, &
              "Data mode")
         call setAttribute(outSpectraListSetH, "FILTER", filter, &
              "Filter used for obtaining the data")
         call setAttribute(outSpectraListSetH, "FRAMTIME", frameTime, &
              "Frame time")
         call setAttribute(outSpectraListSetH, "DEADFRAC", deadFrac, &
              "Frame dead fraction")
         call setAttribute(outSpectraListSetH, "TMEXP_ID", tmExpId, &
              "Exposure Id from telemetry")
         call setAttribute(outSpectraListSetH, "FRMTIME", frmTime, &
              "Frametime from telemetry")
         call setAttribute(outSpectraListSetH, "DATATYPE", dataType, &
              "Data type")
         call setAttribute(outSpectraListSetH, "OBS_ID", obsId, &
              "Observation identifier")
         call setAttribute(outSpectraListSetH, "EXP_ID", expId, &
              "Exposure identifier")
         call setAttribute(outSpectraListSetH, "DATE-OBS", dateObs, &
              "Observation start time")
         call setAttribute(outSpectraListSetH, "DATE-END", dateEnd, &
              "Observation end time")

         ! Add columns with spectra identification
         nBins=nSpectraSources
         write(blockName, '(a)') "SRCLIST"
         outSpectraTableH=addTable(outSpectraListSetH, blockName, nBins, &
              "Table with the extracted spectra list")

         ! Add columns with the spectrum identification number to the (input) source list table 
         if (.not. hasColumn(outSpectraTableH, "SPECTR_NO")) then
            specNoListCol = addColumn(outSpectraTableH, "SPECTR_NO", integer32, "", columnDim, "spectrum No ")
         else
            specNoListCol = column(outSpectraTableH, "SPECTR_NO", modify)
         endif
         specNoListArray => int32data(specNoListCol)
         call setNullValue(specNoListCol, 999999 )
 
         do iSource=0, nSpectraSources-1
            call setToNull(specNoListCol, iSource)
         enddo

         ! X coordinates of the zero-order
         if (.not. hasColumn(outSpectraTableH, "X0")) then
            specX0Col = addColumn(outSpectraTableH, "X0", real32, "", columnDim, "Zero-order X-coord.")
         else
            specX0Col = column(outSpectraTableH, "X0", modify)
         endif
         specX0Array => real32data(specX0Col)
         !call setNullValue(specX0Col, 999999 )
 
         !do iSource=0, nSpectraSources-1
         !   call setToNull(specX0Col, iSource)
         !enddo
         specX0Array=0.0

         ! Y coordinates of the zero-order
         if (.not. hasColumn(outSpectraTableH, "Y0")) then
            specY0Col = addColumn(outSpectraTableH, "Y0", real32, "", columnDim, "Zero-order Y-coord.")
         else
            specY0Col = column(outSpectraTableH, "Y0", modify)
         endif
         specY0Array => real32data(specY0Col)
         !call setNullValue(specY0Col, 999999 )
 
         !do iSource=0, nSpectraSources-1
         !   call setToNull(specY0Col, iSource)
         !enddo
         specY0Array=0.0

         ! X coordinates of the first-order
         if (.not. hasColumn(outSpectraTableH, "X1")) then
            specX1Col = addColumn(outSpectraTableH, "X1", real32, "", columnDim, "First-order X-coord.")
         else
            specX1Col = column(outSpectraTableH, "X1", modify)
         endif
         specX1Array => real32data(specX1Col)
         !call setNullValue(specX1Col, 999999 )
 
         !do iSource=0, nSpectraSources-1
         !   call setToNull(specX1Col, iSource)
         !enddo
         specX1Array=0.0

         ! Y coordinates of the first-order
         if (.not. hasColumn(outSpectraTableH, "Y1")) then
            specY1Col = addColumn(outSpectraTableH, "Y1", real32, "", columnDim, "First-order Y-coord.")
         else
            specY1Col = column(outSpectraTableH, "Y1", modify)
         endif
         specY1Array => real32data(specY1Col)
         !call setNullValue(specY1Col, 999999 )
 
         !do iSource=0, nSpectraSources-1
         !   call setToNull(specY1Col, iSource)
         !enddo
         specY1Array=0.0
         
         ! Add columns with the spectrum identification number to the (input) source list table 
         if (.not. hasColumn(outSpectraTableH, "SPECTR_ID")) then
            specIdCol = addColumn(outSpectraTableH, "SPECTR_ID", integer32, "", &
                 columnDim, "spectrum identifier, as in the SWSRLI-file")
         else
            specIdCol = column(outSpectraTableH, "SPECTR_ID", modify)
         endif
         specIdArray => int32data(specIdCol)
         call setNullValue(specIdCol, 999999 )
 
         do iSource=0, nSpectraSources-1
            call setToNull(specIdCol, iSource)
         enddo

         ! The column with the spectrum origin (the number of the corresponding point-like source)
!         if (.not. hasColumn(outSpectraTableH, "ORIGIN_ID")) then
!            specOriginListCol = addColumn(outSpectraTableH, "ORIGIN_ID", integer32, "", columnDim, "spectrum origin number ")
!         else
!            specOriginListCol = column(outSpectraTableH, "ORIGIN_ID", modify)
!         endif
!         specOriginListArray => int32data(specOriginListCol)
!         call setNullValue(specOriginListCol, 999999 )
!         do iSource=0, nSpectraSources-1
!            call setToNull(specOriginListCol, iSource)
!         enddo
         do iSource=0, nSpectraSources-1
            specNoListArray(iSource)=iSource+1
         enddo
         ! Fill in the table of the extracted spectra
         nAux=0
         do iSource=0, nSources-1
            if (spectrumNoArray(iSource) .lt. 99999 .and. spectrumNoArray(iSource) .gt. 0) then
               specIdArray(nAux)=spectrumNoArray(iSource)
               !specOriginListArray(nAux)=spectrumOriginArray(iSource)
               specX0Array(nAux)=x0aux(iSource)
               specX1Array(nAux)=x1aux(iSource)
               specY0Array(nAux)=y0aux(iSource)
               specY1Array(nAux)=y1aux(iSource)
               nAux=nAux+1
            endif
         enddo ! iSource=0, nSources-1

!***********************************************************************
         !*************************
         ! Release handles on data
         !*************************
         call release(outSet0H)
         call release(grismSetH)
         do i=1, noColumns
            if (inColumnFlag(i) .eq. 1) then
               call release(inColumnH(i))
            endif
         end do
         ! VNY 01.09.2003
         call release(spectrumNoCol)
         call release(spectrumOriginCol)
         ! VNY 01.02.2004
         call release(specNoListCol)
         call release(specX0Col)
         call release(specX1Col)
         call release(specY0Col)
         call release(specY1Col)
         call release(specIdCol)
         !call release(specOriginListCol)
         call release(outSpectraTableH)
         call release(outSpectraListSetH)
         !
         !---VNY 02.09.2003
         deallocate(sourceFWHMaj2)
         deallocate(sourceFWHMin2)
         deallocate(sourceXCoord)
         deallocate(sourceYCoord)
         !
         deallocate(limitLeftMax)
         deallocate(limitRightMax)
         deallocate(sourceShape)
         deallocate(sourceRelation)
!
         deallocate(x0aux)
         deallocate(x1aux)
         deallocate(y0aux)
         deallocate(y1aux)
!
         deallocate(bkgVect)
         deallocate(bkgVect2)
         deallocate(bkgVect3)
         deallocate(bkgVectR)
         deallocate(bkgVectL)
         deallocate(smoothBkgVectR)
         deallocate(smoothBkgVectL)
         deallocate(specVect)
         deallocate(countsScaleFactor)
         ! VNY 17.11.2003
         deallocate (wavelengthArray)
         deallocate (spectrumArray)
         deallocate (spectrumErrArray)
         deallocate (bkgArray)
         deallocate (bkgErrArray)
         deallocate (fluxArray)
         deallocate (fluxErrArray)

      endif ! nSources>0

      ! VNY 14.10.2003 Deleting the columns in the input source list
      !************************************************************************
      if (hasColumn(inTableH, "RATE")) then
         call deleteColumn(inTableH, "RATE")
      endif
      if (hasColumn(inTableH, "RATE_ERR")) then
         call deleteColumn(inTableH, "RATE_ERR")
      endif
      if (hasColumn(inTableH, "BACKGROUND_RATE")) then
         call deleteColumn(inTableH, "BACKGROUND_RATE")
      endif
      if (hasColumn(inTableH, "SIGNIFICANCE")) then
         call deleteColumn(inTableH, "SIGNIFICANCE")
      endif
      if (hasColumn(inTableH, "CORR_RATE")) then
         call deleteColumn(inTableH, "CORR_RATE")
      endif
      if (hasColumn(inTableH, "CORR_RATE_ERR")) then
         call deleteColumn(inTableH, "CORR_RATE_ERR")
      endif
      if (hasColumn(inTableH, "CORR_BACKGROUND_RATE")) then
         call deleteColumn(inTableH, "CORR_BACKGROUND_RATE")
      endif
      if (hasColumn(inTableH, "PSF1_CORR")) then
         call deleteColumn(inTableH, "PSF1_CORR")
      endif
      if (hasColumn(inTableH, "COINCLOSS_CORR")) then
         call deleteColumn(inTableH, "COINCLOSS_CORR")
      endif
      if (hasColumn(inTableH, "APERTURE")) then
         call deleteColumn(inTableH, "APERTURE")
      endif
      !---^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      call release(inTableH)
   else ! hasBlock(sourceListSetH)
      !************************************
      ! Incompatible file - can't continue
      !************************************
      write (errstr, *) "Could not find any compatibile blocks in set ", sourceListFile
      call message(AppMsg, 2)
   endif ! hasBlock(sourceListSetH)
   call release(sourceListSetH)
   !      call release(sourceRegSetH)
!   deallocate(siginput)
 end subroutine omgrism

!*************************************************************
! Subroutines
!*************************************************************

subroutine AddReal32Column(tableSet, array, size, heading, &
        units, comment, nullValue, position)
    type(TableT), intent(in) :: tableSet
    character(len=*), intent(in) :: heading, units, comment
    integer, intent(in) :: size
    real(kind=single), intent(in), dimension(0 : size - 1) :: array
    real(kind=single), intent(in) :: nullValue
    integer(kind=int32), intent(in), optional :: position
    real(kind=single), pointer :: real32Col(:)
    type(ColumnT) :: pColumn
    integer i
    integer :: ndims(1) = (/1/)


     if (.not. hasColumn(tableSet, trim(heading))) then
        if (present(position)) then
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment), position)
        else 
            pColumn = addColumn(tableSet, trim(heading), real32, trim(units), ndims, trim(comment))
        end if
    else
        pColumn = column(tableSet, trim(heading), modify)
    end if
   

    real32Col => real32Data(pcolumn)

    do i = 0 , size - 1
        if (array(i) .eq. nullValue) then
            call setToNull(pcolumn, i)
        else
            real32Col(i) = array(i)
        end if
    enddo
    call release(pcolumn)

end subroutine AddReal32Column

!----
end module omgrism_mod











